/*****************************************************************************
 * File: sak2v2_serial.c
 * Author: Jonathan Gips and Ryan Aylward, MIT Wearable Computing Group
 * http://www.media.mit.edu/wearables/
 * Adapted from hackfest_serial.c, hoarder_serial.c
 * Copyright (C) 2003 MIT Media Lab
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *****************************************************************************/

#include "sak2v2_serial.h"
#include "sak2v2_protocol.h"
#include <sys/time.h>
  
volatile int STOP=FALSE; 
int fd;
struct termios oldtio,newtio;

#define SAK2V2_MAX_BODYLEN 256
#define SAK2V2_NUM_PACKS_TO_WAIT 5

sak2v2_error sak2Error;

typedef struct sak2v2_device {
  int devType;
  int devID;
  int readAddr;
  int bps;
  int spp;
  int rate;
  union {
    int configured;
    int requested;
  } state;

} sak2v2_device;

typedef struct sak2v2_packet {
  char messageID;
  char bodyLen;
  unsigned char body[SAK2V2_MAX_BODYLEN];
} sak2v2_packet;


static int sak2BaseClock = 0;
static sak2v2_packet sampPack;
static struct timeval sampPackTime;
static int bGotSample = 0;
static int bStreaming = 0;


static sak2v2_device sak2DevList[SAK2V2_NUM_DEVICES];

static sak2v2_device hostDevList[SAK2V2_NUM_DEVICES] = { { SAK2V2_DTYPE_I2C, 0xB0, 2, 4, 0, 22, {0}}, /* SAK2V2_ACCEL_0 */
							 { SAK2V2_DTYPE_I2C, 0xB2, 2, 4, 0, 22, {0}}, /* SAK2V2_ACCEL_1 */
							 { SAK2V2_DTYPE_I2C, 0xA2, 2, 4, 0, 400, {0}}, /* SAK2V2_TAG_READER */
							 { SAK2V2_DTYPE_SPECIAL, 0, 0, 2, 0, 100, {0}}, /* SAK2V2_POLAR */
							 { SAK2V2_DTYPE_ANALOG, 0, 0, 1, 0, 10, {0}}, /* SAK2V2_EKG */
							 { SAK2V2_DTYPE_ANALOG, 1, 0, 1, 0, 10, {0}}, /* SAK2V2_GSR */
							 { SAK2V2_DTYPE_ANALOG, 2, 0, 1, 0, 100, {0}}};/* SAK2V2_TEMP */

static int devMap[SAK2V2_NUM_DEVICES]; //map host dev enum to the index in sak2 device list
static int bufOffset[SAK2V2_NUM_DEVICES]; //num bytes to offset into a sample packet to get a sak2 device's data

static int SpinRead(char *buffer, int size);
static int SpinWrite(char *buffer, int size);
static int GetAckOrNack(int *ack);
static int ReadPacket(sak2v2_packet *pack);
static int WritePacket(sak2v2_packet *pack);
static void ProcessSensorPacket(sak2v2_packet *pack);
static void ProcessMetaSensorPacket(sak2v2_packet *pack);
static void UpdateDevMap();

int sak2v2_init_serial(){
  int i;

  fd = open(MODEMDEVICE, O_RDWR | O_NOCTTY ); 
  if (fd <0) {perror(MODEMDEVICE); exit(-1); }
  /* this is a comment */
  tcgetattr(fd,&oldtio); /* save current port settings */
  
  bzero(&newtio, sizeof(newtio));
  newtio.c_cflag = BAUDRATE | CS8 | CLOCAL | CREAD;
  newtio.c_iflag = IGNPAR;
  newtio.c_oflag = 0;
  
  /* set input mode (non-canonical, no echo,...) */
  newtio.c_lflag = 0;
   
  newtio.c_cc[VTIME]    = 0;   /* inter-character timer unused */
  newtio.c_cc[VMIN]     = 1;   /* blocking read until 1 chars received */
  //was 1
  
  tcflush(fd, TCIFLUSH);
  tcsetattr(fd,TCSANOW,&newtio);

  for(i = 0; i < SAK2V2_NUM_DEVICES; i++)
    {
      devMap[i] = -1;
    }

  return 1;
}

int sak2v2_close_serial()
{
  tcsetattr(fd,TCSANOW,&oldtio);
  
  return 1;
}

int sak2v2_set_device(sak2v2_devName devEnum, int enable)
{
  if(devEnum == SAK2V2_NUM_DEVICES) return 0;

  hostDevList[devEnum].state.requested = enable;

  return 1;
}
 
int sak2v2_configure_devices()
{ 
  sak2v2_packet pack;
  int ack, idx, i;

  pack.messageID = SAK2V2_CONFIG_C;
  pack.body[0] = SAK2V2_BASE_SAMP_CLOCK;
  
  idx = 1;

  for(i = 0; i < SAK2V2_NUM_DEVICES; i++)
    {
      if(!hostDevList[i].state.requested) continue;
      
      pack.body[idx] = hostDevList[i].devType;
      pack.body[idx+1] = hostDevList[i].devID;
      pack.body[idx+2] = hostDevList[i].readAddr;
      pack.body[idx+3] = hostDevList[i].bps;
      pack.body[idx+4] = hostDevList[i].rate;
      
      idx+=5;
    }
  
  pack.bodyLen = idx;
  
  if(!WritePacket(&pack)) return 0;
  
  if(!GetAckOrNack(&ack)) return 0;
  
  return ack;
}

int sak2v2_verify_device_configurations()
{
  sak2v2_packet pack;
  int bGotConfig = 0;
  int d, idx, deviceIdx=0, i, w=0;
  
  pack.messageID = SAK2V2_GET_CONFIG_C;
  pack.bodyLen = 1;
  pack.body[0] = 0;

  if(!WritePacket(&pack)) return 0;
  
  
  //Get the response...
  while(!bGotConfig && w++ < SAK2V2_NUM_PACKS_TO_WAIT)
    {
      if(!ReadPacket(&pack)) return 0;
      
      if(pack.messageID == SAK2V2_CONFIG_RET_C)
	{
	  bGotConfig = 1;
	}
    }
  
  idx = 1; //skip base clock for now
  
  for(i = 0; i < SAK2V2_NUM_DEVICES; i++)
    {
      sak2DevList[i].state.configured = 0;
    }
    
  //update sak2 device array
  for(deviceIdx = 0; deviceIdx < SAK2V2_NUM_DEVICES; deviceIdx++)
    {
      
      //check to see if the packet has the last byte we'll
      //need for this device
      if(idx + 4 >= pack.bodyLen) break;

      sak2DevList[deviceIdx].devType = pack.body[idx];
      sak2DevList[deviceIdx].devID = pack.body[idx+1];
      sak2DevList[deviceIdx].readAddr = pack.body[idx+2];
      sak2DevList[deviceIdx].bps = pack.body[idx+3];
      sak2DevList[deviceIdx].rate = pack.body[idx+4];
      
      if(SAK_DEBUG)
	fprintf(stderr,"GET_CONFIG: type=%d, id=%d, add=%d, bps=%d, rate=%d\n", 
		pack.body[idx],  pack.body[idx+1], pack.body[idx+2],
		pack.body[idx+3], pack.body[idx+4]);
      
      sak2DevList[deviceIdx].state.configured = 1;
      idx+=5;

    }
  
  UpdateDevMap();
  
  //Check to see that all the requested devices have been
  //configured on the sak
  
  for(d = 0; d < SAK2V2_NUM_DEVICES; d++)
    {
      if(hostDevList[d].state.requested &&
	 devMap[d] == -1)
	{
	  fprintf(stderr, "Host and sak2 have mismatched device configs!\n");
	  fprintf(stderr, "host %d is not configured.\n", d);
	  return 0;
	}
    }

  return 1;
}

int sak2v2_read_devices()
{
  bGotSample = 0;
  
  if(!bStreaming) return 0;
  
  while(!bGotSample)
    {
      if(!ReadPacket(&sampPack)) return 0;
      
      if(sampPack.messageID == SAK2V2_SENSOR_C) 
	{
	  //fprintf(stderr, "SENSOR\n");
	  ProcessSensorPacket(&sampPack);
	  bGotSample = 1;
	}
      else if(sampPack.messageID == SAK2V2_META_SENSOR_C)
	{
	  //fprintf(stderr, "META\n");
	  ProcessMetaSensorPacket(&sampPack);
	} 
    }

  return 1;
}

int sak2v2_get_samples(sak2v2_devName dev, char **buf_ptr, int *bps, int *num_samp, struct timeval *time)
{
  if(devMap[dev] == -1) return 0;
  
  if(bGotSample)
    {
      *bps = (int) sak2DevList[devMap[dev]].bps;
      
      if(sak2DevList[devMap[dev]].spp == 0)
	{
	  *num_samp = sampPack.body[bufOffset[devMap[dev]]];
	  *buf_ptr = sampPack.body + bufOffset[devMap[dev]] + 1;
	}
      else
	{
	  *num_samp = sak2DevList[devMap[dev]].spp;
	  *buf_ptr = sampPack.body + bufOffset[devMap[dev]];
	}

      *time = sampPackTime;
      
      return 1;
    }      

  return 0;
}

int sak2v2_start_streaming()
{
  sak2v2_packet pack;
  int ack;

  pack.messageID = SAK2V2_START_STREAM_C;
  pack.bodyLen = 1;
  pack.body[0] = 0; //junk

  //printf("write packet\n");
  if(!WritePacket(&pack)) return 0;
  
  ///printf("get ack or nack\n");
  if(!GetAckOrNack(&ack)) return 0;

  if(ack) bStreaming = 1;
  
  return ack;
}

int sak2v2_stop_streaming()
{
  sak2v2_packet pack;
  int ack;

  pack.messageID = SAK2V2_TERM_STREAM_C;
  pack.bodyLen = 1;
  pack.body[0] = 0; //junk
  
  if(!WritePacket(&pack)) return 0;
  
  if(!GetAckOrNack(&ack)) return 0;

  if(ack) bStreaming = 0;
  
  return ack;
}

int sak2v2_echo(int size)
{
  sak2v2_packet pack, echoPack;
  int i, bGotEcho = 0, w=0;

  pack.messageID = SAK2V2_ECHO_C;
  pack.bodyLen = size;

  for(i=0; i<size; i++)
    {
      pack.body[i] = i;
    }
  
  
  if(!WritePacket(&pack)) return 0;

  while(!bGotEcho && w++ < SAK2V2_NUM_PACKS_TO_WAIT)
    {
      if(!ReadPacket(&echoPack)) return 0;
      
      if(echoPack.messageID == SAK2V2_ECHO_C) 
	{
	  bGotEcho = 1;
	  
	  if(echoPack.bodyLen != size) return 0;

  
	  for(i=0; i<size; i++)
	    {
	      if(echoPack.body[i] != i) return 0;
	    }
	}
      else if(echoPack.messageID == SAK2V2_META_SENSOR_C)
	{
	  //fprintf(stderr, "META\n");
	  ProcessMetaSensorPacket(&echoPack);
	} 
    }
  
  return bGotEcho;
}

int sak2v2_clear_configuration()
{
  sak2v2_packet pack;
  int ack, i;

  pack.messageID = SAK2V2_CLEAR_C;
  pack.bodyLen = 1;
  pack.body[0] = 0; //junk

  //printf("write packet\n");
  if(!WritePacket(&pack)) return 0;
  
  ///printf("get ack or nack\n");
  if(!GetAckOrNack(&ack)) return 0;

  
  for(i = 0; i < SAK2V2_NUM_DEVICES; i++)
    {
      hostDevList[i].state.requested = 0;
      
      if(ack) 
	{
	  sak2DevList[i].state.configured = 0;
	}
    }

  return ack;
}


//---------------------------------------------------------
// Local Helper Functions
//---------------------------------------------------------
int SpinRead(char *buffer, int size)
{
  int readBytes = 0;
  int newBytes;
  fd_set readSet;
  struct timeval tv;
  int retval;
  int i;

  FD_ZERO(&readSet);
  FD_SET(fd, &readSet);

  while(readBytes < size)
    {
      tv.tv_sec = 5;
      tv.tv_usec = 0;

      retval = select(fd+1, &readSet, NULL, NULL, &tv);

      if(retval)
	{
	  //fprintf(stderr, "<read:");
	  newBytes = read(fd, buffer + readBytes, size - readBytes);
	  //fprintf(stderr, "byte>\n");
	}
      else
	{
	  if(SAK_DEBUG) 
	    {
	      fprintf(stderr, "READ ERROR: got %d bytes of %d\n", readBytes, size);
	      
	      for(i = 0; i < readBytes; i++)
		{
		  fprintf(stderr, "%d ", buffer[i]);
		}	    
	      fprintf(stderr, "\n");
	    }
	  //set disconnect flag
	  //sleep?
	  return 0;
	}
      
      readBytes += newBytes;
    }
  
  return 1;
}

int SpinWrite(char *buffer, int size)
{
  int writtenBytes = 0;
  int newBytes;

  while(writtenBytes < size)
    {
      newBytes = write(fd, buffer + writtenBytes, size - writtenBytes);
      
      if(newBytes < 0)
	{
	  if(SAK_DEBUG) fprintf(stderr, "WRITE ERROR\n");
	  //set disconnect flag
	  //sleep?
	  return 0;
	}
      
      writtenBytes += newBytes;
    }
  
  return 1;

}

int ReadPacket(sak2v2_packet *pack)
{
  char buff[64];
  char checksum, calc_checksum = 0;
  int bReadPacket = 0, i;
  int readState = 0;

  //fprintf(stderr, "R");

  if(!pack) return 0;
  
  while(!bReadPacket)
    {
      switch(readState) {
      case 0: //start byte state
	if(SAK_DEBUG) fprintf(stderr, "s");
	calc_checksum = 0;
	
	if(!SpinRead(buff,1)) return 0;
	if(buff[0] == SAK2V2_PACKET_START_C) readState++;
	break;
      case 1: //messageID
	if(SAK_DEBUG) printf("m");
	if(!SpinRead(&pack->messageID,1)) return 0;
	readState++;
	break;
      case 2: // body length
	if(SAK_DEBUG) printf("l");
	if(!SpinRead(&pack->bodyLen,1)) return 0;
	readState++;
	break;
      case 3: //body
	if(SAK_DEBUG) printf("b");
	if(!SpinRead(pack->body, pack->bodyLen)) return 0;
	readState++;
	break;
      case 4://checksum
	if(SAK_DEBUG) printf("c");
	if(!SpinRead(&checksum, 1)) return 0;
	readState++;
      case 5://stop byte
	if(SAK_DEBUG) printf("S");
	if(!SpinRead(buff,1)) return 0;
	if(buff[0] != SAK2V2_PACKET_STOP_C)
	  readState = 0;
	else readState++;
	break;
      case 6://check the checksum  
	if(SAK_DEBUG) printf("=");
	for(i = 0; i < pack->bodyLen; i++)
	  {
	    calc_checksum ^= pack->body[i];
	  }
	
	if(checksum != calc_checksum) readState = 0;
	else bReadPacket = 1;
	break;
      default:
	return 0;
      }
    }

  if(SAK_DEBUG) printf("packet read: mess=%c, len=%d\n", pack->messageID, pack->bodyLen);
  
  return 1;
}

int WritePacket(sak2v2_packet *pack)
{
  char start_c = SAK2V2_PACKET_START_C;
  char end_c = SAK2V2_PACKET_STOP_C;
  char checksum = 0;
  int i;

  if(!SpinWrite(&start_c, 1)) return 0;
  
  if(!SpinWrite(&pack->messageID, 1)) return 0;

  if(!SpinWrite(&pack->bodyLen, 1)) return 0;
  
  if(!SpinWrite(pack->body, pack->bodyLen)) return 0;

  //compute checksum and write
  for(i = 0; i < pack->bodyLen; i++)
    {
      checksum ^= pack->body[i];
    }

  if(!SpinWrite(&checksum, 1)) return 0;

  if(!SpinWrite(&end_c, 1)) return 0;
  
  if(SAK_DEBUG) printf("packet written: mess=%c, len=%d\n", pack->messageID, pack->bodyLen);
  return 1;
}


int GetAckOrNack(int *ack)
{
  sak2v2_packet pack;
  int bGotReturn = 0, w=0;

  if(!ack) return 0;

  //get the ack or nack
  while(!bGotReturn && w++ < SAK2V2_NUM_PACKS_TO_WAIT)
    {
      if(!ReadPacket(&pack)) return 0;
      
      if(pack.messageID == SAK2V2_ACK_C ||
	 pack.messageID == SAK2V2_NACK_C)
	{
	  bGotReturn = 1;
	}
      else if(pack.messageID == SAK2V2_META_SENSOR_C) 
	ProcessMetaSensorPacket(&pack);
    }
  
  *ack = (pack.messageID == SAK2V2_ACK_C);
  
  return 1;
}



void ProcessSensorPacket(sak2v2_packet *pack)
{
  int idx = 0, deviceIdx, varSamples;
    
  if(pack->bodyLen <= 1) return;

  //stamp arrival time
  gettimeofday(&sampPackTime, NULL);
  
  //figure out the offsets into the the sensor buffer
  //for each of the sak2 devices.  This requires looking
  //at the spp for each device to determine if the number
  //of samples varies
  for(deviceIdx = 0; deviceIdx < SAK2V2_NUM_DEVICES; deviceIdx++)
    {
      if(sak2DevList[deviceIdx].state.configured)
	{
	  bufOffset[deviceIdx] = idx;

	  if(sak2DevList[deviceIdx].spp == 0)
	    {
	      varSamples = pack->body[idx++];
	      
	      idx += (varSamples * sak2DevList[deviceIdx].bps);
	    }
	  else
	    idx += (sak2DevList[deviceIdx].spp * sak2DevList[deviceIdx].bps);
	}
    }
}



void ProcessMetaSensorPacket(sak2v2_packet *pack)
{
  int idx, i, deviceIdx;

  //reset the sak2 device array
  for(i = 0; i < SAK2V2_NUM_DEVICES; i++)
    {
      sak2DevList[i].state.configured = 0;
    }
 
  if(pack->bodyLen <= 0) return;

  //get the new clock
  sak2BaseClock = pack->body[0];
  idx = 1;
  
  //update new devices into the sak2 array
  for(deviceIdx = 0; deviceIdx < SAK2V2_NUM_DEVICES; deviceIdx++)
    {
      //check to see if the packet has the last byte we'll
      //need for this device
      if(idx + 4 >= pack->bodyLen) break;

      sak2DevList[deviceIdx].devType = pack->body[idx];
      sak2DevList[deviceIdx].devID = pack->body[idx+1];
      sak2DevList[deviceIdx].readAddr = pack->body[idx+2];
      sak2DevList[deviceIdx].bps = pack->body[idx+3];
      sak2DevList[deviceIdx].spp = pack->body[idx+4];
            
      sak2DevList[deviceIdx].state.configured = 1;

      if(SAK_DEBUG) 
	fprintf(stderr,"META: type=%d, id=%d, add=%d, bps=%d, spp=%d\n", 
		pack->body[idx], pack->body[idx+1], pack->body[idx+2],
		pack->body[idx+3], pack->body[idx+4]);
	
      idx+=5;
    }
  
  UpdateDevMap();
}

void UpdateDevMap()
{
  int h, s;

  for(h = 0; h < SAK2V2_NUM_DEVICES; h++)
    {
      devMap[h] = -1;
      
      for(s = 0; s < SAK2V2_NUM_DEVICES; s++)
	{
	  if(sak2DevList[s].state.configured == 1 &&
	     hostDevList[h].state.requested == 1 &&
	     hostDevList[h].devType == sak2DevList[s].devType &&
	     hostDevList[h].devID == sak2DevList[s].devID)
	    {
	      if(SAK_DEBUG) printf("sak2: %d is host: %d\n", s, h);

	      devMap[h] = s;
	      break;
	    }
	}
    }
}


