/*****************************************************************************
*
* Author: Jonathan Gips, MIT Wearable Computing Group
* http://www.media.mit.edu/wearables/
* Copyright (C) 2005 MIT Media Lab
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*
*****************************************************************************
* Implementation of an XML parser, based on the 'expat' library
* 
*****************************************************************************/

#include "XMLObject.hpp"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <assert.h>
#include <ctype.h>

#include "expat.h"


const int XMLObject::XML_BUFF_SIZE = 200000;

char *XMLObject::MyStrndup(const char *s, int len)
{
  char *buf = (char*) malloc(sizeof(char)*(len+1));
  memset(buf, 0, sizeof(char)*(len+1));
  memcpy(buf, s, sizeof(char)*len);
  
  return buf;
}

XMLObject::XMLObject(void) :
  parser(NULL), parentObj(NULL) {}

XMLObject::~XMLObject(void) {}

void XMLObject::pushXMLObject(XMLObject *obj, const char *name, const char **atts)
{
  if(obj != NULL) {
    obj->setParent(this);
    obj->setParser(parser);
    XML_SetUserData(parser, obj);
    obj->startXMLElement(name, atts);
  }
}

void XMLObject::popXMLObject()
{
  XMLObject *obj = getParent();
  
  if(obj != NULL)
    XML_SetUserData(obj->getParser(), obj);
  else {
    //Skip everything else 
    XML_SetUserData(getParser(), NULL);
    XML_SetElementHandler(parser, StartXMLElementNOP, EndXMLElementNOP);
    XML_SetCharacterDataHandler(parser, XMLCharacterDataHandlerNOP);
  }
}

  
int XMLObject::loadXML(char *filename)
{
  FILE *xmlFD = fopen(filename, "r");
  if (xmlFD == NULL) {
    return 1;
  }
  
  if(parser) {
    XML_ParserFree(parser);
  }

  parser = XML_ParserCreate(NULL);
  
  XML_SetUserData(parser, this);

  //This sets VIRTUAL start and end element parsers
  XML_SetElementHandler(parser, StartXMLElementWrapper, EndXMLElementWrapper);
  XML_SetCharacterDataHandler(parser, XMLCharacterDataHandlerWrapper);
 
  char buf[XML_BUFF_SIZE];
  memset(buf, 0, XML_BUFF_SIZE);
  bool done;

  do {
    size_t len = fread(buf, 1, sizeof(buf), xmlFD);
    done = feof(xmlFD);
    
    if (XML_Parse(parser, buf, len, done) == XML_STATUS_ERROR) {
      fprintf(stderr,
              "%s at line %d\n",
              XML_ErrorString(XML_GetErrorCode(parser)),
              XML_GetCurrentLineNumber(parser));
      return 1;
    }
  } while (!done);
  XML_ParserFree(parser);
  return 0;
}

void XMLObject::StartXMLElementWrapper(void *thisObj, const char *name, const char **atts)
{
  XMLObject *xObj = (XMLObject*)thisObj;
  
  xObj->startXMLElement(name, atts);
}


void XMLObject::EndXMLElementWrapper(void *thisObj, const char *name)
{
  XMLObject *xObj = (XMLObject*)thisObj;
  
  xObj->endXMLElement(name);
}

void XMLObject:: XMLCharacterDataHandlerWrapper(void *thisObj, const char *s, int len)
{
  XMLObject *xObj = (XMLObject*)thisObj;
  
  s = skipCharacterDataWS(s);
  
  if(!isascii(s[0])) return;

  xObj->processXMLCharacterData(s, len);
}

inline
const char *XMLObject::skipCharacterDataWS(const char *s)
{
  while(isspace(s[0])) s++;

  return s;
}

//---------------------------------------------------------------


XMLVector::XMLVector(unsigned int d) : 
  dim(d), v(NULL)
{ 
  XMLObject();
  if(dim > 0)
    v = NewVector(dim);
}
  
XMLVector::~XMLVector() 
{ 
  clear();
}

void XMLVector::clear() 
{
  if(v) FreeVector(v);
  v = NULL;
}

void XMLVector::startXMLElement(const char *name, const char **atts) 
{
  xml_state = VECTOR;
}

void XMLVector::endXMLElement(const char *name) 
{ 
  popXMLObject();
}

void XMLVector::processXMLCharacterData(const char *s, int len) 
{
  char *s_ptr(NULL);
  
  switch(xml_state) {
  case VECTOR:
    for(unsigned int i = 0; i < dim; i++) {
      VectorSet(v,i,strtod(s, &s_ptr));
      s = s_ptr;
    }
    break;
  default:
    break;
  }
  xml_state = SKIP;
}

//---------------------------------------------------------------


XMLMatrix::XMLMatrix(unsigned int r, unsigned int c) : 
  rowIdx(-1),row(r), column(c), m(NULL)
{ 
  XMLObject();
  if(row > 0 && column > 0)
    m = NewMatrix(row, column);
}
  
XMLMatrix::~XMLMatrix() 
{ 
  clear(); 
}

void XMLMatrix::clear() 
{
  if(m) FreeMatrix(m);
  m = NULL;
  rowIdx = 0;
}


bool XMLMatrix::GetTranspose(Matrix *m, double *trans) 
{
  if(m == NULL) return false;

  for(unsigned int i = 0; i < m->rows; i++) {
    for(unsigned int j = 0; j < m->cols; j++) {
      trans[i*m->cols + j] = m->values[j*m->rows + i];
    }
  }

  return true;
}

void XMLMatrix::startXMLElement(const char *name, const char **atts) 
{
  if(!strcasecmp(name, "matrix")) {
    xml_state = MATRIX;
    clear();
  }
  else if(!strcasecmp(name, "numrows")) {
    xml_state = NUMROWS;
  } 
  else if(!strcasecmp(name, "numcols")) {
    xml_state = NUMCOLS;
  } 
  else if(!strcasecmp(name, "row")) {
    xml_state = ROW;
    if(m == NULL && row > 0 && column > 0) {
      m = NewMatrix(row, column);
      rowIdx = 0;
    } 
    else if(m == NULL) assert(0); //premature row elem
  }
}

void XMLMatrix::endXMLElement(const char *name) 
{ 
  if(!strcasecmp(name, "matrix") ||
     (strcasecmp(name, "numrows") &&
      strcasecmp(name, "numcols") &&
      strcasecmp(name, "row")) ) {
    popXMLObject(); //assume end tag
  }
}

void XMLMatrix::processXMLCharacterData(const char *s, int len) 
{
  const char *s_ptr(NULL);
  char *s_ptr2(NULL);
  double val = -999;

  switch(xml_state) {
  case NUMROWS:
    row = atoi(s);
    rowIdx = 0;
    break;
  case NUMCOLS:
    column = atoi(s);
    break;
  case ROW:
    s_ptr = s;
    for(unsigned int i = 0; i < column; i++) {
      val = strtod(s_ptr, &s_ptr2);
      MatrixSet(m,rowIdx,i,val);
      s_ptr = s_ptr2;
    }
    rowIdx++;
    break;
  default:
    break;
  }
  xml_state = SKIP; //only one parsing necessary per state
  
}
