/*****************************************************************************
*
* Author: Wen Dong, MIT Wearable Computing Group
* http://www.media.mit.edu/wearables/
* Copyright (C) 2005 MIT Media Lab
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*****************************************************************************
* This is the implementation of the influence model. It is based on the Gnu 
* Scientific Liberary (GSL). An interface to Enchantment is implemented in
* IClassifier
*****************************************************************************/

#define HAVE_INLINE
#define GSL_RANGE_CHECK_OFF 1
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <gsl/gsl_nan.h>
#include <gsl/gsl_math.h>
#include <gsl/gsl_sf_exp.h>
#include <gsl/gsl_vector.h>
#include <gsl/gsl_matrix.h>
#include <gsl/gsl_blas.h>
#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>
#include <gsl/gsl_linalg.h>
#include "hd_influence.h"
//#include <dmalloc.h>


struct _hd_influence{
  /* number of latent states for each chain */
  gsl_vector * n_states; 
  gsl_vector *s_off;

  /* discrete output -> number of observation states */
  /* continuous output -> Gaussian dimensionality */
  gsl_vector * n_outputs;  
  gsl_vector * o_type;


  /* influence parameters */
  gsl_matrix * infl;
  gsl_matrix * A;
  gsl_matrix ** B;
  gsl_matrix * D;
  gsl_vector * prior;
  gsl_matrix *** cov;

  gsl_matrix * N;

  /* sufficient statistics accumulator*/
  gsl_matrix * xi;
  gsl_vector * pi;
  gsl_vector * sw;
  gsl_matrix ** B0;
  gsl_matrix *** cov0;
} ;

struct _hd_sample{
  gsl_vector * n_s;
  gsl_vector * s_off;
  gsl_vector * n_o;
  gsl_vector * obs_off;
  gsl_vector * o_t;
  gsl_matrix * s;
  gsl_matrix * o;

  /* sufficient statistics */
  gsl_matrix * alpha;
  gsl_matrix * beta;
  gsl_matrix * scale;
  gsl_matrix * gamma;
  gsl_matrix * obslik;

} ;

int hd_matrix_normalize(gsl_matrix *in){
  int i;
  for(i=0;i<in->size1;i++){
    gsl_vector_view tmp = gsl_matrix_row(in,i);
    double sum = gsl_blas_dasum(&tmp.vector);
    gsl_vector_scale(&tmp.vector,1.0f/(sum+1e-10));
  }
  return 0;
}

hd_influence * hd_influence_init(int n_chains,
                                 double * n_states, 
                                 double * n_outputs,
                                 double * o_type){
  int i,k,n_total_states;
  gsl_vector_view tmp1,tmp2,tmp3;

  hd_influence* infl = (hd_influence*)malloc(sizeof(hd_influence));
  memset(infl,0,sizeof(hd_influence));

  tmp1 = gsl_vector_view_array(n_states,n_chains);
  tmp2 = gsl_vector_view_array(n_outputs,n_chains);
  tmp3 = gsl_vector_view_array(o_type,n_chains);
  infl->n_states = gsl_vector_alloc(n_chains);
  gsl_vector_memcpy(infl->n_states, &tmp1.vector);
  infl->s_off = gsl_vector_alloc(n_chains);
  gsl_vector_set(infl->s_off,0,0);
  for(i=1;i<n_chains;i++)
    gsl_vector_set(infl->s_off,i,gsl_vector_get(infl->s_off,i-1)+
                   n_states[i-1]);
  infl->n_outputs = gsl_vector_alloc(n_chains);
  gsl_vector_memcpy(infl->n_outputs, &tmp2.vector);
  infl->o_type = gsl_vector_alloc(n_chains);
  gsl_vector_memcpy(infl->o_type, &tmp3.vector);

  for(i=0,n_total_states=0; i<n_chains; i++) n_total_states += n_states[i];
  infl->A = gsl_matrix_alloc(n_total_states,n_total_states);
  infl->D = gsl_matrix_alloc(n_chains,n_chains);
  infl->prior = gsl_vector_alloc(n_total_states);
  infl->pi = gsl_vector_alloc(n_total_states);
  infl->sw = gsl_vector_alloc(n_total_states);
  infl->infl = gsl_matrix_alloc(n_total_states,n_total_states);
  infl->xi = gsl_matrix_alloc(n_total_states,n_total_states);

  infl->N = gsl_matrix_alloc(n_total_states,n_total_states);
  gsl_matrix_set_all(infl->N,0);
  for(i=0; i<n_chains; i++){
    int ns_i = gsl_vector_get(infl->n_states,i);
    int off_i = gsl_vector_get(infl->s_off,i);
    gsl_matrix_view tmp=gsl_matrix_submatrix(infl->N,off_i,off_i,ns_i,ns_i);
    gsl_matrix_set_all(&tmp.matrix,1.0f);
  }

  infl->B = (gsl_matrix**)malloc(n_chains*sizeof(gsl_matrix*));
  infl->B0 = (gsl_matrix**)malloc(n_chains*sizeof(gsl_matrix*));
  for(i=0;i<n_chains;i++){
    infl->B[i] = gsl_matrix_alloc(n_states[i],n_outputs[i]);
    infl->B0[i] = gsl_matrix_alloc(n_states[i],n_outputs[i]);
  }

  infl->cov = (gsl_matrix***)malloc(n_chains*sizeof(gsl_matrix***));
  infl->cov0 = (gsl_matrix***)malloc(n_chains*sizeof(gsl_matrix***));
  for(i=0;i<n_chains;i++) if(o_type[i]=='c'){
    infl->cov[i] = (gsl_matrix**)malloc(n_states[i]*sizeof(gsl_matrix**));
    infl->cov0[i] = (gsl_matrix**)malloc(n_states[i]*sizeof(gsl_matrix**));
    for(k=0;k<n_states[i];k++){
      infl->cov[i][k] = gsl_matrix_alloc(n_outputs[i],n_outputs[i]);
      infl->cov0[i][k] = gsl_matrix_alloc(n_outputs[i],n_outputs[i]);
    }
  }

  hd_influence_reset_ess(infl);

  return infl;
}

int hd_influence_get_nchains(hd_influence *infl){
  return infl->n_states->size;
}

int hd_influence_get_num_states(hd_influence *infl, double * ns){
  gsl_vector_view n=gsl_vector_view_array(ns,infl->n_states->size);
  gsl_vector_memcpy(&n.vector, infl->n_states);
  return 0;
}

int hd_influence_get_num_outputs(hd_influence *infl,double *n_o){
  gsl_vector_view n=gsl_vector_view_array(n_o,infl->n_outputs->size);
  gsl_vector_memcpy(&n.vector, infl->n_outputs);
  return 0;
}

int hd_influence_get_o_types(hd_influence *infl,double *o_types){
  gsl_vector_view o=gsl_vector_view_array(o_types,infl->o_type->size);
  gsl_vector_memcpy(&o.vector, infl->o_type);
  return 0;
}

int hd_influence_set_prior(hd_influence * infl, double *prior){
  gsl_vector_view p0 = gsl_vector_view_array(prior,infl->infl->size1);
  gsl_vector_memcpy(infl->prior, &p0.vector);
  return 0;
}

int hd_influence_get_prior(hd_influence * infl, double *prior){
  gsl_vector_view p0 = gsl_vector_view_array(prior,infl->infl->size1);
  gsl_vector_memcpy(&p0.vector, infl->prior);
  return 0;
}

int hd_influence_set_D(hd_influence * infl, double * D){
  int nc = infl->n_states->size;
  gsl_matrix_view D0 = gsl_matrix_view_array(D, nc, nc);
  gsl_matrix_memcpy(infl->D, &D0.matrix); 
  return 0;
}

int hd_influence_get_D(hd_influence * infl, double * D){
  int nc = infl->n_states->size;
  gsl_matrix_view D0 = gsl_matrix_view_array(D, nc, nc);
  gsl_matrix_memcpy(&D0.matrix,infl->D); 
  return 0;
}

int hd_influence_set_A(hd_influence * infl, int c1, int c2, double *A){
  int ns1=gsl_vector_get(infl->n_states,c1);
  int ns2=gsl_vector_get(infl->n_states,c2);
  gsl_matrix_view A0 = gsl_matrix_view_array(A,ns1,ns2);
  int off1 = gsl_vector_get(infl->s_off, c1);
  int off2 = gsl_vector_get(infl->s_off, c2);
  gsl_matrix_view A2 = gsl_matrix_submatrix(infl->A,off1,off2,ns1,ns2);
  gsl_matrix_memcpy(&A2.matrix,&A0.matrix);
  return 0;
}

int hd_influence_get_A(hd_influence * infl, int c1, int c2, double *A){
  int ns1=gsl_vector_get(infl->n_states,c1);
  int ns2=gsl_vector_get(infl->n_states,c2);
  gsl_matrix_view A0 = gsl_matrix_view_array(A,ns1,ns2);
  int off1 = gsl_vector_get(infl->s_off, c1);
  int off2 = gsl_vector_get(infl->s_off, c2);
  gsl_matrix_view A2 = gsl_matrix_submatrix(infl->A,off1,off2,ns1,ns2);
  gsl_matrix_memcpy(&A0.matrix,&A2.matrix);
  return 0;
}

int hd_influence_set_B(hd_influence * infl, int c, double * B){
  int ns = gsl_vector_get(infl->n_states,c);
  int no = gsl_vector_get(infl->n_outputs,c);
  gsl_matrix_view B0 = gsl_matrix_view_array(B,ns,no);
  gsl_matrix_memcpy(infl->B[c],&B0.matrix);
  return 0;
}

int hd_influence_get_B(hd_influence * infl, int c, double * B){
  int ns = gsl_vector_get(infl->n_states,c);
  int no = gsl_vector_get(infl->n_outputs,c);
  gsl_matrix_view B0 = gsl_matrix_view_array(B,ns,no);
  gsl_matrix_memcpy(&B0.matrix,infl->B[c]);
  return 0;
}

int hd_influence_set_cov(hd_influence * infl, int c, int s, double * cov){
  int n_o = gsl_vector_get(infl->n_outputs,c);
  gsl_matrix_view cov0 = gsl_matrix_view_array(cov,n_o,n_o);
  gsl_matrix_memcpy(infl->cov[c][s],&cov0.matrix);
  return 0;
}

int hd_influence_get_cov(hd_influence * infl, int c, int s, double * cov){
  int n_o = gsl_vector_get(infl->n_outputs,c);
  gsl_matrix_view cov0 = gsl_matrix_view_array(cov,n_o,n_o);
  gsl_matrix_memcpy(&cov0.matrix,infl->cov[c][s]);
  return 0;
}

int hd_influence_set_H(hd_influence * infl){
  int nc=infl->D->size1,i,j;
  gsl_matrix_memcpy(infl->infl,infl->A);
  for(i=0; i<nc; i++){
    int offi = gsl_vector_get(infl->s_off,i);
    int nsi = gsl_vector_get(infl->n_states,i);

    for(j=0; j<nc; j++){
      int offj = gsl_vector_get(infl->s_off,j);
      int nsj = gsl_vector_get(infl->n_states,j);
      gsl_matrix_view tmp=gsl_matrix_submatrix(infl->infl,offi,offj,nsi,nsj);
      gsl_matrix_scale(&tmp.matrix,gsl_matrix_get(infl->D,i,j));

    }
  }

  return 0;
}

int hd_influence_get_H(hd_influence * infl, double * H){
  int sns = infl->infl->size1;
  gsl_matrix_view H0 = gsl_matrix_view_array(H, sns, sns);
  gsl_matrix_memcpy(&H0.matrix, infl->infl);
  return 0;
}

int hd_influence_reset_ess(hd_influence * infl){

  int i;
  for(i=0;i<infl->n_states->size; i++){
    gsl_matrix_set_all(infl->B0[i],0);
  }
  for(i=0;i<infl->n_states->size;i++) 
    if(gsl_vector_get(infl->o_type,i)=='c'){
      int k;
      for(k=0;k<gsl_vector_get(infl->n_states,i);k++){
        gsl_matrix_set_all(infl->cov0[i][k],0);
      }
    }

  gsl_vector_set_all(infl->pi,0);
  gsl_matrix_set_all(infl->xi,0);
  return 0;
}

void hd_influence_free(hd_influence * infl){
  int i;
  
  for(i=0; i<infl->n_states->size; i++){
    gsl_matrix_free(infl->B[i]);
    gsl_matrix_free(infl->B0[i]);
  }
  free(infl->B);
  free(infl->B0);

  for(i=0;i<infl->n_states->size;i++) 
    if(gsl_vector_get(infl->o_type,i)=='c'){
      int k;
      for(k=0;k<gsl_vector_get(infl->n_states,i);k++){
        gsl_matrix_free(infl->cov[i][k]);
        gsl_matrix_free(infl->cov0[i][k]);
      }
      free(infl->cov[i]);
      free(infl->cov0[i]);
    }
  free(infl->cov);
  free(infl->cov0);

  gsl_matrix_free(infl->infl);
  gsl_matrix_free(infl->xi);
  gsl_matrix_free(infl->N);
  gsl_matrix_free(infl->A);
  gsl_matrix_free(infl->D);

  gsl_vector_free(infl->prior);
  gsl_vector_free(infl->pi);
  gsl_vector_free(infl->sw);

  gsl_vector_free(infl->n_states);
  gsl_vector_free(infl->s_off);
  gsl_vector_free(infl->n_outputs);
  gsl_vector_free(infl->o_type);

  free(infl);
}

int hd_influence_print(FILE * fd, hd_influence * infl){
  int i,j,k;

  fprintf(fd,"this influence model has %d chains",infl->D->size1);
  fprintf(fd,"\n");

  fprintf(fd,"number of states for each chain ");
  gsl_vector_fprintf(stdout, infl->n_states, "%2.0f ");
  fprintf(fd,"\n");

  fprintf(fd,"number of outputs for each chain ");
  gsl_vector_fprintf(fd, infl->n_outputs, "%2.0f ");
  fprintf(fd,"\n");

  /*
  fprintf(fd,"the D matrix is\n");
  for(i=0;i<infl->D->size1;i++){
    gsl_vector_view row = gsl_matrix_row(infl->D,i);
    for(k=0;k<(&row.vector)->size;k++)
      fprintf(fd,"%8.4f",gsl_vector_get(&row.vector,k));
    fprintf(fd,"\n");
  }
  fprintf(fd,"\n");
  */

  fprintf(fd,"the prior is\n");
  for(i=0;i<infl->prior->size;i++){
    fprintf(fd,"%8.4f",gsl_vector_get(infl->prior,i));
  }
  fprintf(fd,"\n");

  fprintf(fd,"the influence matrix is\n");
  for(i=0;i<infl->infl->size1;i++){
    gsl_vector_view row = gsl_matrix_row(infl->infl,i);
    for(k=0;k<(&row.vector)->size;k++)
      fprintf(fd,"%8.4f",gsl_vector_get(&row.vector,k));
    fprintf(fd,"\n");
  }
  fprintf(fd,"\n");

  fprintf(fd,"the output matrices are\n");
  for(i=0;i<infl->D->size1;i++){
    fprintf(fd,"B[%d]\n",i);
    for(j=0;j<infl->B[i]->size1;j++){
      gsl_vector_view row = gsl_matrix_row(infl->B[i],j);
      int k;
      for(k=0;k<(&row.vector)->size;k++)
        fprintf(fd,"%8.4f",gsl_vector_get(&row.vector,k));
      fprintf(fd,"\n");
    }
    fprintf(fd,"\n");
  }

  fprintf(fd,"covariance matrices are\n");
  for(i=0;i<infl->D->size1;i++){
    if(gsl_vector_get(infl->o_type,i)=='d') continue;
    fprintf(fd,"cov[%d]\n",i);
    for(j=0;j<gsl_vector_get(infl->n_outputs,i);j++){
      int m;
      for(m=0;m<gsl_vector_get(infl->n_states,i);m++){
        gsl_vector_view row = gsl_matrix_row(infl->cov[i][m],j);
        int k;
        for(k=0;k<(&row.vector)->size;k++)
          fprintf(fd,"%8.4f",gsl_vector_get(&row.vector,k));
        fprintf(fd,"\t");
      }
      fprintf(fd,"\n");
    }
    fprintf(fd,"\n");
  }

  /*
  fprintf(fd,"the xi matrix is\n");
  for(i=0;i<infl->xi->size1;i++){
    gsl_vector_view row = gsl_matrix_row(infl->xi,i);
    for(k=0;k<(&row.vector)->size;k++)
      fprintf(fd,"%8.4f",gsl_vector_get(&row.vector,k));
    fprintf(fd,"\n");
  }
  fprintf(fd,"\n");
  */
  fprintf(fd,"\n");
  return 0;
}

int hd_sample_gaussian(int n_samples,
                       gsl_vector * mu, 
                       gsl_matrix * cov, 
                       gsl_matrix * sampl){
  int M = cov->size1;
  int N = cov->size2;
  int i,j;
  gsl_rng * r = gsl_rng_alloc(gsl_rng_default);
  gsl_matrix * s2 = gsl_matrix_alloc(M,n_samples);

  /* assert(M==N); */
  gsl_matrix * U = gsl_matrix_alloc(M,N);
  gsl_matrix * V = gsl_matrix_alloc(N,N);
  gsl_vector_view tmp1 = gsl_matrix_diagonal(V);
  gsl_vector * S = gsl_vector_alloc(N);
  gsl_vector * work = gsl_vector_alloc(N);
  gsl_matrix_memcpy(U,cov);
  gsl_linalg_SV_decomp(U,V,S,work);

  gsl_matrix_set_all(V,0); // reuse V
  gsl_vector_memcpy(&tmp1.vector,S);
  for(i=0;i<M;i++) gsl_vector_set(&tmp1.vector,i,sqrt(gsl_vector_get(S,i)));

  //gsl_rng_env_setup();
  gsl_rng_set(r,clock()/CLOCKS_PER_SEC);
  for(i=0;i<M;i++) for(j=0;j<n_samples;j++){
    double u = gsl_ran_gaussian(r,1);
    gsl_matrix_set(sampl,i,j, u*2-1);
  }
  gsl_rng_free(r);

  gsl_blas_dgemm(CblasNoTrans,CblasNoTrans,1.0f,V,sampl,0.0f,s2);

  for(j=0;j<n_samples;j++){
    tmp1 = gsl_matrix_column(sampl,j);
    gsl_vector_memcpy(&tmp1.vector,mu);
  }
  gsl_blas_dgemm(CblasNoTrans,CblasNoTrans,1.0f,U,s2,1.0f,sampl);
  
  gsl_matrix_free(U);
  gsl_matrix_free(V);
  gsl_vector_free(S);
  gsl_vector_free(work);
  gsl_matrix_free(s2);
  return 0;
}

int hd_influence_sample(hd_influence *infl, hd_sample *sampl){
  int i=0, j, off_s = 0;
  const int nc=infl->D->size1, T = sampl->s->size2;
  gsl_rng * r;

  /* latent states */
  gsl_vector * tmp1 = gsl_vector_alloc(infl->prior->size);
  gsl_vector * tmp2 = gsl_vector_alloc(infl->prior->size);
  gsl_vector_set_all(tmp1,0.0f);

  gsl_rng_env_setup();
  r = gsl_rng_alloc(gsl_rng_default);
  gsl_rng_set(r,clock()/CLOCKS_PER_SEC);
  for(j=0,off_s=0;j<nc;j++){
    int ns = gsl_vector_get(infl->n_states,j);
    gsl_vector_view cdt = gsl_vector_subvector(infl->prior,off_s,ns);
    gsl_ran_discrete_t * g = gsl_ran_discrete_preproc(ns,cdt.vector.data);
    int n = gsl_ran_discrete(r,g); // result
    gsl_ran_discrete_free(g);
    gsl_matrix_set(sampl->s,j,i,n); 
    gsl_vector_set(tmp1,off_s+n,1);
    off_s += ns;
  }
  for(i=1; i<T; i++){
    gsl_blas_dgemv(CblasTrans, 1, infl->infl,tmp1, 0,tmp2);
    off_s = 0;
    gsl_vector_set_all(tmp1,0.0f);
    for(j=0;j<nc;j++){
      int ns = gsl_vector_get(infl->n_states,j);
      gsl_vector_view cdt = gsl_vector_subvector(tmp2,off_s,ns);
      gsl_ran_discrete_t * g = gsl_ran_discrete_preproc(ns,cdt.vector.data);
      int n = gsl_ran_discrete(r,g);
      gsl_ran_discrete_free(g);
      gsl_matrix_set(sampl->s,j,i,n); // result
      gsl_vector_set(tmp1,off_s+n,1);
      off_s += ns;
    }
  }
  /* output */
  for(j=0;j<nc;j++){
    unsigned char o_type = gsl_vector_get(infl->o_type,j);
    int off_o = gsl_vector_get(sampl->obs_off,j);
    int n_o = gsl_vector_get(sampl->n_o,j);
    int ns = gsl_vector_get(infl->n_states,j);
    if(o_type=='c'){
      for(i=0;i<ns;i++){
        gsl_vector_view mu = gsl_matrix_row(infl->B[j],i);
        gsl_matrix * cov = infl->cov[j][i];
        gsl_matrix * tmp;
        int n_sampl = 0,k,m;
        for(k=0;k<T;k++) if(gsl_matrix_get(sampl->s,j,k)==i) n_sampl++;
        if(n_sampl==0) continue;
        tmp = gsl_matrix_alloc(n_o,n_sampl);
        hd_sample_gaussian(n_sampl,&mu.vector,cov,tmp);
        for(k=0,m=0;k<T;k++) if(gsl_matrix_get(sampl->s,j,k)==i) {
          gsl_vector_view tmp2 = gsl_matrix_column(tmp,m++);
          gsl_vector_view tmp3 = gsl_matrix_column(sampl->o,k);
          gsl_vector_view tmp4 = gsl_vector_subvector(&tmp3.vector,off_o,n_o);
          gsl_vector_memcpy(&tmp4.vector,&tmp2.vector);
        }
        gsl_matrix_free(tmp);
      }
    }else if(o_type=='d'){
      for(i=0; i<T; i++){
        int ns = gsl_vector_get(infl->n_states,j);
        int off_o = gsl_vector_get(sampl->obs_off,j);
        gsl_vector_view cdt =
          gsl_matrix_row(infl->B[j],gsl_matrix_get(sampl->s,j,i));
        gsl_ran_discrete_t * g=gsl_ran_discrete_preproc(ns,cdt.vector.data);
        int n = gsl_ran_discrete(r,g); // result
        gsl_matrix_set(sampl->o,off_o,i,n);
        gsl_ran_discrete_free(g);
      }
    }
  }

  gsl_vector_free(tmp1);
  gsl_vector_free(tmp2);
  gsl_rng_free(r);

  return 000;
}

void hd_sample_free(hd_sample *sampl){
  gsl_vector_free(sampl->n_o);
  gsl_vector_free(sampl->obs_off);
  gsl_vector_free(sampl->n_s);
  gsl_vector_free(sampl->s_off);
  gsl_vector_free(sampl->o_t);
  gsl_matrix_free(sampl->s);
  gsl_matrix_free(sampl->o);

  gsl_matrix_free(sampl->obslik);
  gsl_matrix_free(sampl->alpha);
  gsl_matrix_free(sampl->beta);
  gsl_matrix_free(sampl->gamma);
  gsl_matrix_free(sampl->scale);

  free(sampl);
}

int hd_sample_get_nchains(hd_sample *sampl){
  return sampl->n_s->size;
}

int hd_sample_get_num_states(hd_sample *sampl, double *n_states){
  gsl_vector_view n_s = gsl_vector_view_array(n_states,sampl->n_s->size);
  gsl_vector_memcpy(&n_s.vector,sampl->n_s);
  return 000;
}

int hd_sample_get_num_outputs(hd_sample *sampl,double *n_outputs){
  gsl_vector_view n_o = gsl_vector_view_array(n_outputs,sampl->n_o->size);
  gsl_vector_memcpy(&n_o.vector,sampl->n_o);
  return 000;
}

int hd_sample_get_o_types(hd_sample *sampl, double * o_types){
  gsl_vector_view o_t = gsl_vector_view_array(o_types,sampl->o_t->size);
  gsl_vector_memcpy(&o_t.vector,sampl->o_t);
  return 000;
}

int hd_sample_get_s(hd_sample *sampl, int chain, int t,double *s){
  *s = gsl_matrix_get(sampl->s,chain,t);
  return 0;
}

int hd_sample_set_s(hd_sample *sampl, int chain, int t,double s){
  gsl_matrix_set(sampl->s,chain,t,s);
  return 0;
}

int hd_sample_get_o(hd_sample *sampl, int chain, int t, double *o){
  int obs_off = gsl_vector_get(sampl->obs_off,chain);
  int n_o = gsl_vector_get(sampl->n_o, chain);
  gsl_vector_view src0 = gsl_matrix_column(sampl->o,t);
  gsl_vector_view src = gsl_vector_subvector(&src0.vector,obs_off,n_o);
  gsl_vector_view dst = gsl_vector_view_array(o,n_o);
  gsl_vector_memcpy(&dst.vector,&src.vector);
  return 0;
}

int hd_sample_set_o(hd_sample *sampl, int chain, int t, double *o){
  int obs_off = gsl_vector_get(sampl->obs_off,chain);
  int n_o = gsl_vector_get(sampl->n_o, chain);
  gsl_vector_view dst0 = gsl_matrix_column(sampl->o,t);
  gsl_vector_view dst = gsl_vector_subvector(&dst0.vector,obs_off,n_o);
  gsl_vector_view src = gsl_vector_view_array(o,n_o);
  gsl_vector_memcpy(&dst.vector,&src.vector);
  return 0;
}

int hd_sample_get_gamma(hd_sample * sampl, int chain, int t, double * g){
  int s_off = gsl_vector_get(sampl->s_off,chain);
  int n_s = gsl_vector_get(sampl->n_s, chain);
  gsl_vector_view src0 = gsl_matrix_row(sampl->gamma,t);
  gsl_vector_view src = gsl_vector_subvector(&src0.vector,s_off,n_s);
  gsl_vector_view dst = gsl_vector_view_array(g,n_s);
  gsl_vector_memcpy(&dst.vector,&src.vector);
  return 0;
}

hd_sample * hd_sample_alloc(hd_influence *infl, int T){
  int nc = infl->D->size1, sns = infl->infl->size1;
  hd_sample * sampl = (hd_sample*)malloc(sizeof(hd_sample));
  int total_o = 0,i;
  sampl->n_o = gsl_vector_alloc(nc);
  gsl_vector_memcpy(sampl->n_o,infl->n_outputs);

  sampl->n_s = gsl_vector_alloc(nc);
  gsl_vector_memcpy(sampl->n_s,infl->n_states);
  sampl->s_off = gsl_vector_alloc(nc);
  gsl_vector_memcpy(sampl->s_off,infl->s_off);

  sampl->o_t = gsl_vector_alloc(nc);
  gsl_vector_memcpy(sampl->o_t,infl->o_type);

  sampl->s = gsl_matrix_alloc(nc,T);
  gsl_matrix_set_all(sampl->s,GSL_NAN);
  for(i=0;i<nc;i++)
    if(gsl_vector_get(infl->o_type,i)=='c')
      total_o += gsl_vector_get(infl->n_outputs,i);
    else
      total_o ++;

  for(i=0;i<nc;i++){
    if(gsl_vector_get(infl->o_type,i)=='c')
      gsl_vector_set(sampl->n_o,i,gsl_vector_get(infl->n_outputs,i));
    else
      gsl_vector_set(sampl->n_o,i,1);
  }
  sampl->obs_off = gsl_vector_alloc(total_o);
  gsl_vector_set(sampl->obs_off,0,0);
  for(i=1;i<nc;i++)
    if(gsl_vector_get(infl->o_type,i)=='c'){
      int tmp1 = gsl_vector_get(sampl->obs_off,i-1);
      int tmp2 = gsl_vector_get(sampl->n_o,i-1);
      gsl_vector_set(sampl->obs_off,i,tmp1+tmp2);
    }else{
      gsl_vector_set(sampl->obs_off,i,gsl_vector_get(sampl->obs_off,i-1)+1);
    }

  sampl->o = gsl_matrix_alloc(total_o,T);
  gsl_matrix_set_all(sampl->o, GSL_NAN);

  sampl->obslik = gsl_matrix_alloc(sns,T);
  sampl->alpha = gsl_matrix_alloc(T,sns);
  sampl->beta = gsl_matrix_alloc(sns,T);
  sampl->scale = gsl_matrix_alloc(T,sns);
  sampl->gamma = gsl_matrix_alloc(T,sns);

  return sampl;
}

int hd_sample_print(FILE * fd, hd_sample *sampl){

  int T = sampl->s->size2;
  int nc = sampl->s->size1;
  int i,j;

  for(i=0;i<T;i++){
    for(j=0;j<nc;j++){
      double tmp = gsl_matrix_get(sampl->s,j,i);
      fprintf(fd,"%10.2f",tmp);
    }
    for(j=0;j<sampl->o->size1;j++){
      double tmp = gsl_matrix_get(sampl->o,j,i);
      fprintf(fd, "%10.6f",tmp);
    }
    fprintf(fd,"\n");
  }

  return 0;
}

int hd_sample_print_statistics(FILE * fd, hd_sample *sampl){
  int i,j;

  fprintf(fd,"#### obslik ####\n");
  for(i=0;i<sampl->obslik->size2;i++){
    for(j=0;j<sampl->obslik->size1;j++)
      fprintf(fd,"%8.6f ",gsl_matrix_get(sampl->obslik,j,i));
    fprintf(fd,"\n");
  }
  fprintf(fd,"\n");

  fprintf(fd,"#### alpha ####\n");
  for(i=0;i<sampl->alpha->size1;i++){
    for(j=0;j<sampl->alpha->size2;j++)
      fprintf(fd,"%8.6f ",gsl_matrix_get(sampl->alpha,i,j));
    fprintf(fd,"\n");
  }
  fprintf(fd,"\n");

  fprintf(fd,"#### scale ####\n");
  for(i=0;i<sampl->scale->size1;i++){
    for(j=0;j<sampl->scale->size2;j++)
      fprintf(fd,"%8.6f ",gsl_matrix_get(sampl->scale,i,j));
    fprintf(fd,"\n");
  }
  fprintf(fd,"\n");

  fprintf(fd,"#### beta ####\n");
  for(i=0;i<sampl->beta->size2;i++){
    for(j=0;j<sampl->beta->size1;j++)
      fprintf(fd,"%8.6f ",gsl_matrix_get(sampl->beta,j,i));
    fprintf(fd,"\n");
  }
  fprintf(fd,"\n");

  fprintf(fd,"#### gamma ####\n");
  for(i=0;i<sampl->gamma->size1;i++){
    for(j=0;j<sampl->gamma->size2;j++)
      fprintf(fd,"%8.6f ",gsl_matrix_get(sampl->gamma,i,j));
    fprintf(fd,"\n");
  }
  fprintf(fd,"\n");

  return 0;
}

int hd_influence_maximize_params(hd_influence * infl){

  int i,j;
  int n_chains = infl->n_states->size;
  gsl_matrix * coeff2=gsl_matrix_alloc(n_chains,infl->infl->size1);
  gsl_matrix * D0 = gsl_matrix_alloc(n_chains,n_chains);
  gsl_matrix * left;

  /* T */
  gsl_matrix_set_all(coeff2,0);
  for(i=0;i<n_chains;i++){
    int ns = gsl_vector_get(infl->n_states,i);
    int s_off = gsl_vector_get(infl->s_off,i);
    gsl_matrix_view tmp = gsl_matrix_submatrix(coeff2,i,s_off,1,ns);
    gsl_matrix_set_all(&tmp.matrix,1);
  }
  left = gsl_matrix_alloc(n_chains,infl->infl->size1);
  gsl_blas_dgemm(CblasNoTrans,CblasNoTrans,1.0f,coeff2,infl->xi,0.0f,left);
  gsl_blas_dgemm(CblasNoTrans,CblasTrans,1.0f,left,coeff2,0.0f,infl->D);
  /* hack ! slow down increment of D */
  for(i=0; i<n_chains; i++) for(j=0; j<n_chains; j++){
    double tmp = gsl_matrix_get(infl->D,i,j);
    tmp = gsl_log1p((2*M_E-1)*tmp);
    gsl_matrix_set(infl->D,i,j,tmp);
  }
  gsl_matrix_transpose_memcpy(D0,infl->D);
  hd_matrix_normalize(D0);
  gsl_matrix_transpose_memcpy(infl->D,D0);
  gsl_matrix_free(coeff2);
  gsl_matrix_free(D0);
  gsl_matrix_free(left);

  /* A */
  gsl_matrix_memcpy(infl->A,infl->xi);
  for(i=0; i<n_chains; i++){
    int nsi = gsl_vector_get(infl->n_states,i);
    int off_i = gsl_vector_get(infl->s_off,i);
    for(j=0; j<n_chains; j++){
      int nsj = gsl_vector_get(infl->n_states,j);
      int off_j = gsl_vector_get(infl->s_off,j);
      gsl_matrix_view tmp = gsl_matrix_submatrix(infl->A,off_i,off_j,nsi,nsj);
      hd_matrix_normalize(&tmp.matrix);
    }
  }

  /* H */
  hd_influence_set_H(infl);

  /* B */
  for(i=0;i<n_chains;i++){
    int o_type = gsl_vector_get(infl->o_type,i);
    int s_off = gsl_vector_get(infl->s_off,i);
    int ns = gsl_vector_get(infl->n_states,i);
    gsl_matrix_set_all(infl->B[i],0);
    if(o_type=='c'){
      gsl_matrix_memcpy(infl->B[i],infl->B0[i]);
      for(j=0;j<ns;j++) gsl_matrix_memcpy(infl->cov[i][j],infl->cov0[i][j]);

      /* normalize */
      for(j=0;j<ns;j++){
        double denom = gsl_vector_get(infl->sw,s_off+j)+.000001;
        const double prior = .01f;
        gsl_vector_view diag = gsl_matrix_diagonal(infl->cov[i][j]);
        gsl_vector_view b = gsl_matrix_row(infl->B[i],j);
        gsl_vector_scale(&b.vector,1/denom);
        gsl_matrix_scale(infl->cov[i][j],1/denom);
        gsl_blas_dger(-1,&b.vector,&b.vector,infl->cov[i][j]);

        gsl_vector_add_constant(&diag.vector,prior);
      }
    }else{
      gsl_matrix_memcpy(infl->B[i],infl->B0[i]);
      hd_matrix_normalize(infl->B[i]);
    }
  }

  /* prior */
  gsl_blas_dgemv(CblasTrans, 1, infl->N,infl->pi,0.0f,infl->prior);
  gsl_vector_div(infl->pi,infl->prior);
  gsl_vector_memcpy(infl->prior, infl->pi);

  return 000;
}

inline int hd_matrix_pinv(gsl_matrix * A, gsl_matrix * inv_A){
  int i, M = A->size1, N = A->size2;
  gsl_matrix *U = gsl_matrix_alloc(M,N);
  gsl_matrix *V = gsl_matrix_alloc(N,N);
  gsl_vector *S = gsl_vector_alloc(N);
  gsl_vector *work = gsl_vector_alloc(N);
  gsl_matrix *S0 = gsl_matrix_alloc(N,N);
  gsl_vector_view diag_S = gsl_matrix_diagonal(S0);
  gsl_matrix * left = gsl_matrix_alloc(N,N);
  gsl_matrix_memcpy(U,A);
  gsl_linalg_SV_decomp(U,V,S,work);
  gsl_matrix_set_all(S0,0);
  for(i=0;i<N;i++) gsl_vector_set(&diag_S.vector,i,1/gsl_vector_get(S,i)); 

  gsl_blas_dgemm(CblasNoTrans,CblasNoTrans,1.0f,V,S0,0.0f,left);
  gsl_blas_dgemm(CblasNoTrans,CblasTrans,1.0f,left,U,0.0f,inv_A);
  
  gsl_matrix_free(U);
  gsl_matrix_free(V);
  gsl_matrix_free(S0);
  gsl_matrix_free(left);
  gsl_vector_free(work);
  gsl_vector_free(S);
  return 0;
}

int hd_gaussian_prob(gsl_vector * mu, 
                     gsl_matrix * cov, 
                     gsl_matrix * sampl,
                     gsl_vector * prob){
  int n_samples = sampl->size2;
  int nd = sampl->size1;
  double denom;
  int i;
  gsl_matrix *icov = gsl_matrix_alloc(cov->size2,cov->size1);
  /* assert(cov->size1==cov->size2) */

  gsl_matrix * LU = gsl_matrix_alloc(cov->size1,cov->size2);
  gsl_permutation * p = gsl_permutation_alloc(cov->size1);
  int signum;
  gsl_matrix *sampl0 = gsl_matrix_alloc(sampl->size1,sampl->size2);
  gsl_vector * right = gsl_vector_alloc(nd);
  gsl_matrix_memcpy(LU,cov);
  gsl_linalg_LU_decomp(LU,p,&signum);
  denom = 1/(.000001+sqrt(gsl_pow_int(2*M_PI,nd)*fabs(gsl_linalg_LU_det(LU,signum))));

  hd_matrix_pinv(cov,icov);

  gsl_matrix_memcpy(sampl0,sampl);
  for(i=0;i<n_samples;i++){
    double dot=0;
    gsl_vector_view tmp = gsl_matrix_column(sampl0,i);
    if( gsl_isnan(gsl_blas_dasum(&tmp.vector)) ){
      gsl_vector_set(prob,i,.000001); /* no evidence */
      continue;
    }
    gsl_vector_sub(&tmp.vector,mu);
    gsl_blas_dgemv(CblasNoTrans, 1, icov,&tmp.vector,0,right);
    gsl_blas_ddot(&tmp.vector,right,&dot);
    /*gsl_vector_set(prob,i,gsl_sf_exp(-.5*dot));*/
    gsl_vector_set(prob,i,exp(-.5*dot));
  }
  gsl_vector_scale(prob,denom);
  
  gsl_matrix_free(icov);
  gsl_matrix_free(LU);
  gsl_permutation_free(p);
  gsl_matrix_free(sampl0);
  gsl_vector_free(right);
  return 0;
}

int obs2obslik(hd_influence * infl, hd_sample * ev){
  int i,j;
  int T = ev->s->size2;
  int nc = ev->s->size1;
  for(j=0;j<nc;j++){
    int o_type = gsl_vector_get(infl->o_type,j);
    int s_off = gsl_vector_get(infl->s_off,j);
    int ns = gsl_vector_get(infl->n_states,j);
    int n_o = gsl_vector_get(infl->n_outputs,j);
    int obs_off = gsl_vector_get(ev->obs_off,j);
    if(o_type=='c'){
      for(i=0;i<ns;i++){
          gsl_vector_view p = gsl_matrix_row(ev->obslik,s_off+i);
          gsl_vector_view mu = gsl_matrix_row(infl->B[j],i);
          gsl_matrix_view o = gsl_matrix_submatrix(ev->o,obs_off,0,n_o,T);
          hd_gaussian_prob(&mu.vector,infl->cov[j][i],&o.matrix,&p.vector);
      }
    }else{
      for(i=0;i<T;i++){
        double o = gsl_matrix_get(ev->o,obs_off,i);
        gsl_vector_view tmp1 = gsl_matrix_column(ev->obslik,i);
        gsl_vector_view tmp3 = gsl_vector_subvector(&tmp1.vector,s_off,ns);
        if(gsl_isnan(o)){
          gsl_vector_set_all(&tmp3.vector,.000001);/* no evidence */
          continue;
        }else{
          gsl_vector_view tmp2 = gsl_matrix_column(infl->B[j],o);
          gsl_vector_memcpy(&tmp3.vector,&tmp2.vector);
          /*gsl_vector_scale(&tmp3.vector,1/gsl_blas_dasum(&tmp3.vector));*/
        }
      }
    }

    /* latent state already provided */
    for(i=0;i<T;i++){
      double s = gsl_matrix_get(ev->s,j,i);
      if(!gsl_isnan(s)){
        gsl_vector_view tmp1 = gsl_matrix_column(ev->obslik,i);
        gsl_vector_view tmp3 = gsl_vector_subvector(&tmp1.vector,s_off,ns);
        gsl_vector_set_all(&tmp3.vector,0);
        gsl_vector_set(&tmp3.vector,s,1);
      }
    }

  }

  return 0;
}

int hd_influence_enter_evidence(hd_influence * infl, hd_sample * ev){
  /* alpha, beta */
  int i,j,k;
  int T=ev->s->size2,total_s = infl->A->size1;
  gsl_vector_view tmp1,tmp2,tmp3,tmp4;
  gsl_vector * tmp = gsl_vector_alloc(total_s);
  gsl_matrix * xi0 = gsl_matrix_alloc(total_s,total_s);

  obs2obslik(infl,ev);

  gsl_matrix_set_all(ev->alpha,0.0f);
  gsl_matrix_set_all(ev->beta,0.0f);
  gsl_matrix_set_all(ev->gamma,0.0f);
  gsl_matrix_set_all(ev->scale,0.0f);
  
  /* \alpha_0 <- prior */
  i=0;
  tmp2 = gsl_matrix_row(ev->alpha,i);
  tmp3 = gsl_matrix_column(ev->obslik,i);
  tmp4 = gsl_matrix_row(ev->scale,i);
  gsl_vector_memcpy(&tmp2.vector,infl->prior);
  /* \alpha_t <- \alpha_t.* obslik_t */
  gsl_vector_mul(&tmp2.vector,&tmp3.vector);

  /* scale_t <- \alpha_t\times N */
  gsl_blas_dgemv(CblasTrans, 1, infl->N,&tmp2.vector,1.0f,&tmp4.vector);
  /* \alpha_t <- \alpha_t.*scale_t */
  gsl_vector_div(&tmp2.vector,&tmp4.vector);


  for(i=1;i<T;i++){
    /* \alpha_t <- \alpha_{t-1}\times H */
    tmp1 = gsl_matrix_row(ev->alpha,i-1);
    tmp2 = gsl_matrix_row(ev->alpha,i);
    tmp3 = gsl_matrix_column(ev->obslik,i);
    tmp4 = gsl_matrix_row(ev->scale,i);
    gsl_blas_dgemv(CblasTrans,1,infl->infl,&tmp1.vector,0.0f,&tmp2.vector);
    /* \alpha_t <- \alpha_t.* obslik_t */
    gsl_vector_mul(&tmp2.vector,&tmp3.vector);

    /* scale_t <- \alpha_t\times N */
    gsl_blas_dgemv(CblasTrans, 1, infl->N,&tmp2.vector,0.0f,&tmp4.vector);
    /* \alpha_t <- \alpha_t.*scale_t */
    gsl_vector_div(&tmp2.vector,&tmp4.vector);

  }

  /* \beta */
  gsl_matrix_set_all(ev->beta,1.0f);
  /*
  tmp1 = gsl_matrix_column(ev->beta,T-1);
  gsl_vector_set_all(&tmp1.vector,1);
  for(i=T-2;i>=0;i--){
    gsl_vector_view tmp1 = gsl_matrix_column(ev->beta,i+1);
    gsl_vector_view tmp2 = gsl_matrix_column(ev->beta,i);
    gsl_vector_view tmp3 = gsl_matrix_column(ev->obslik,i+1);
    gsl_vector_view tmp4 = gsl_matrix_row(ev->scale,i+1);
    gsl_vector_memcpy(tmp,&tmp1.vector);

    gsl_vector_mul(tmp,&tmp3.vector);
    gsl_vector_div(tmp,&tmp4.vector);
    gsl_blas_dgemv(CblasNoTrans,1,infl->infl,tmp,1,&tmp2.vector);
  }
  gsl_vector_free(tmp);
  */

  /* \gamma = \alpha .* \beta */
  gsl_matrix_transpose_memcpy(ev->gamma,ev->beta);
  gsl_matrix_mul_elements(ev->gamma,ev->alpha);

  /* \sum \xi */
  for(i=0; i<T-1; i++){
    int j;
    
    gsl_matrix_memcpy(xi0,infl->infl);
    for(j=0;j<infl->xi->size2;j++){
      double factor = gsl_matrix_get(ev->alpha,i,j);
      gsl_vector_view l = gsl_matrix_row(xi0,j);
      gsl_vector_scale(&l.vector,factor);
    }
    for(j=0;j<infl->xi->size1;j++){
      double factor = gsl_matrix_get(ev->obslik,j,i+1)/
	gsl_matrix_get(ev->scale,i+1,j)*gsl_matrix_get(ev->beta,j,i+1);
      gsl_vector_view r = gsl_matrix_column(xi0,j);
      gsl_vector_scale(&r.vector,factor);
      
    }
    gsl_matrix_add(infl->xi,xi0);
  }
  gsl_matrix_free(xi0);

  tmp1 = gsl_matrix_row(ev->gamma,0);
  gsl_vector_add(infl->pi,&tmp1.vector);

  /* B0 & cov0 */
  for(i=0;i<infl->D->size1;i++){
    int o_type = gsl_vector_get(infl->o_type,i);
    int s_off = gsl_vector_get(infl->s_off,i);
    int ns = gsl_vector_get(infl->n_states,i);
    int obs_off = gsl_vector_get(ev->obs_off,i);
    int n_o = gsl_vector_get(ev->n_o,i);
    if(o_type=='c'){
      gsl_vector_view sw = gsl_vector_subvector(infl->sw,s_off,ns);
      gsl_vector_set_all(&sw.vector,0);
      for(j=0;j<T;j++){
        /* B0 */
        gsl_vector_view o2 = gsl_matrix_column(ev->o,j);
        gsl_vector_view o = gsl_vector_subvector(&o2.vector,obs_off,n_o);
	if( !gsl_isnan(gsl_blas_dasum(&o.vector)) ){
	  gsl_vector_view g2 = gsl_matrix_row(ev->gamma,j);
	  gsl_vector_view g = gsl_vector_subvector(&g2.vector,s_off,ns);
	  gsl_blas_dger(1,&g.vector,&o.vector,infl->B0[i]);
	  gsl_vector_add(&sw.vector,&g.vector);
	  /* cov0 */
	  for(k=0;k<ns;k++){
	    double w = gsl_vector_get(&g.vector,k);
	    gsl_blas_dger(w,&o.vector,&o.vector,infl->cov0[i][k]);
	  }
	}
      }
    }else{
      for(j=0;j<T;j++){
        int k = gsl_matrix_get(ev->o,obs_off,j);
	if( !gsl_isnan(k) ){ 
	  gsl_vector_view dst = gsl_matrix_column(infl->B0[i], k);
	  gsl_vector_view src0 = gsl_matrix_row(ev->gamma,j);
	  gsl_vector_view src = gsl_vector_subvector(&src0.vector,s_off,ns);
	  gsl_vector_add(&dst.vector,&src.vector);
	}
      }
    }
  }

  return 0;
}

double hd_influence_loglikelihood(hd_influence * infl, hd_sample * ev){
  int i,j, T=ev->s->size2, nc=ev->s->size1;
  double ll;
  for(i=0;i<T;i++) for(j=0;j<nc;j++){
    int s_off = gsl_vector_get(infl->s_off,j);
    ll += log(gsl_matrix_get(ev->scale,i,s_off));
  }
  return ll;
}

int hd_influence_find_mpe(hd_influence *infl,hd_sample *sampl){
  int i,k, T=sampl->alpha->size1, n_chains=infl->D->size1;

  for(i=0;i<n_chains;i++){
    int ns = gsl_vector_get(infl->n_states,i);
    int s_off = gsl_vector_get(infl->s_off,i);
    gsl_matrix_view tmp = gsl_matrix_submatrix(sampl->gamma,0,s_off,T,ns);
    for(k=0;k<T;k++){
      gsl_vector_view tmp1 = gsl_matrix_row(&tmp.matrix,k);
      gsl_matrix_set(sampl->s,i,k,gsl_vector_max_index(&tmp1.vector));
    }
  }
  return 000;
}

struct _hd_sample_view{
  hd_sample * buff;
  hd_sample sampl;

  int ndx;
  int dt;

  gsl_matrix_view s;
  gsl_matrix_view o;

  /* sufficient statistics */
  gsl_matrix_view alpha;
  gsl_matrix_view beta;
  gsl_matrix_view scale;
  gsl_matrix_view gamma;
  gsl_matrix_view obslik;
};

hd_sample_view *hd_sample_view_alloc(hd_influence *infl, int buff_size, 
				     int window_size){
  hd_sample_view * view = (hd_sample_view*)malloc(sizeof(hd_sample_view));
  view->buff = hd_sample_alloc(infl, buff_size);
  view->ndx = 0;
  view->dt = window_size;
  return view;
}

void hd_sample_view_free(hd_sample_view *sampl_view){
  free(sampl_view->buff);
  free(sampl_view);
}

hd_sample * hd_sample_view_shift(hd_sample_view * sampl_v, int tShift){

  hd_sample * sampl = sampl_v->buff;
  int T = sampl->s->size2;
  int sns = sampl->beta->size1;
  int nc = sampl->s->size1;
  int sno = sampl->o->size1;

  int t0 = (sampl_v->ndx + tShift) % T;
  int t1 = (sampl_v->ndx + tShift + sampl_v->dt) % T;
  sampl_v->ndx = t0;
  if(t0+sampl_v->dt > T){
    gsl_matrix * s2 = gsl_matrix_alloc(nc,T);
    gsl_matrix * o2 = gsl_matrix_alloc(sno,T);
    gsl_matrix_view s_1 = gsl_matrix_submatrix(sampl->s,0,t0,nc,T-t0);
    gsl_matrix_view s_2 = gsl_matrix_submatrix(sampl->s,0,0,nc,t1);
    gsl_matrix_view o_1 = gsl_matrix_submatrix(sampl->o,0,t0,sno,T-t0);
    gsl_matrix_view o_2 = gsl_matrix_submatrix(sampl->o,0,0,sno,t1);
    gsl_matrix_view s2_1 = gsl_matrix_submatrix(s2,0,0,nc, T-t0);
    gsl_matrix_view s2_2 = gsl_matrix_submatrix(s2,0,T-t0,nc,t1);
    gsl_matrix_view o2_1 = gsl_matrix_submatrix(o2,0,0,sno,T-t0);
    gsl_matrix_view o2_2 = gsl_matrix_submatrix(o2,0,T-t0,sno,t1);
    gsl_matrix_view s_a = gsl_matrix_submatrix(sampl->s,0,0,nc,t1+T-t0);
    gsl_matrix_view o_a = gsl_matrix_submatrix(sampl->o,0,0,sno,t1+T-t0);
    gsl_matrix_view s2_a = gsl_matrix_submatrix(s2,0,0,nc,t1+T-t0);
    gsl_matrix_view o2_a = gsl_matrix_submatrix(o2,0,0,sno,t1+T-t0);
    gsl_matrix_memcpy(&s2_1.matrix,&s_1.matrix);
    gsl_matrix_memcpy(&s2_2.matrix,&s_2.matrix);
    gsl_matrix_memcpy(&o2_1.matrix,&o_1.matrix);
    gsl_matrix_memcpy(&o2_2.matrix,&o_2.matrix);
    gsl_matrix_memcpy(&s_a.matrix,&s2_a.matrix);
    gsl_matrix_memcpy(&o_a.matrix,&o2_a.matrix);
    gsl_matrix_free(s2);
    gsl_matrix_free(o2);
    t0 = 0;
    sampl_v->ndx = 0;
    gsl_matrix_set_all(sampl->alpha,0);
    gsl_matrix_set_all(sampl->beta,0);
    gsl_matrix_set_all(sampl->gamma,0);
    gsl_matrix_set_all(sampl->scale,0);
    gsl_matrix_set_all(sampl->obslik,0);
  }
  
  sampl_v->sampl.s_off = sampl->s_off;
  sampl_v->sampl.n_s = sampl->n_s;
  sampl_v->sampl.n_o = sampl->n_o;
  sampl_v->sampl.obs_off = sampl->obs_off;
  sampl_v->s = gsl_matrix_submatrix(sampl->s,0,t0,nc,sampl_v->dt);
  sampl_v->sampl.s = &sampl_v->s.matrix;
  sampl_v->o = gsl_matrix_submatrix(sampl->o,0,t0,sno,sampl_v->dt);
  sampl_v->sampl.o = &sampl_v->o.matrix;
  sampl_v->alpha = gsl_matrix_submatrix(sampl->alpha,t0,0,sampl_v->dt,sns);
  sampl_v->sampl.alpha = &sampl_v->alpha.matrix;
  sampl_v->beta = gsl_matrix_submatrix(sampl->beta,0,t0,sns,sampl_v->dt);
  sampl_v->sampl.beta = &sampl_v->beta.matrix;
  sampl_v->gamma = gsl_matrix_submatrix(sampl->gamma,t0,0,sampl_v->dt,sns);
  sampl_v->sampl.gamma = &sampl_v->gamma.matrix;
  sampl_v->scale = gsl_matrix_submatrix(sampl->scale,t0,0,sampl_v->dt,sns);
  sampl_v->sampl.scale = &sampl_v->scale.matrix;
  sampl_v->obslik = gsl_matrix_submatrix(sampl->obslik,0,t0,sns,sampl_v->dt);
  sampl_v->sampl.obslik = &sampl_v->obslik.matrix;

  return &(sampl_v->sampl);
}

