/*************************************************
 * File: SignalRecorder.h
 * Author: Mark Blum (mark@media.mit.edu)
 * Birth Date: 05/12/05
 *
 * Project: ASSIST
 * 
 * 
 * Copyright (C) 2005 MIT Media Lab
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 ************************************************
 * SignalRecorder GUI-application that takes care of recording data and
 * starting and stopping classifiers. Upon request it will launch
 * required processes with system calls and will kill them on termination.
 * It includes the following intellegent components:
 * - Sensor file size, battery and CPU logging
 * - Managing sensor calibration using MITeCalibrate
 * - Manages picture taking
 * - polls the shutter button if requested
 * - runs the WiFi sniffing procedure
 * - manages online labeling
 * - computes the interest prediction algorithm and logs moments of interst
 ***********************************************/

#ifndef GroupMedia_H
#define GroupMedia_H
#include "SignalRecorder_base.h"

#include <stdio.h>
#include <qdialog.h>
#include <pthread.h>
#include <semaphore.h>
#include <qwidget.h>
#include <qobject.h>
#include <qtabwidget.h>
#include <qtimer.h>

extern "C" {
#include "linux_audio.h"
}
#include <qlabel.h>
#include <qstring.h>
#include <qlineedit.h>
#include <qcheckbox.h>
#include <qapplication.h>

#define NUM_CHAINS 4
#define MAX_LABELS_PER_CHAIN 10
#define RING_BUFFER_SIZE 50

class GroupMedia : public GroupMediaBase
{ 
    Q_OBJECT

public:
    GroupMedia( QWidget* parent = 0, const char* name = 0, WFlags f =0 );
    ~GroupMedia();
   
struct timeval label_timestamp;
          
protected:

    char* name;
    char outputFile[64];
    
private:
    
    // variables for interest measurments
    char interestPoints[NUM_CHAINS][MAX_LABELS_PER_CHAIN];
    char* labels[NUM_CHAINS][MAX_LABELS_PER_CHAIN];
    char currentClassification[NUM_CHAINS];
    char currentSuperState[NUM_CHAINS];
    char classificationRingBuffer[NUM_CHAINS][RING_BUFFER_SIZE];
    int ringBufferIndex;
    long lastMomentOfInterest; // tv_sec
    int interestThreshold;
    float currentInterest;
    float transitionInterest;
    
    QTimer *captureImageCounter;
    QTimer *checkSensorCounter;
    QTimer *buttonPollCounter;
    QTimer *showClassificationCounter;
    QTimer *wifiCounter;
    QTimer *calibrationCounter;
    int pollingButton;
    int capturingImage;
    int cameraDriver;
    int wifiState;
    FILE* labelFileID;
    int audioFileSize, MITe1FileSize, MITe2FileSize, MITe3FileSize, MITe4FileSize, wifiFileSize;
    int calibrating;
              
    int SetupEnchant();
    int TestServer();
    int DiscEnchant();
    void CheckName();
    //int CheckSignal(cSignal* sig); //char* UNL);
    int GetFileSize(char* filename);
    void CheckBattery();
    void CheckCPU();
    void SetStandbyLED(int status);
    void KillAllProcesses();

public slots:
 virtual void Quit();  
 virtual void StartRecording();
 virtual void StopRecording();
 virtual void StartClassifier();
 virtual void StopClassifier();
 virtual void LaunchServer();
 virtual void SelectPosture(int id);
 virtual void SelectActivity(int id);
 virtual void CheckAllSignals();
 virtual void CaptureImage(char* fileName);
 virtual void CaptureImageSlot();
 virtual void ShowClassification();
 virtual void PlayBuzzer();
 virtual void WifiSniffer();
 virtual void Calibrate();
 virtual void setInterestThreshold(int k);

private slots:
 virtual void PollButton();
 void InitInterestMeasurement();
 void MomentOfInterest();



};

#endif // GroupMedia_H
