/*************************************************
 * File: linux_audio.c
 * Author: unknown
 * Birth Date: unknown
 * Modified by: Mark Blum (mark@media.mit.edu)
 *
 * Project: ASSIST
 * 
 * 
 * Copyright (C) 2005 MIT Media Lab
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 ************************************************
 * This file is originally from the 'inference' modul. It is included
 * in the SignalRecorder application only for the possibility to play
 * sound on request.
 ***********************************************/

#include "linux_audio.h"

/* Global storage for speed, channels, and format information */

char *linux_audio_rdevice=NULL;
int linux_audio_rspeed = 0;
int linux_audio_rchannels = 0;
int linux_audio_rformat = 0;
int linux_audio_rdescript = -1;

char *linux_audio_wdevice=NULL;
int linux_audio_wspeed = 0;
int linux_audio_wchannels = 0;
int linux_audio_wformat = 0;
int linux_audio_wdescript = -1;

/* Configures the linux audio device for output.

   Default parameter values are used if arguments are NULL or zero.

   Returns 1 on success, zero on failure.
   File descriptor is stored in linux_audio_wdescript.  */

void linux_audio_close()
{
  if (linux_audio_wdescript != -1) {
    close(linux_audio_wdescript);
    linux_audio_wdescript=-1;
  } 
}  
    
int linux_init_write_sound(char *device,
			   int speed,
			   int channels,
			   int format) {  

				/* File descriptor previously used */
  if (linux_audio_wdescript != -1) {
    close(linux_audio_wdescript);
    linux_audio_wdescript=-1;
  }
				/* Device name previous assigned */
  if (device) {
    if (linux_audio_wdevice) {
      free(linux_audio_wdevice);
    }
    linux_audio_wdevice = strndup(device,64);
  }
  else {
    linux_audio_wdevice = strndup(DEFAULT_OUTPUT_AUDIO_DEVICE,64);
  }
				/* Store device parameters */
  linux_audio_wspeed = speed ? speed : DEFAULT_AUDIO_SPEED;
  linux_audio_wchannels = channels ? channels : DEFAULT_AUDIO_CHANNELS;
  linux_audio_wformat = format ? format : DEFAULT_AUDIO_FORMAT;
  
				/* Open the audio device */
  linux_audio_wdescript = open(linux_audio_wdevice, O_WRONLY | O_NONBLOCK);
  if( linux_audio_wdescript == -1)  {
    char buff2[128];
    sprintf(buff2,"failed to open device %s",linux_audio_wdevice);
    perror(buff2);
    return 0;
  }
				/* Configure the device through ioctls */
  if( ioctl( linux_audio_wdescript, SNDCTL_DSP_SETFMT , &linux_audio_wformat)==-1)  {
    perror("ioctl(\"SNDCTL_DSP_SETFMT\")");
    return 0;
  }

  
  if( ioctl( linux_audio_wdescript, SNDCTL_DSP_CHANNELS , &linux_audio_wchannels)==-1)  {
    perror("ioctl(\"SNDCTL_DSP_CHANNELS\")");
    return 0;
  }

  if( ioctl( linux_audio_wdescript, SNDCTL_DSP_SPEED , &linux_audio_wspeed)==-1)  {
    perror("ioctl(\"SNDCTL_DSP_SPEED\")");
    return 0;
  }
  
  return 1;
}

/* Uses the configuration information in a cSignal to configure the
   audio output system */

int linux_init_write_sound_signal(Signal *signal,
				  char *device,
				  int format) {
  int speed;
  int channels;

  if (signal) {
    speed = SignalGetSampleRate(signal);
    channels = SignalGetLength(signal,1);
  }
  else {
    speed = channels = 0;
  }
  
  return linux_init_write_sound(device, speed, channels, format);
}

/* Configure the audio input device */

int linux_init_read_sound(char *device,
			  int *speed,
			  int *channels,
			  int format) {

				/* File descriptor previously used */
  if (linux_audio_rdescript != -1) {
    close(linux_audio_rdescript);
    linux_audio_rdescript=-1;
  }
				/* Device name previous assigned */
  if (device) {
    if (linux_audio_rdevice) {
      free(linux_audio_rdevice);
    }
    linux_audio_rdevice = strndup(device,64);
  }
  else {
    linux_audio_rdevice = strndup(DEFAULT_INPUT_AUDIO_DEVICE,64);
  }

				/* Store device parameters */
  linux_audio_rspeed = (speed && *speed) ? *speed : DEFAULT_AUDIO_SPEED;
  linux_audio_rchannels = (channels && *channels) ? *channels :
    DEFAULT_AUDIO_CHANNELS;
  linux_audio_rformat = format ? format : DEFAULT_AUDIO_FORMAT;

				/* Open the audio device */
  linux_audio_rdescript = open(linux_audio_rdevice, O_RDONLY);
  if( linux_audio_rdescript == -1)  {
    char buff2[128];
    sprintf(buff2,"failed to open device %s",linux_audio_rdevice);
    perror(buff2);
    return 0;
  }
	
  if( ioctl( linux_audio_rdescript, SNDCTL_DSP_SETFMT , &linux_audio_rformat)==-1)  {
    perror("ioctl(\"SNDCTL_DSP_SETFMT\")");
    return 0;
  }

  if( ioctl( linux_audio_rdescript, SNDCTL_DSP_CHANNELS , &linux_audio_rchannels)==-1)  {
    perror("ioctl(\"SNDCTL_DSP_CHANNELS\")");
    return 0;
  }

  if( ioctl( linux_audio_rdescript, SNDCTL_DSP_SPEED , &linux_audio_rspeed)==-1)  {
    perror("ioctl(\"SNDCTL_DSP_SPEED\")");
    return 0;
  }

  if (channels)
    *channels = linux_audio_rchannels;
  if (speed)
    *speed = linux_audio_rspeed;
  return 1;
}


/* This plays back raw wave recorded at 22050, 16 bit, mono */ 
/* Does this function work or make sense in this context?  What if the
   audio settings don't match?  Shouldn't the configuration be read
   from the WAV header? */

int linux_play_wave(char *file_name) {

  signed short sound_buffer[DEFAULT_AUDIO_BUFFER_SIZE];
  int wave = 0;
  int length;
  
  if (linux_audio_wdescript == -1) {
    fprintf(stderr,"linux_play_wave: linux audio output not configured\n");
    return 0;
  }

    
  wave = open(file_name, O_RDONLY);
  if( wave == -1)  {
    perror("open(\"filename\")");
    return 0;
  }

  while( (length = read( wave,
			 sound_buffer,
			 DEFAULT_AUDIO_BUFFER_SIZE )) && (length != EOF) ) {
				/* read from the file, write to the device */
    write( linux_audio_wdescript, sound_buffer, length);
  }
  close (wave);
  return 1;
}

/* This function reads audio samples (using the current settings) into
   the signed short buffer pointed to by *buffer, of length size.  It
   returns the number of samples read.  */

int linux_audio_read( signed short *buffer, int size) {
  int length;

  if (linux_audio_rdescript == -1) {
    fprintf(stderr,"linux_audio_read: linux audio input not configured\n");
    return 0;
  }
    
  if (!buffer ) {
    perror("linux_audio_read: null buffer arg\n");
    return 0;
  }
  length = read(linux_audio_rdescript, buffer, size*sizeof(signed short));
  return length/sizeof(signed short);
}

/* This function plays audio (using the current settings) from the
   signed short buffer pointed to by *buffer, of length size.  It
   returns the number of samples played. */

int linux_audio_play( signed short *buffer, int size) {
  int length;

  if (linux_audio_wdescript == -1) {
    fprintf(stderr,"linux_audio_read: linux audio input not configured\n");
    return 0;
  }

  if (!buffer ) {
    perror("linux_audio_play: null buffer arg\n");
    return 0;
  }
  length = write(linux_audio_wdescript, buffer, size*sizeof(signed short));
  return length/sizeof(signed short);
}
  

