/*******************************************************************
 * Name: Jonathan Gips
 * 
 * Copyright (C) 2002 MIT Media Lab,
 *               2004 AWare Technologies, Inc.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 *
 * -------------------
 * File: darray.c  
 * ---------------
 * This file implements the DArray abstraction. It provides a versatile
 * solution to the problem of generic containers in C. The implementation
 * also grows pretty elegantly, but it does not shrink. This could be
 * changed if desired, but for many applications it is not necessary, I
 * would believe, especially if client data is appropriately freed.
 ********************************************************************/

#include <enchantment/darray.h>
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>

#define DEFAULT_ALLOC_LEN 8

static void GrowArray(DArray array);

/* Struct: DArrayImplementation
 * ----------------------------
 * Contains all of the information relevant to a DArray.
 * This includes, most importantly, a void * that will
 * point to a chunk of memory that will hold array elements
 * but will be traversed manually, not as a conventional array,
 * to allow for mutable types.
 *
 * ElemFreeFn is a function pointer that is dereferenced by
 * functions when it is necessary to free certain memory
 * that only the client would be able to know about.
 */

typedef struct DArrayImplementation {
	int sizeBytes;
	
	int numElemsToAllocate;
	int alloc_len;
	int logic_len;

	ArrayElementFreeFn elemFreeFn;

	void *array_p;
} DArrayImplementation;

/* ArrayNew
 * --------
 * Basically, this function allocates space for the DArray
 * and does some value checking.  If crucial values are not
 * specified, an alert kills the program. Otherwise, default
 * values may be substituted.
 */

DArray ArrayNew(int elemSize, int numElemsToAllocate, 
                ArrayElementFreeFn elemFreeFn) {

	DArrayImplementation *d_array;

	assert(elemSize > 0);

	d_array = (DArrayImplementation*)malloc(sizeof(DArrayImplementation));

	d_array->sizeBytes = elemSize;

	if(numElemsToAllocate == 0) d_array->numElemsToAllocate = DEFAULT_ALLOC_LEN;
	else d_array->numElemsToAllocate = numElemsToAllocate;

	d_array->array_p = malloc(elemSize*d_array->numElemsToAllocate);

	d_array->alloc_len = d_array->numElemsToAllocate;
	d_array->logic_len = 0;
	
	d_array->elemFreeFn = elemFreeFn;

	return d_array;
}

/* ArrayFree
 * ---------
 * This function frees all the memory associated with the
 * DArray, both client and implementation related.  First,
 * the clients function, elemFreeFn, is called on all of the
 * elements in the DArray.  Then, the memory that held those
 * elemements is released, and finally the entire 
 * DArrayImplementation struct is freed.
 */

void ArrayFree(DArray array) {
	void *elem_ptr;
	int i;

	if(array->elemFreeFn != NULL){
	  for(i = 0; i < array->logic_len; i++) {
	    elem_ptr = (char *) array->array_p +
	      array->sizeBytes*i;
	   ( *(array->elemFreeFn))(elem_ptr);
	  }
	}

	free(array->array_p);

	free(array);
}

int ArrayLength(const DArray array) {return array->logic_len;}


/* ArrayNth
 * --------
 * A simple function that is exemplary of the basic methods used
 * throughout this ADT's implementation.  Asserts are used for
 * bounds checking the client.  And, void *s are cast as char *
 * so that basic pointer arithmetic can be carried out.
 */

void *ArrayNth(DArray array, int n) {
  assert(n >= 0 && n < array->logic_len);

	return (char *) array->array_p + array->sizeBytes*n;
}

/* ArrayAppend
 * -----------
 * Appends a element onto the end of the DArray. If necessary,
 * the DArray will grow.
 */

void ArrayAppend(DArray array, const void *newElem) {
	void *new_elem_loc;

	assert(newElem != NULL);

	if(array->logic_len == array->alloc_len) GrowArray(array);

	new_elem_loc = (char *) array->array_p + array->sizeBytes*array->logic_len;
	memcpy(new_elem_loc, newElem, array->sizeBytes);
       
    array->logic_len++;
}

/* ArrayInsertAt
 * -------------
 * Inserts a new element somewhere in the DArray and moves the
 * elements beyond it one place higher.  It will also Grow the 
 * array if necessary. Memory need not be moved if the place 
 * where the element is being inserted is the last place on 
 * the DArray.
 */

void ArrayInsertAt(DArray array, const void *newElem, int n) {
  int mem_size;
  void *mem_to_shift, *new_mem_loc, *new_ptr;
	
  assert(n >= 0 && n <= array->logic_len);
  
  if(array->logic_len == array->alloc_len) GrowArray(array);
  
  new_ptr = mem_to_shift = (char *) array->array_p + n*array->sizeBytes;
  new_mem_loc = (char *) mem_to_shift + array->sizeBytes;
  mem_size = (array->logic_len - n)*array->sizeBytes;
  
  if(n != array->logic_len) memmove(new_mem_loc, mem_to_shift,
				    mem_size);
  
  memcpy(new_ptr, newElem, array->sizeBytes);
  
  array->logic_len++;
}

/* ArrayDeleteAt
 * ------------
 * Works in the opposite manner of the ArrayInsertAt.
 * The client memory freeing function is called before
 * the data is wiped out.
 */

void ArrayDeleteAt(DArray array, int n) {
  int mem_size;
  void *mem_to_shift, *delete_mem_loc;
  
  assert(n >= 0 && n <= array->logic_len - 1);
  
  delete_mem_loc = (char *) array->array_p + n*array->sizeBytes;
  if(array->elemFreeFn != NULL) (*(array->elemFreeFn))(delete_mem_loc);

  if(n != array->logic_len - 1) {
    mem_to_shift = (char *) delete_mem_loc + array->sizeBytes;
    mem_size = (array->logic_len - n - 1)*array->sizeBytes;
    memmove(delete_mem_loc, mem_to_shift, mem_size);
  }
  
  array->logic_len--;
}

void ArrayRemoveAt(DArray array, int n) {
  int mem_size;
  void *mem_to_shift, *delete_mem_loc;
  
  assert(n >= 0 && n <= array->logic_len - 1);
  
  delete_mem_loc = (char *) array->array_p + n*array->sizeBytes;

  if(n != array->logic_len - 1) {
    mem_to_shift = (char *) delete_mem_loc + array->sizeBytes;
    mem_size = (array->logic_len - n - 1)*array->sizeBytes;
    memmove(delete_mem_loc, mem_to_shift, mem_size);
  }
  
  array->logic_len--;
}

/* ArrayReplaceAt
 * -------------
 * More simple than ArrayDeleteAt, this function calls
 * elemFreeFn on the memory to be written over, and then
 * fills in the new memory.
 */
void ArrayReplaceAt(DArray array, const void *newElem, int n) {
  void *elem_ptr;
  
  assert(n >= 0 && n <= array->logic_len - 1);
  
  elem_ptr = (char *) array->array_p + n*array->sizeBytes;
  
  if(array->elemFreeFn != NULL) (*(array->elemFreeFn))(elem_ptr);
  
  memcpy(elem_ptr, newElem, array->sizeBytes);
}

/* ArraySort
 * ---------
 * Uses qsort to sort the Array according to the ordering
 * of comparator, which returns values analogous to strcmp.
 */

void ArraySort(DArray array, ArrayCompareFn comparator) {
  assert(comparator != NULL);
  
  qsort(array->array_p, array->logic_len, array->sizeBytes,
	comparator); 
}

/* ArraySearch
 * -----------
 * More complicated than the others, this function has two different
 * control sequences.  If the Array is already sorted, then bsearch is 
 * used to quickly traverse the data and come up with a match to key, 
 * according to the use of comparator. Otherwise, a slower linear search
 * is carried out over all the elements.
 */

int ArraySearch(DArray array, const void *key, ArrayCompareFn comparator,
		int fromIndex, int isSorted) {
  
  int i, comp_val;
  void *elem_ptr;
  
  assert(comparator != NULL);
  
  if(fromIndex < 0) fromIndex = 0;
  else if(fromIndex > array->logic_len - 1) return NOT_FOUND;
  
  if(isSorted) {
    elem_ptr = bsearch(key, (char *) array->array_p + fromIndex
		       *array->sizeBytes, array->logic_len, 
		       array->sizeBytes, comparator);
    if(elem_ptr != NULL) return ((char *) elem_ptr - (char *)
				 array->array_p)/array->sizeBytes;
  }
  else {
    for(i = fromIndex; i < array->logic_len; i++) {
      elem_ptr = (char *) array->array_p + i*array->sizeBytes;
      comp_val = (*comparator)(elem_ptr, key);
      if(comp_val == 0) return i;
    }
  }
  return NOT_FOUND;
}

/* ArrayMap
 * --------
 * This function will take an ArrayMapFn and apply it to every
 * element of the array.  Not that much special about it.
 */

void ArrayMap(DArray array, ArrayMapFn fn, void *clientData) {
  int i;
  
  assert(fn != NULL);
  
  for(i = 0; i < array->logic_len; i++) {
		(*fn)((char *) array->array_p + i*array->sizeBytes, clientData);
  }
}

/* GrowArray
 * ---------
 * Reallocates the data pointed at by the array_p pointer into a larger 
 * chunk of memory.  This allows the allocated length to be increased.
 */

static void GrowArray(DArray array) {
  int new_size;
  
  new_size = (array->alloc_len + array->numElemsToAllocate)*array->sizeBytes;
  
  array->array_p = realloc(array->array_p, new_size);
  
  array->alloc_len += array->numElemsToAllocate;
}

int ArrayEq(const void *a, const void *b)
{
  return *(void **)a - *(void **)b;
}
