/*************************************************
 * File: introduction.c
 * Birth Date: 4/23/04
 *
 * Project: Enchantment IPC
 * 
 * Based on Enchantment IPC Whitepaper
 *   by Edward Keyes.
 * 
 * Copyright (C) 2003 MIT Media Lab
 *               2004 AWare Technologies, Inc.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 *
 *************************************************/

#include "introduction.h"
#include "socket.h"
#include "ipc.h"

#include <string.h>
#include <stdlib.h>
#include <stdio.h>

#define DEBUG_INTRODUCTION 0

int IntroSend(Socket *socket, const IntroMessage *message)
{
  unsigned len;
  char *key;
  TypeID type;
  if (SocketNoWrite(socket))
    {
      if (DEBUG_INTRODUCTION)
	printf("Not writing %s to file\n", message->format->key);
      return 1;
    }
  key = message->format->key;
  len = strlen(key);
  if (message->format->hasIndex)
    {
      key = malloc(len + 3);
      strcpy(key, message->format->key);
      key[len] = '-';
      key[len + 1] = message->index + '0';
      key[len + 2] = '\0';
      len += 2;
    }
  else
    key = message->format->key;
  SocketSendRecord(socket, key, len);
  if (message->format->hasIndex)
    free(key);
  if (message->format->expected == T_EMPTY)
    type = message->type;
  else
    type = message->format->expected;
  WriteType(socket, type);
  WriteValue(socket, type, message->value);
  if (DEBUG_INTRODUCTION)
    printf("Sent %p property '%s'\n", socket, message->format->key);
  return 1;
}

int IntroRecv(Socket *socket, IntroFormat **table, IntroMessage *message)
{
  int saw_invalid = 0;
  message->format = NULL;
  do
    {
      char *buffer;
      size_t size;
      int i;
      int valid = 1;
      if (SocketNoRead(socket))
	{
	  if (DEBUG_INTRODUCTION)
	    printf("Not reading from file\n");
	  // this is actually a file...
	  return 0;
	}
      if (!SocketRecvRecord(socket, &buffer, &size))
	{
	  if (DEBUG_INTRODUCTION)
	    printf("Got disconnect from %p\n", socket);
	  return saw_invalid;
	}
      if (!size)
	{
	  if (DEBUG_INTRODUCTION)
	    printf("Change from %p\n", socket);
	  return saw_invalid;
	}
      for (i = 0; table[i]; i++)
	{
	  int keylen = strlen(table[i]->key);
	  int index = -1;
	  if ((!strncmp(buffer, table[i]->key, keylen)) &&
	      (table[i]->hasIndex ?
	       (buffer[keylen] == '-' && 
		((index = (buffer[keylen + 1] - '0')) >= 0 && index <= 9) &&
		size == keylen + 2) :
	       size == keylen))
	    {
	      message->index = index;
	      message->format = table[i];
	      break;
	    }
	}
      if (!table[i])
	{
	  valid = 0;
	  if (DEBUG_INTRODUCTION)
	    printf("Unknown property: '%s'\n", buffer);
	}
      ReadType(socket, &message->type);
      if (valid && table[i]->expected != T_EMPTY &&
	  table[i]->expected != message->type)
	{
	  if (DEBUG_INTRODUCTION)
	    printf("Unexpected type for property '%s' expected %x was %x\n",
		   buffer, table[i]->expected, message->type);
	  valid = 0;
	}
      ReadValue(socket, message->type, &message->value);
      if (!valid)
	{
	  if (DEBUG_INTRODUCTION)
	    printf("Invalid interaction\n");
	  saw_invalid = -1;
	  continue;
	}
      if (DEBUG_INTRODUCTION)
	printf("%p Got a '%s' property\n", socket, table[i]->key);
      return 1;
    }
  while (1);
}

void IntroFree(IntroMessage *message)
{
  if (message->format && message->format->expected == T_STRING)
    free(message->value.string_val);
  message->format = NULL;
}

int IntroDone(Socket *socket)
{
  if (SocketNoWrite(socket))
    return 1;
  if (DEBUG_INTRODUCTION)
    printf("Done sending to %p\n", socket);
  SocketSendRecord(socket, NULL, 0);
  SocketFlush(socket);
  return 1;
}
