/*************************************************
 * File: ipc.c
 * Author: Jonathan Gips
 * Birth Date: 6/27/01
 *
 * Project: Enchantment IPC
 * 
 * Based on Enchantment IPC Whitepaper
 *   by Edward Keyes.
 * 
 * Copyright (C) 2002, 2003, 2004 MIT Media Lab
 *               2004 AWare Technologies, Inc.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 *
 *************************************************/

#include <enchantment/dstring.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <assert.h>
#include <string.h>

#ifdef SOLARIS
#include <sys/uio.h>
#endif

#include "ipc.h"
#include "socket.h"

#define MAX_NAME_LEN 80

// Note: this is used converting in both directions, under the assumption
// that host byte order is a self-inverting permutation of network byte order.
// This is true of big, little, and even pdp.

// There's at least ARM which has little-endian words, but combines them
// big-endian to make 8-byte floating point values. That is: 56781234

// This doesn't deal properly with the combination big-endian byte
// order and little-endian float word order.

// This also handles changing element sizes; the buffer must be large
// enough for the larger size.

void ntoh(char *buffer, size_t ineltsize, size_t outeltsize, 
	  unsigned length, TypeID type)
{
  int i;
  long j;
  double f;
#if BYTE_ORDER == LITTLE_ENDIAN
  char buf[16];
  if (ineltsize > 1)
    {
      for (i = 0; i < length; i++)
	{
#if __FLOAT_WORD_ORDER == __BIG_ENDIAN
	  if ((type & T_NUMERIC_FORMAT_MASK) == T_FLOAT_FORMAT)
	    {
	      int k;
	      for (j = 0; j < ineltsize; j += 4)
		{
		  for (k = 0; k < 4; k++)
		    {
		      buf[j + k] = buffer[j + 3 - k + i * ineltsize];
		    }
		}
	    }
	  else
#endif
	    {
	      for (j = 0; j < ineltsize; j++)
		{
		  buf[j] = buffer[ineltsize - j - 1 + i * ineltsize];
		}
	    }
	  memcpy(buffer + i * ineltsize, buf, ineltsize);
	}
    }
#endif
  if (ineltsize == outeltsize)
    return;
  if (ineltsize < outeltsize)
    {
      for (i = length - 1; i + 1; i--)
	{
	  switch (type & T_NUMERIC_FORMAT_MASK)
	    {
	    case T_INT_FORMAT:
	      if (ineltsize == 1)
		j = ((signed char *)buffer)[i];
	      else if (ineltsize == 2)
		j = ((short *)buffer)[i];
	      else
		j = ((int *)buffer)[i];
	      if (outeltsize == 1)
		((signed char *)buffer)[i] = j;
	      else if (outeltsize == 2)
		((short *)buffer)[i] = j;
	      else
		((int *)buffer)[i] = j;
	      break;
	    case T_FLOAT_FORMAT:
	      if (ineltsize == 4)
		f = ((float *)buffer)[i];
	      else
		f = ((double *)buffer)[i];
	      if (outeltsize == 4)
		((float *)buffer)[i] = f;
	      else
		((double *)buffer)[i] = f;
	      break;
	    default:
	      ;
	      // Type not supported for resizing...
	    }
	}
    }
  else
    {
      for (i = 0; i < length; i++)
	{
	  switch (type & T_NUMERIC_FORMAT_MASK)
	    {
	    case T_INT_FORMAT:
	      if (ineltsize == 1)
		j = ((signed char *)buffer)[i];
	      else if (ineltsize == 2)
		j = ((short *)buffer)[i];
	      else
		j = ((int *)buffer)[i];
	      if (outeltsize == 1)
		((signed char *)buffer)[i] = j;
	      else if (outeltsize == 2)
		((short *)buffer)[i] = j;
	      else
		((int *)buffer)[i] = j;
	      break;
	    case T_FLOAT_FORMAT:
	      if (ineltsize == 4)
		f = ((float *)buffer)[i];
	      else
		f = ((double *)buffer)[i];
	      if (outeltsize == 4)
		((float *)buffer)[i] = f;
	      else
		((double *)buffer)[i] = f;
	      break;
	    default:
	      printf("Problem with types\n");
	      // Type not supported for resizing...
	    }
	}
    }
}

void convcopy(char *in, char *out, size_t ineltsize, size_t outeltsize, 
	      unsigned length, TypeID type)
{
  long j;
  double f;
  switch (type & T_NUMERIC_FORMAT_MASK)
    {
    case T_INT_FORMAT:
      if (ineltsize == 1)
	j = *((int8_t *)in);
      else if (ineltsize == 2)
	j = *((int16_t *)in);
      else
	j = *((int32_t *)in);
      if (outeltsize == 1)
	*((int8_t *)out) = j;
      else if (outeltsize == 2)
	*((int16_t *)out) = j;
      else
	*((int32_t *)out) = j;
      break;
    case T_FLOAT_FORMAT:
      if (ineltsize == 4)
	f = *((float *)in);
      else
	f = *((double *)in);
      if (outeltsize == 4)
	*((float *)out) = f;
      else
	*((double *)out) = f;
      break;
    default:
      printf("Problem with types\n");
    }
}

typedef struct
{
  char *data;
  size_t size;
} Unknown;

void ReadString(Socket *socket, char **result, size_t *size_p)
{
  size_t size;
  unsigned char part;
  SocketRecv(socket, &part, 1);
  size = part;
  *result = (char *)malloc(size + 1);
  SocketRecv(socket, *result, size);
  (*result)[size] = '\0';
  if (part)
    SocketRecv(socket, &part, 1);
  if (size_p)
    *size_p = size;
}

void ReadNumeric(Socket *socket, TypeID type, EnchantData *value)
{
  unsigned char buffer[8];
  size_t size = 1 << (type & 0x3);
  SocketRecv(socket, buffer, size);
  ntoh(buffer, size, size, 1, type);
  switch (type & 0xF03)
    {
    case 0x0:
      value->int_val = *((int8_t *)buffer);
      break;
    case 0x1:
      value->int_val = *((int16_t *)buffer);
      break;
    case 0x2:
      value->int_val = *((int32_t *)buffer);
      break;
    case 0x102:
      value->float_val = *((float *)buffer);
      break;
    case 0x103:
      value->doubleptr_val = malloc(sizeof(double));
      *value->doubleptr_val = *((double *)buffer);
      break;
    }
}

void ReadType(Socket *socket, TypeID *type)
{
  unsigned char buffer[4];
  SocketRecv(socket, buffer, 4);
  *type = 
    (((((buffer[0] << 8) + buffer[1]) << 8) + buffer[2]) << 8) + buffer[3];
}

void ReadNetsock(Socket *socket, NetSocketLocation **addr_p)
{
  unsigned char temp;
  unsigned char len;
  int i;
  char *name;
  unsigned short port;

  SocketRecv(socket, (char *) &port, 2);
  SocketRecv(socket, (char *) &temp, 1);
  SocketRecv(socket, (char *) &len, 1);
  switch (temp)
    {
    case 0:
      ReadString(socket, &name, NULL);
      *addr_p = CreateNetSocketLocation(name, port);
      free(name);
      break;
    case 84:
      *addr_p = malloc(sizeof(NetSocketLocation));
      (*addr_p)->port = port;
      SocketRecv(socket, (char *) &(*addr_p)->ip, 4);
      break;
    default:
      // No idea...
      for (i = 0; i < len; i++)
	SocketRecv(socket, (char *) &temp, 1);
      *addr_p = NULL;
      return;
    }
  NetSocketLocationMerge(*addr_p, SocketRemote(socket));
}

void ReadTimeval(Socket *socket, struct timeval *time_p)
{
  unsigned char buffer[8];
  SocketRecv(socket, buffer, 8);
  ntoh(buffer, 4, 4, 2, T_INT32);
  time_p->tv_sec = ((int *) buffer)[0];
  time_p->tv_usec = ((int *) buffer)[1] / 1000;
}

void ReadTensor(Socket *socket, TypeID type, EnchantData *value)
{
  unsigned char buffer[8];
  // Note: we do the endianness conversion of the data in the tensor's
  // storage, because nothing else could be using it yet.
  switch ((type >> 12) & 0xF)
    {
    case 1:
      SocketRecv(socket, buffer, 4);
      ntoh(buffer, 4, 4, 1, T_INT32);
      value->vector_val = NewVector(*((unsigned *)buffer));
      SocketRecv(socket, (unsigned char *)value->vector_val->values, 
		 8 * value->vector_val->dimension);
      ntoh((unsigned char *)value->vector_val->values, 8, 8, 
	   value->vector_val->dimension,
	   T_DOUBLE);
      break;
    case 2:
      SocketRecv(socket, buffer, 8);
      ntoh(buffer, 4, 4, 2, T_INT32);
      value->matrix_val = NewMatrix(((unsigned *)buffer)[0],
				    ((unsigned *)buffer)[1]);
      SocketRecv(socket, (unsigned char *)value->matrix_val->values, 
		 8 * value->matrix_val->rows * value->matrix_val->cols);
      ntoh((unsigned char *)value->matrix_val->values, 8, 8, 
	   value->matrix_val->rows * value->matrix_val->cols,
	   T_DOUBLE);
      break;
    }
}

void ReadCharacter(Socket *socket, int *val)
{
  unsigned char buffer[4];
  SocketRecv(socket, buffer, 4);
  ntoh(buffer, 4, 4, 1, T_INT32);
  *val = ((int *) buffer)[0];
}

void ReadBool(Socket *socket, int *val)
{
  unsigned char c;
  SocketRecv(socket, &c, 1);
  *val = !!c;
}

void ReadUnknown(Socket *socket, EnchantData *value)
{
  Unknown *result = malloc(sizeof(Unknown));
  ReadString(socket, &result->data, &result->size);
  value->empty_val = result;
}

void ReadValue(Socket *socket, TypeID type, EnchantData *value)
{
  switch (GetFormat(type))
    {
    case T_EMPTY:
      break;
    case T_NUMERIC:
      if (type & 0xF000)
	ReadTensor(socket, type, value);
      else
	ReadNumeric(socket, type, value);
      break;
    case T_STRING:
      ReadString(socket, &value->string_val, NULL);
      break;
    case T_TYPE:
      ReadType(socket, &value->type_val);
      break;
    case T_NETSOCK:
      ReadNetsock(socket, &value->netsock_val);
      break;
    case T_BRANCH:
      // Note: not used by ReadNode, which does special things in this case
      break;
    case T_TIMEVAL:
      value->timeval_val = malloc(sizeof(struct timeval));
      ReadTimeval(socket, value->timeval_val);
      break;
    case T_CHARACTER:
      ReadCharacter(socket, &value->int_val);
      break;
    case T_BOOL:
      ReadBool(socket, &value->int_val);
      break;
    default:
      printf("Got unsupported type %x\n", type);
      if (type >= T_UNKNOWN)
	ReadUnknown(socket, value);
      else
	printf("XXXX Should understand it?\n");
    }
}

void WriteNumeric(Socket *socket, TypeID type, EnchantData value)
{
  unsigned char buffer[8];
  size_t size = 1 << (type & 0x3);
  switch (type & 0xF03)
    {
    case 0x0:
      *buffer = value.int_val;
      break;
    case 0x1:
      *((int16_t *)buffer) = value.int_val;
      break;
    case 0x2:
      *((int32_t *)buffer) = value.int_val;
      break;
    case 0x102:
      *((float *)buffer) = value.float_val;
      break;
    case 0x103:
      *((double *)buffer) = *value.doubleptr_val;
      break;
    }
  ntoh(buffer, size, size, 1, type);
  SocketSend(socket, buffer, size);
}

void WriteTensor(Socket *socket, TypeID type, EnchantData value)
{
  unsigned char buffer[8];
  unsigned i, j;
  Vector *vector;
  Matrix *matrix;
  switch ((type >> 12) & 0xF)
    {
    case 1:
      vector = value.vector_val;
      *((unsigned int *)buffer) = vector->dimension;
      ntoh(buffer, 4, 4, 1, T_INT32);
      SocketSend(socket, buffer, 4);
      for (i = 0; i < vector->dimension; i++)
	{
	  *((double *)buffer) = VectorRef(vector, i);
	  ntoh(buffer, 8, 8, 1, T_DOUBLE);
	  SocketSend(socket, buffer, 8);
	}
      break;
    case 2:
      matrix = value.matrix_val;
      ((unsigned int *)buffer)[0] = matrix->rows;
      ((unsigned int *)buffer)[1] = matrix->cols;
      ntoh(buffer, 4, 4, 2, T_INT32);
      SocketSend(socket, buffer, 8);
      for (i = 0; i < matrix->rows; i++)
	{
	  for (j = 0; j < matrix->cols; j++)
	    {
	      *((double *)buffer) = MatrixRef(matrix, i, j);
	      ntoh(buffer, 8, 8, 1, T_DOUBLE);
	      SocketSend(socket, buffer, 8);
	    }
	}
      break;
    }
}

void WriteString(Socket *socket, const char *value, size_t length)
{
  unsigned char sz;
  while (length > 0)
    {
      if (length > 255)
	sz = 255;
      else
	sz = length;
      SocketSend(socket, &sz, 1);
      SocketSend(socket, value, sz);
      length -= sz;
      value += sz;
    }
  sz = 0;
  SocketSend(socket, &sz, 1);
}

void WriteNetsock(Socket *socket, NetSocketLocation *addr)
{
  unsigned char temp;
  SocketSend(socket, (char *) &addr->port, 2);
  temp = 84;
  SocketSend(socket, (char *) &temp, 1);
  temp = 0x04;
  SocketSend(socket, (char *) &temp, 1);
  SocketSend(socket, (char *) &addr->ip, 4);
}

void WriteType(Socket *socket, TypeID type)
{
  unsigned char buffer[4];
  buffer[0] = (type >> 24) & 0xFF;
  buffer[1] = (type >> 16) & 0xFF;
  buffer[2] = (type >> 8) & 0xFF;
  buffer[3] = (type >> 0) & 0xFF;
  SocketSend(socket, buffer, 4);
}

void WriteTimeval(Socket *socket, struct timeval *time_p)
{
  unsigned char buffer[8];
  ((int *) buffer)[0] = time_p->tv_sec;
  ((int *) buffer)[1] = time_p->tv_usec * 1000;
  ntoh(buffer, 4, 4, 2, T_INT32);
  SocketSend(socket, buffer, 8);
}

void WriteCharacter(Socket *socket, int character)
{
  unsigned char buffer[4];
  ((int *) buffer)[0] = character;
  ntoh(buffer, 4, 4, 1, T_INT32);
  SocketSend(socket, buffer, 4);
}

void WriteBool(Socket *socket, int val)
{
  char c = !!val;
  SocketSend(socket, &c, 1);
}

static void writeSubNode(Socket *socket, NodeType *node, IPCNodeProc proc,
			 void *data)
{
  WriteType(socket, node->type);
  if (GetFormat(node->type) != T_BRANCH)
    {
      WriteValue(socket, node->type, node->data);
    }
  else
    {
      unsigned index;
      NodeType *child;
      for (index = 0; (child = BranchIndex(node, index)); index++)
	{
	  if (proc)
	    child = (*proc)(data, child);
	  if (child)
	    {
	      WriteString(socket, child->name, strlen(child->name));
	      writeSubNode(socket, child, proc, data);
	    }
	}
      WriteString(socket, NULL, 0);
    }
}

static void writeBranch(Socket *socket, EnchantData value)
{
  unsigned index;
  NodeType *child;
  for (index = 0; (child = BranchValueIndex(value, index)); index++)
    {
      WriteString(socket, child->name, strlen(child->name));
      writeSubNode(socket, child, NULL, NULL);
    }
  WriteString(socket, NULL, 0);
}

void WriteUnknown(Socket *socket, EnchantData value)
{
  Unknown *val = value.empty_val;
  WriteString(socket, val->data, val->size);
}

void WriteValue(Socket *socket, TypeID type, EnchantData value)
{
  switch (GetFormat(type))
    {
    case T_EMPTY:
      break;
    case T_NUMERIC:
      if (type & 0xF000)
	WriteTensor(socket, type, value);
      else
	WriteNumeric(socket, type, value);
      break;
    case T_STRING:
      WriteString(socket, value.string_val, strlen(value.string_val));
      break;
    case T_TYPE:
      WriteType(socket, value.type_val);
      break;
    case T_TIMEVAL:
      WriteTimeval(socket, value.timeval_val);
      break;
    case T_NETSOCK:
      WriteNetsock(socket, value.netsock_val);
      break;
    case T_BRANCH:
      // Note: not used by WriteNode, which does special things in this case
      writeBranch(socket, value);
      break;
    case T_BOOL:
      WriteBool(socket, value.int_val);
      break;
    case T_CHARACTER:
      WriteCharacter(socket, value.int_val);
      break;
    default:
      if (type >= T_UNKNOWN)
	WriteUnknown(socket, value);
      else
	printf("XXXX Unknown low type %x?\n", type);
    }
}

void WriteNode(Socket *socket, NodeType *node, IPCNodeProc proc, void *data)
{
  if (proc)
    {
      node = (*proc)(data, node);
      if (!node)
	return; // XXXX shouldn't happen
    }
  
  writeSubNode(socket, node, proc, data);
}

void ReadNode(Socket *socket, NodeType **node_p, IPCNodeProc proc, void *data)
{
  TypeID type;
  EnchantData value;
  NodeType *node;

  ReadType(socket, &type);
  if (!type)
    {
      *node_p = NULL;
      return;
    }
  if (GetFormat(type) != T_BRANCH)
    {
      ReadValue(socket, type, &value);
      node = CreateTypedNode("root", type, &value);
    }
  else
    {
      char *name;
      size_t len;
      node = CreateBranchNode("root", type);
      do
	{
	  NodeType *child;
	  ReadString(socket, &name, &len);
	  if (!len)
	    {
	      free(name);
	      break;
	    }
	  ReadNode(socket, &child, proc, data);
	  child->name = InstanceStringN(name, len);
	  free(name);
	  InsertNode(node, child);
	}
      while (1);
    }
  if (proc)
    node = (*proc)(data, node);
  *node_p = node;
  return;
}
