/*************************************************
 * File: socket.c
 * Author: Daniel Barkalow
 * Birth Date: 2/4/03
 *
 * Project: Enchantment IPC
 * 
 * Based on Enchantment IPC Whitepaper
 *   by Edward Keyes.
 * 
 * Copyright (C) 2003 MIT Media Lab
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 *
 *************************************************/

#include "socket.h"
#include <netinet/in.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/socket.h>
#include <string.h>
#include <fcntl.h>
#include <sys/time.h>
#include <netdb.h>
#include "ipc.h"
#include <errno.h>

#include <stdio.h>

#include <signal.h>

#define DEBUG_SOCKETS 0

#ifdef WANT_SSL
//#include "sec.h"

#include <openssl/bio.h>
#include <openssl/err.h>
//#include <openssl/rand.h>
#include <openssl/ssl.h>
//#include <openssl/x509v3.h>

struct Identity
{
  SSL_CTX *ctx;
};

typedef BIO *sockd;

#define sockd_fd(sockd) (BIO_get_fd((sockd), NULL))

#define sockd_write(sockd, buf, size) BIO_write(sockd, buf, size)

#define sockd_read(sockd, buf, size) BIO_read(sockd, buf, size)

#define sockd_free(sockd) BIO_free_all(sockd)

#else // !WANT_SSL

typedef int sockd;

#define sockd_fd(sockd) ((long int) sockd)

#define sockd_write(sockd, buf, size) write(sockd, buf, size)

#define sockd_read(sockd, buf, size) read(sockd, buf, size)

#define sockd_free(sockd) close(sockd)

#endif

struct Socket
{
  union
  {
    sockd single;
    DArray multi;
  } bio;
  unsigned multipoint : 1;
  unsigned writing : 1;
  unsigned inObject : 1;
  unsigned atEnd : 1;
  unsigned mayRead : 1;
  unsigned mayWrite : 1;
  char *inbuf;
  unsigned inbuflen;
  unsigned inbufpos;
  char *outbuf;
  unsigned outbuflen;
  unsigned outbufpos;
  char *name;
  int err;
  NetSocketLocation *local; // memory owned by the socket unless parent
  Socket *parent;
  union 
  {
    NetSocketLocation *remote;
    int lsnr;
  } ext;
};

struct SocketFactory
{
  int fd;
  int allowUnreliable;
  NetSocketLocation *local;
};

/* Returns a Socket with any necessary defaults set */
static Socket *allocSocket()
{
  Socket *ret = malloc(sizeof(Socket));
  ret->inbuf = malloc(8);
  ret->inbuflen = 8;
  ret->inbufpos = 0;
  ret->outbuf = malloc(8);
  ret->outbuflen = 8;
  ret->outbufpos = 0;
  ret->local = NULL;
  ret->ext.remote = NULL;
  ret->multipoint = 0;
  ret->name = NULL;
  ret->writing = 1;
  ret->err = 0;
  ret->inObject = 0;
  ret->parent = NULL;
  ret->mayRead = 1;
  ret->mayWrite = 1;
  return ret;
}

static int totalOut = 0;
static int totalIn = 0;
static int totalOpsOut = 0;
static int totalOpsIn = 0;

#define BUFMAX 2048


Socket *SocketConnect(NetSocketLocation *target)
{
  Socket *ret;
  int fd;
  sockd sbio;
  socklen_t addrLen = sizeof(struct sockaddr_in);
  struct sockaddr_in addr;

  fd = socket(AF_INET, SOCK_STREAM, 0);
  if (fd == -1)
    {
      return NULL;
    }
  /*printf("Connect to %x:%x = %d.%d.%d.%d:%d\n",
	 target->ip, target->port,
	 ntohl(target->ip) >> 24,
	 (ntohl(target->ip) >> 16) & 255,
	 (ntohl(target->ip) >> 8) & 255,
	 (ntohl(target->ip)) & 255,
	 ntohs(target->port));*/
  addr.sin_family = AF_INET;
  addr.sin_port = target->port;
  addr.sin_addr.s_addr = target->ip;
  if (connect(fd, (struct sockaddr *) &addr, sizeof(struct sockaddr_in)) == -1)
    {
      return NULL;
    }

#ifdef WANT_SSL
  sbio = BIO_new_socket(fd, 1);
#else
  sbio = fd;
#endif

  ret = allocSocket();
  ret->bio.single = sbio;

  ret->ext.remote = malloc(sizeof(NetSocketLocation));
  *ret->ext.remote = *target;
  getsockname(fd, (struct sockaddr *) &addr, &addrLen);
  ret->local = CreateNetSocketLocation(NULL, ntohs(addr.sin_port));
  return ret;
}

SocketFactory *SocketListen(int port, int allowUnreliable)
{
  SocketFactory *ret;
  int fd;
  struct sockaddr_in addr;

  bzero(&addr, sizeof(struct sockaddr_in));
  addr.sin_family = AF_INET;
  addr.sin_port = htons(port);
  addr.sin_addr.s_addr = htonl(INADDR_ANY);

  fd = socket(AF_INET, SOCK_STREAM, 0);
  if (fd == -1)
    return NULL;
  if (bind(fd, (struct sockaddr *) &addr, sizeof(struct sockaddr_in)) < 0)
    {
      return NULL;
    }
  if (listen(fd, 5) == -1)
    {
      return NULL;
    }

  fcntl(fd, F_SETFL, O_NONBLOCK);

  ret = malloc(sizeof(SocketFactory));
  ret->fd = fd;
  ret->allowUnreliable = allowUnreliable;
  ret->local = CreateNetSocketLocation(NULL, ntohs(addr.sin_port));
  return ret;
}

Socket *SocketAccept(SocketFactory *factory)
{
  Socket *ret;
  int fd;
  sockd abio;
  struct sockaddr_in addr;
  socklen_t size = sizeof(struct sockaddr_in);

  fd = accept(factory->fd, (struct sockaddr *) &addr, &size);
  if (fd == -1)
    {
      if (errno != EAGAIN)
	  printf("Error accepting %p(%d): %s\n", factory, factory->fd, 
	       strerror(errno));
      return NULL;
    }

#ifdef WANT_SSL
  abio = BIO_new_socket(fd, 1);
#else
  abio = fd;
#endif

  ret = allocSocket();
  ret->bio.single = abio;
  ret->ext.remote = malloc(sizeof(NetSocketLocation));
  ret->ext.remote->ip = addr.sin_addr.s_addr;
  ret->ext.remote->port = addr.sin_port;
  ret->local = CopyNetSocketLocation(factory->local);
  return ret;
}

Socket *MultipointListen(int port, int allowUnreliable)
{
  Socket *ret;
  int fd;
  struct sockaddr_in addr;
  socklen_t addrLen = sizeof(struct sockaddr_in);

  bzero(&addr, sizeof(struct sockaddr_in));
  addr.sin_family = AF_INET;
  addr.sin_port = htons(port);
  addr.sin_addr.s_addr = htonl(INADDR_ANY);

  fd = socket(AF_INET, SOCK_STREAM, 0);
  if (fd == -1)
    return NULL;
  if (bind(fd, (struct sockaddr *) &addr, sizeof(struct sockaddr_in)) == -1)
    return NULL;
  if (listen(fd, 5) == -1)
    return NULL;
  if (getsockname(fd, (struct sockaddr *) &addr, &addrLen) == -1)
    return NULL;
  fcntl(fd, F_SETFL, O_NONBLOCK);

  ret = allocSocket();
  ret->multipoint = 1;
  ret->bio.multi = ArrayNew(sizeof(sockd), 2, NULL);
  ret->ext.lsnr = fd;
  ret->writing = 0;
  ret->local = CreateNetSocketLocation(NULL, ntohs(addr.sin_port));
  return ret;
}

Socket *MultipointAccept(Socket *socket)
{
  int fd;
  sockd abio;
  struct sockaddr_in addr;
  socklen_t size = sizeof(addr);
  Socket *ret;

  fd = accept(socket->ext.lsnr, 
	      (struct sockaddr*) &addr, &size);
  if (fd == -1)
    {
      if (errno != EAGAIN)
	printf("Error: %s\n", strerror(errno));
      return NULL;
    }
  printf("Accepted client on fd %d\n", fd);
  
#ifdef WANT_SSL
  abio = BIO_new_socket(fd, 1);
#else
  abio = fd;
#endif

  ret = allocSocket();
  ret->bio.single = abio;
  ret->ext.remote = malloc(sizeof(NetSocketLocation));
  ret->ext.remote->ip = addr.sin_addr.s_addr;
  ret->ext.remote->port = addr.sin_port;
  ret->local = socket->local;
  ret->parent = socket;
  return ret;
}

static void socketJoin(Socket *parent, Socket *child)
{
  // This deals with getting a child socket folded back into the
  // parent socket when the child socket is closed.
  // Doesn't handle cleaning up the child socket.
  //printf("Put %d into common client list.\n", child->fd.single);
  ArrayAppend(parent->bio.multi, &(child->bio.single));
}

Socket *SocketFile(int fd)
{
  int flags;
  Socket *ret = allocSocket();
  sockd fbio;
#ifdef WANT_SSL
  fbio = BIO_new_fd(fd, 1);
#else
  fbio = fd;
#endif
  
  flags = fcntl(fd, F_GETFL);

  switch (flags & O_ACCMODE)
    {
    case O_RDONLY:
      ret->mayWrite = 0;
      break;
    case O_WRONLY:
      ret->mayRead = 0;
      break;
    }

  ret->bio.single = fbio;
  return ret;
}

static int doSocketSend(Socket *socket, const char *buffer, size_t size)
{
  int success = 1;
  int ret;
  int i;
  sockd sbio;

  void *sig = signal(SIGPIPE, SIG_IGN);
  if (socket->multipoint)
    {
      //printf("multisend %d->%x\n", size, socket);
      for (i = 0; i < ArrayLength(socket->bio.multi); i++)
	{
	  sbio = *((sockd *)ArrayNth(socket->bio.multi, i));
	  ret = sockd_write(sbio, buffer, size);
	  if (ret != size)
	    {
	      success = 0; // ????
	      printf("Lost client on fd %ld\n", sockd_fd(sbio));
	      sockd_free(sbio);
	      ArrayDeleteAt(socket->bio.multi, i);
	      i--;
	    }
	}
      ret = size;
    }
  else
    {
      //printf("send %d->%x\n", size, socket);
      ret = sockd_write(socket->bio.single, buffer, size);
    }

  if (ret > 0)
    {
      totalOpsOut++;
      totalOut += ret;
    }
  else if (ret < 0)
    {
      socket->err = errno;
    }
  else
    socket->err = 0;

  if (sig != SIG_ERR)
    signal(SIGPIPE, sig);
  if (!success)
    { 
      return -1; // Before, success would be returned despite socket send failure.
    }
  return ret;
}

int SocketSend(Socket *socket, const char *buffer, size_t size)
{
  // This is the case where we don't buffer:
  //  if it is not in an object, and it's local or multipoint outside 
  //  of start/end
  int objsz;
  if (!socket->inObject &&
      (socket->multipoint ? socket->writing : !socket->ext.remote))
    return doSocketSend(socket, buffer, size);
  if (!size)
    return size;
  if (size + socket->outbufpos < BUFMAX)
    {
      // We can fit the data in the buffer
      if (size + socket->outbufpos > socket->outbuflen)
	{
	  while (size + socket->outbufpos > socket->outbuflen)
	    socket->outbuflen *= 2;
	  socket->outbuf = realloc(socket->outbuf, socket->outbuflen);
	}
      memcpy(socket->outbuf + socket->outbufpos, buffer, size);
      socket->outbufpos += size;
      return size;
    }

  // Here we can't fit the data in the buffer, so we'll send it all.

  if (socket->outbufpos)
    {
      // Already have some data in the buffer
      if (socket->inObject)
	{
	  objsz = htonl(socket->outbufpos);
	  doSocketSend(socket, (char *) &objsz, 4);
	}
      doSocketSend(socket, socket->outbuf, socket->outbufpos);
      if (size < BUFMAX)
	{
	  // With the old data gone, the new data fits.

	  // ???? Prove or fix: might this not fit?
	  if (size > socket->outbuflen)
	    {
	      printf("Copy out of range!");
	    }
	  memcpy(socket->outbuf, buffer, size);
	  socket->outbufpos = size;
	  return size;
	}
      else
	{
	  // Cleared the buffer
	  socket->outbufpos = 0;
	}
    }

  // We have to send this chunk, too.

  if (socket->inObject)
    {
      objsz = htonl(size);
      doSocketSend(socket, (char *) &objsz, 4);
    }
  return doSocketSend(socket, buffer, size);
}

int SocketSendRecord(Socket *socket, const char *buffer, size_t size)
{
  char buf[4];
  int sz;
  //printf("sending a record of size %d\n", size);
  *((unsigned long *)buf) = htonl(size);
  sz = SocketSend(socket, buf, 4);
  if (sz != 4)
    return 0;
  sz = SocketSend(socket, buffer, size);
  if (sz != size)
    return 0;
  return 1;
}

int SocketBeginObject(Socket *socket)
{
  if (socket->inObject)
    return 0;
  socket->atEnd = 0;
  socket->inObject = 1;
  return 1;
}

int SocketEndObject(Socket *socket)
{
  int size = 0;
  int ret;
  if (!socket->inObject)
    return 0;
  socket->inObject = 0;
  if (!socket->writing)
    return 1;
  ret = SocketSend(socket, (char *) &size, 4);
  return ret == 4;
}

int SocketFlush(Socket *socket)
{
  int ret = 0;
  if (socket->writing && socket->outbufpos)
    {
      ret = doSocketSend(socket, socket->outbuf, socket->outbufpos);
      socket->outbufpos = 0;
    }
  return ret;
}

void SocketSendStart(Socket *socket)
{
  if (!socket->multipoint)
    return;
  socket->writing = 1;
}

int SocketSendEnd(Socket *socket)
{
  int ret;
  if (!socket->multipoint)
    return 0;
  ret = SocketFlush(socket);
  socket->writing = 0;
  return ret;
}

static int doSocketRecv(Socket *socket, char *buffer, size_t size)
{
  int ret = 1;
  int total = 0;
  // We might get a SIGPIPE if we read using SSL
  void *sig = signal(SIGPIPE, SIG_IGN);
  //printf("recv %d<-%p\n", size, socket);
  while (size > total && ret > 0)
    {
      do
	{
	  ret = sockd_read(socket->bio.single, buffer + total, size - total);
	}
      while (ret < 0 &&
#ifdef WANT_SSL
	     BIO_should_retry(socket->bio.single)
#else
	     0
#endif
	     );
      if (ret > 0)
	{
	  total += ret;
	}
      else if (ret < 0)
	{
#ifdef WANT_SSL
	  SSL *ssl = NULL;
	  BIO_get_ssl(socket->bio.single, &ssl);
	  if (ssl)
	    {
	      SSL_load_error_strings();
	      ERR_print_errors_fp(stderr);
	      printf("doSocketRecv: Got %d (%s)\n", ret, 
		     ERR_error_string(SSL_get_error(ssl, ret), NULL));
	    }
	  else
	    {
	      printf("doSocketRecv: Got %d (no connection)\n", ret);
	    }
#else
	  printf("doSocketRecv: Got %d (%s)\n", ret, strerror(errno));
#endif
	}
      else
	{
	  if (DEBUG_SOCKETS)
	    printf("doSocketRecv: Got EOF\n");
	}
    }
  if (ret > 0)
    {
      totalOpsIn++;
      totalIn += total;
    }
  else if (ret < 0)
    {
      socket->err = errno;
      //printf("%s\n", strerror(errno));
    }
  else
    {
      socket->err = 0;
    }
  //printf("total: %d.\n", total);
  if (sig != SIG_ERR)
    signal(SIGPIPE, sig);
  return total;
}

int SocketRecv(Socket *socket, char *buffer, size_t size)
{
  int objsz;
  if (socket->inObject)
    {
      while (!socket->atEnd && socket->inbufpos < size)
	{
	  doSocketRecv(socket, (char *) &objsz, 4);
	  if (objsz)
	    {
	      doSocketRecv(socket, socket->inbuf + socket->inbufpos, ntohl(objsz));
	      socket->inbufpos += ntohl(objsz);
	    }
	  else
	    socket->atEnd = 1;
	}
      memcpy(buffer, socket->inbuf, size);
      memmove(socket->inbuf, socket->inbuf + size, socket->inbufpos - size);
      socket->inbufpos -= size;
      return size;
    }
  return doSocketRecv(socket, buffer, size);
}

int SocketRecvRecord(Socket *socket, char **buffer, size_t *size)
{
  char buf[4];
  if (SocketRecv(socket, buf, 4) != 4)
    return 0;
  *size = ntohl(*(unsigned long *)buf);
  if (socket->inbuflen < *size + 1)
    {
      free(socket->inbuf);
      while (socket->inbuflen < *size + 1)
	socket->inbuflen *= 2;
      socket->inbuf = malloc(socket->inbuflen);
    }
  if (SocketRecv(socket, socket->inbuf, *size) != *size)
    return 0;
  *buffer = socket->inbuf;
  (*buffer)[*size] = '\0';
  return 1;
}

int SocketRecvMulti(Socket *socket, char *buffer, size_t size, int *client)
{
  int i;
  sockd sbio;
  int ret;
  fd_set fds;
  int maxFD = -1, fd;
  struct timeval timeout;

  FD_ZERO(&fds);
  timeout.tv_sec = 0;
  timeout.tv_usec = 0;
  for (i = 0; i < ArrayLength(socket->bio.multi); i++)
    {
      fd = sockd_fd(*((sockd *)ArrayNth(socket->bio.multi, i)));
      FD_SET(fd, &fds);
      if (fd > maxFD)
	maxFD = fd;
    }
  if (maxFD < 0)
    return -1;
  select(maxFD + 1, &fds, NULL, NULL, &timeout);
  for (i = 0; i < ArrayLength(socket->bio.multi); i++)
    {
      sbio = *((sockd *)ArrayNth(socket->bio.multi, i));
      if (FD_ISSET(sockd_fd(sbio), &fds))
	{
	  ret = sockd_read(sbio, buffer, size);
	  if (ret == size)
	    {
	      *client = sockd_fd(sbio);
	      return ret;
	    }
	}
    }
  return -1;
}

NetSocketLocation *CopyNetSocketLocation(NetSocketLocation *orig)
{
  NetSocketLocation *location = malloc(sizeof(NetSocketLocation));
  memcpy(location, orig, sizeof(NetSocketLocation));
  return location;
}

NetSocketLocation *CreateNetSocketLocation(const char *hostname, unsigned port)
{
  NetSocketLocation *location;
  struct hostent *hent;
  if (hostname)
    {
      hent = gethostbyname(hostname);
      if (!hent || !hent->h_addr_list[0])
	{
	  printf("Couldn't look up %s\n", hostname);
	  return NULL;
	}
      else
       {
         location = malloc(sizeof(NetSocketLocation));
         location->ip = ((struct in_addr *) hent->h_addr_list[0])->s_addr;
       }
    }
  else
    {
      location = malloc(sizeof(NetSocketLocation));
      location->ip = 0;
    }
  location->port = htons(port);
  //printf("created location %x:%d\n", location->ip, location->port);
  return location;
}

void SocketIPMerge(uint32_t *target, uint32_t *host)
{
  // We want to replace the ip address if it is 0 (had no clue),
  // or if it is 127.*.*.* (localhost).
  if (*target == 0 || ((unsigned char *)target)[0] == 127)
    *target = *host;
}

void NetSocketLocationMerge(NetSocketLocation *target, NetSocketLocation *host)
{
  /*
  PrintLocation(target);
  printf(" from ");
  PrintLocation(host);
  */
  SocketIPMerge(&target->ip, &host->ip);
  //PrintLocation(target);
}

void SocketSelect(fd_set *fds, DArray socketP_Array, SocketFactory *factory)
{
  int len, i, maxFD = 0;
  if (socketP_Array)
    len = ArrayLength(socketP_Array);
  else
    len = 0;
  FD_ZERO(fds);
  if (factory)
    {
      FD_SET(factory->fd, fds);
      maxFD = factory->fd;
    }
  for (i = 0; i < len; i++)
    {
      Socket *sock = *((Socket **) ArrayNth(socketP_Array, i));
      int fd;
      if (sock->multipoint)
	fd = sock->ext.lsnr;
      else
	fd = sockd_fd(sock->bio.single);
      FD_SET(fd, fds);
      if (maxFD < fd)
	maxFD = fd;
    }
  //note: will block forever when sock array is empty and no factory
  select(maxFD+1, fds, NULL, NULL, NULL);
}

int SocketReadReady(Socket *sock, unsigned waitUsec)
{
  int fd;
  struct timeval timeout;
  fd_set fds;
  timeout.tv_sec = 0;
  timeout.tv_usec = waitUsec;

  fd = sockd_fd(sock->bio.single);
  FD_ZERO(&fds);
  FD_SET(fd, &fds);
  return select(fd + 1, &fds, NULL, NULL, &timeout);
}

int SocketIncluded(fd_set *fds, Socket *sock)
{
  int fd;
  if (sock->multipoint)
    fd = sock->ext.lsnr;
  else
    fd = sockd_fd(sock->bio.single);
  return FD_ISSET(fd, fds);
}

int SocketInclude(fd_set *fds, Socket *sock)
{
  int fd;
  if (sock->multipoint)
    fd = sock->ext.lsnr;
  else
    fd = sockd_fd(sock->bio.single);
  FD_SET(fd, fds);
  return fd;
}

void SocketFactoryClose(SocketFactory *factory)
{
  close(factory->fd);
  free(factory->local);
  free(factory);
}

void SocketClose(Socket *socket)
{
  void *sig;
  SocketFlush(socket);
  if (socket->parent)
    {
      socketJoin(socket->parent, socket);
    }
  else if (socket->multipoint)
    {
      int i;
      for (i = 0; i < ArrayLength(socket->bio.multi); i++)
	{
	  sockd bio = *((sockd *)ArrayNth(socket->bio.multi, i));
	  // close here
	  sig = signal(SIGPIPE, SIG_IGN);
	  sockd_free(bio);
	  if (sig != SIG_ERR)
	    signal(SIGPIPE, sig);
	}
      ArrayFree(socket->bio.multi);
      close(socket->ext.lsnr);
      // in this case, we own the local address
      free(socket->local);
    }
  else
    {
      // close here
      sig = signal(SIGPIPE, SIG_IGN);
      sockd_free(socket->bio.single);
      if (sig != SIG_ERR)
	signal(SIGPIPE, sig);
      free(socket->local);
    }
  free(socket->inbuf);
  free(socket->outbuf);
  if (socket->name)
    free(socket->name);
  if (!socket->multipoint)
    free(socket->ext.remote);
  free(socket);
}

char *SocketName(Socket *socket)
{
  if (!socket->name)
    {
      socket->name = malloc(7 + 5 + 1); /* 5 digits of fd */
      sprintf(socket->name, "socket %ld", sockd_fd(socket->bio.single));
    }
  return socket->name;
}

unsigned SocketEndpoints(Socket *socket)
{
  return ArrayLength(socket->bio.multi);
}

NetSocketLocation *SocketLocal(Socket *socket)
{
  return socket->local;
}
NetSocketLocation *SocketRemote(Socket *socket)
{
  return socket->ext.remote;
}

void PrintLocation(NetSocketLocation *location)
{
  unsigned int ip = ntohl(location->ip);
  printf("%d.%d.%d.%d:%d\n",
	 ip >> 24, (ip >> 16) & 255, (ip >> 8) & 255, ip & 255,
	 ntohs(location->port));
}
  
int SocketError(Socket *socket)
{
  return socket->err;
}

int SocketJoinMultipoint(Socket *socket, Socket *multi)
{
  socket->parent = multi;
  free(socket->local);
  socket->local = multi->local;
  return 1;
}

#if SOCK_STATS
void PrintSockStats()
{
  printf("Total IN: %d bytes/%d ops, Total OUT: %d bytes/%d ops.\n", 
	 totalIn, totalOpsIn, totalOut, totalOpsOut);
}
#else
void PrintSockStats()
{
}
#endif

char *SocketFingerprint(Socket *socket)
{
  return NULL;
}

#ifdef WANT_SSL
static int sec_init(BIO **biop, int is_server, Identity *identity)
{
  BIO *sslbio;
  SSL_CTX *ctx;

  if (!identity)
    {
      identity = FallbackIdentity();
      if (!identity)
	return -1;
    }
  ctx = identity->ctx;

  sslbio = BIO_new_ssl(ctx, !is_server);
  BIO_push(sslbio, *biop);
  BIO_set_close(sslbio, BIO_CLOSE);
  
  if (BIO_do_handshake(sslbio) <= 0)
    {
      fprintf(stderr, "Error in SSL handshake\n");
      ERR_print_errors_fp(stderr);
      return -1;
    }
  
  *biop = sslbio;
  
  return 1;
}

#endif

int SocketSecure(Socket *socket, Identity *identity, int is_server)
{
  int ret;
#ifdef WANT_SSL
  SocketFlush(socket);
  //printf("Starting encryption\n");
  ret = sec_init(&socket->bio.single, is_server, identity);
  if (ret >= 0)
    {
      //printf("Connection encrypted\n");
    }
#else
  ret = -1;
#endif
  if (ret < 0)
    {
      printf("Warning: secure connection failed\n");
    }
  return ret;
}

int SocketIsSecure(Socket *socket)
{
#ifdef WANT_SSL
  SSL *ssl = NULL;
  BIO_get_ssl(socket->bio.single, &ssl);
  return ssl != NULL;
#else
  return 0;
#endif
}

#ifdef WANT_SSL
static char *get_fingerprint(X509 *cert) 
{
  int j;
  unsigned int n;
  unsigned char md[EVP_MAX_MD_SIZE];
  char *ret, *posn;
  
  if (!X509_digest(cert,EVP_md5(),md,&n))
    {
      return NULL;
    }
  ret = malloc(n * 3);
  posn = ret;
  for (j=0; j<(int)n; j++)
    {
      sprintf(posn,"%02X%s", md[j], (j+1 == (int)n) ? "" : ":");
      posn += 3;
    }
  return ret;
}
#endif

char *SocketRemoteFingerprint(Socket *socket)
{
#ifdef WANT_SSL
  SSL *ssl = NULL;
  X509 *cert;
  char *ret;
  BIO_get_ssl(socket->bio.single, &ssl);
  if (!ssl)
    return NULL;
  cert = SSL_get_peer_certificate(ssl);
  if (!cert)
    return NULL;
  ret = get_fingerprint(cert);
  X509_free(cert);
  return ret;
#else
  return NULL;
#endif
}

int SocketNoWrite(Socket *socket)
{
  return !socket->mayWrite;
}

int SocketNoRead(Socket *socket)
{
  return !socket->mayRead;
}
