/*************************************************
 * File: unl.c
 * Birth Date: Jul 29, 2004
 *
 * Project: Enchantment IPC
 * 
 * Based on Enchantment IPC Whitepaper
 *   by Edward Keyes.
 *
 * Copyright (C) 2002, 2003, 2004 MIT Media Lab,
 *               2004 AWare Technologies, Inc.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 * 
 *************************************************/

#include <enchantment/unl.h>
#include <enchantment/types.h>

#include <enchantment/dstring.h>

#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#define MAX_NAME_LEN 80

const char *ExtractUNLServer(const char *UNL)
{
  char *delimPtr;
  char *slashPtr;
  int serverLen;

  if (!UNL) 
    return NULL;

  // get a pointer to the delimiter
  delimPtr = strchr(UNL, ':');
  slashPtr = strchr(UNL, '/');

  // no : or there's a / before the first :
  if (!delimPtr || (slashPtr && slashPtr < delimPtr))
    return NULL;

  // get length of server name
  serverLen = (int) (delimPtr - UNL);
  
  // create dstring out of server name
  return InstanceStringN(UNL, serverLen);
}

const char *ExtractUNLPath(const char *UNL)
{
  const char *delimPtr, *pathPtr = NULL, *slashPtr;
  char *temp = NULL;
  const char *ret;
  int pathLen;

  int changed = 0;

  if (!UNL)
    return NULL;

  // get a pointer to the server delimiter
  delimPtr = strchr(UNL, ':');
  slashPtr = strchr(UNL, '/');
  if (!delimPtr || (slashPtr && slashPtr < delimPtr))
    pathPtr = UNL;
  else
    pathPtr = delimPtr + 1;

  // get length of path
  pathLen = strlen(pathPtr);
  
  if (*pathPtr && *pathPtr != '/')
    {
      changed = 1;
      pathLen++;
    }

  // Remove final slashes
  while (pathLen && pathPtr[pathLen - 1] == '/')
    pathLen--;

  if (changed)
    {
      temp = malloc(pathLen);
      if (*pathPtr && *pathPtr != '/')
	{
	  *temp = '/';
	  memcpy(temp + 1, pathPtr, pathLen - 1);
	}
      else
	memcpy(temp, pathPtr, pathLen);
      pathPtr = temp;
    }

  ret = InstanceStringN(pathPtr, pathLen);

  if (temp)
    free(temp);
  
  return ret;
}

const char *ExtractUNLEndNode(const char *UNL)
{
  const char *serverDelimPtr = NULL, *lastDelimPtr;
  char buff[MAX_NAME_LEN];

  if (!UNL)
    return NULL;
  
  lastDelimPtr = strrchr(UNL, '/'); // last '/'
   
  // Get rid of any end '/'
  if (lastDelimPtr && lastDelimPtr[1] == '\0') 
    {
      bzero(buff, MAX_NAME_LEN);
      strncpy(buff, UNL, lastDelimPtr - UNL);
      UNL = buff;
      
      lastDelimPtr = strrchr(UNL, '/');
      if (lastDelimPtr && lastDelimPtr[1] == '\0') 
	return NULL; //only one delim at end is ok
    }

  serverDelimPtr = strchr(UNL, ':');

  if (!serverDelimPtr) 
    {
      serverDelimPtr = UNL-1;    /* This is an eeeevil hack */
    }
  if (!lastDelimPtr) // the root ("") or alias case (like ~jgips)
    {
      if(serverDelimPtr[1] == '\0') //*nodeName = NULL;
	{
	  char path_delim = '/';
	  return InstanceStringN(&path_delim, 1); //root
	}
      else 
	return InstanceString(serverDelimPtr + 1); //alias
    }
  else
    {
      // last node is after the last delimiter
      return InstanceString(lastDelimPtr + 1);
    }
}

const char *ExtractParentUNL(const char *UNL)
{
  char *lastDelimPtr;

  if (!UNL)
    return NULL;

  lastDelimPtr = strrchr(UNL, '/'); 
  //check for trailing delimiter
  if (lastDelimPtr && (*(lastDelimPtr + 1) == '\0')) 
    {
      *lastDelimPtr = '\0';
      lastDelimPtr = strrchr(UNL, '/');
    }

  if (!lastDelimPtr) //we were already at the root
    {
      return NULL;
    }

  return InstanceStringN(UNL, (((lastDelimPtr - UNL) + 1)));
}

const char *CreateUNL(const char *serverName, const char *pathName)
{
  char buff[MAX_NAME_LEN];
  const char *colon, *slash;
  
  if (!serverName) 
    return NULL;

  if (!pathName || (pathName[0] == '/' && pathName[1] == '\0'))
    {
      sprintf(buff, "%s:/", serverName);
    }
  else 
    {
      colon = strchr(pathName, ':');
      slash = strchr(pathName, '/');
      if (colon && (!slash || colon < slash))
	{
	  return InstanceString(pathName);
	}
      if (pathName[0] != '/')
	sprintf(buff, "%s:/%s", serverName, 
		pathName);
      else
	sprintf(buff, "%s:%s", serverName, pathName);
    }

  return InstanceString(buff);
}

const char *DescendsFrom(const char *anc, const char *des)
{
  size_t skip = strlen(anc);
  if (!strncmp(anc, des, skip) && (!skip || des[skip] == '/'))
    return InstanceStringN(des + skip, strlen(des) - skip);
  return NULL;
}

const char *AppendUNLPath(const char *baseUNL, const char *appPath)
{
  char buff[MAX_NAME_LEN], *lastDelim;
  
  if (!baseUNL)
    return NULL;

  if (!appPath) 
    return baseUNL;

  memset(buff, 0, MAX_NAME_LEN);
  sprintf(buff, "%s/", baseUNL);

  lastDelim = strrchr(buff, '/');

  // If path is preceded by delim...
  if (appPath[0] == '/') 
    {
      // ... and there is an extra delim on base now
      if (lastDelim[-1] == '/')
	sprintf(lastDelim, "%s", appPath + 1);
      // ... and there is one delim on base now
      else 
	sprintf(lastDelim + 1, "%s", appPath + 1);
    }
  // Else if path is not preceded by delim...
  else
    {
      // ... and there is now an extra delim on base
      if (lastDelim[-1] == '/')
	sprintf(lastDelim, "%s", appPath);
      // ... and there is one delim on base
      else
	sprintf(lastDelim + 1, "%s", appPath);
    }

  return InstanceString(buff);
}

const char *GetUNLHead(const char *path)
{
  char *delimPtr;
  
  if (!path)
    return NULL;
  
  if (path[0] == '/') 
    path++;

  delimPtr = strchr(path, '/');

  if (delimPtr)
    {
      return InstanceStringN(path, (int)delimPtr - (int)path);
    }
  else 
    {
      return InstanceString(path);
    }
}


const char *GetUNLTail(const char *path)
{
  char *delimPtr;

  if (!path)
    return NULL;
  
  if (path[0] == '/') 
    path++;
  
  delimPtr = strchr(path, '/');

  if (delimPtr)
    {
      return InstanceString(delimPtr);
    }
  else 
    {
      return NULL;
    }
}
