/*************************************************
 * File: whiteboard.c
 * Author: Jonathan Gips
 * Birth Date: 7/02/01
 *
 * Project: Enchantment IPC
 * 
 * Based on Enchantment IPC Whitepaper
 *   by Edward Keyes.
 * 
 * Copyright (C) 2002, 2003, 2004 MIT Media Lab
 *               2004 AWare Technologies, Inc.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 *
 *************************************************/

#include "protocol.h"
#include "introduction.h"

#include <enchantment/whiteboard.h>
#include <enchantment/dstring.h>
#include <enchantment/unl.h>
#include "ipc.h"

#include "whiteboard-protocol.h"
#include "optionset.h"

#include "socket.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <assert.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>

#include <enchantment/identity.h>

#define USE_THREADS yes

#define SHOW_PROTOCOL 0

#ifdef USE_THREADS
#include <pthread.h>
#include <semaphore.h>
#define lock(name) pthread_mutex_lock(&(name))
#define unlock(name) pthread_mutex_unlock(&(name))
#else
#define lock(name)
#define unlock(name)
#endif

#define SERVER_ERROR     1
#define CLIENT_ERROR     1

#define SERVER_ENV_VAR "ENCHANTHOST"

#define FOLLOW_SYMLINK (1 << 0)
#define DROP_AUTH (1 << 1)

#define CONTROLLED_LINK 1

#define ENCHANT_LOCAL 0x501

typedef struct
{
  const char *callType;
  const char *returnUNL;
  int result;
  NodeType *dataNode;
  OptionSet others;
} returnInfo;

typedef struct
{
  unsigned subID;
  const char *UNL;
} SubInfo;

typedef struct
{
  const char *name;
#ifdef USE_THREADS
  pthread_mutex_t requestLock;
  pthread_mutex_t ioLock;
  pthread_cond_t statusWait;
#endif
  returnInfo status;
  Socket *socket;
  DArray subs;
} ServerInfo;

/* Locking requirements
 *
 *  Only one thread may alter the connection lists at a time.
 *
 *  Only one thread may make a request to a server at a time, and no
 *  other thread may make a request to that server until the return
 *  info is provided.
 *
 *  Only one thread may read from a server at a time, including getting
 *  updates.
 *
 * Locking design
 *
 *  There is a lock for the connection lists, required for all access to them.
 *
 *  There is a lock for the update queue, required for all access.
 *
 *  There is a lock for each connection, required to send data to it.
 *
 *  There is a second lock for each connection, required to read from it.
 *
 *  You may not acquire the write lock while holding the read lock.
 *
 *  You may not hold the connection list lock while taking another lock.
 *
 *  ???? Update/other?
 */

static int ipcInit = 0;

#ifdef USE_THREADS
static pthread_mutex_t connectionLock;
static pthread_mutex_t updateLock;
static pthread_mutex_t updateQueueLock;
#endif

static DArray servArray = NULL;
static DArray sockArray = NULL;

static unsigned updateSequence = 0;
static unsigned updatesSent = 0;
static DArray updateQueue = NULL; // of UpdateInfo

static int quiet_client_ipc = 0;

static int kUNLArrayInitSize = 10;
// Global String Constants
static const char *s_return = NULL, *s_update = NULL, *s_terminal = NULL;
static const char *defaultServer = NULL;

#define DEBUG_UPDATE_QUEUE 0

typedef struct
{
  const char *publisher;
  const char *server;
  const char *path;
  NodeType *node;
  unsigned subID;
} UpdateInfo;

static void initializeData()
{
  if (ipcInit)
    return;
  s_update = InstanceString(S_UPDATE);
  s_return = InstanceString(S_RETURN);
  s_terminal = InstanceString(S_TERMINAL);
  sockArray = ArrayNew(sizeof(Socket*), 3, NULL);
  servArray = ArrayNew(sizeof(ServerInfo *), 3, NULL);
  updateQueue = ArrayNew(sizeof(UpdateInfo *), 5, NULL);
#ifdef USE_THREADS
  pthread_mutex_init(&updateQueueLock, NULL);
  pthread_mutex_init(&connectionLock, NULL);
#endif
  ipcInit = 1;
  if (!defaultServer)
    {
      defaultServer = getenv(SERVER_ENV_VAR);
      if (!defaultServer)
	defaultServer = "localhost";
    }
}

/**** Update queue management ****/

static int EnqueueUpdate(const char *server, const char *publisher, 
			 NodeType *node,
			 const char *UNL, unsigned subID)
{
  UpdateInfo *info = malloc(sizeof(UpdateInfo));
  info->server = server;
  info->path = UNL;
  info->node = node;
  info->publisher = publisher;
  info->subID = subID;
  if (DEBUG_UPDATE_QUEUE)
    printf("Enqueue %p at %s:%s\n", node, info->server, info->path);

  lock(updateQueueLock);
  updateSequence++;
  ArrayAppend(updateQueue, &info);
  unlock(updateQueueLock);
  return 1;
}

/**** Reading data from servers ****/

/* We need to get at least one message from this socket. 
 * Called with ioLock.
 */
static int doUnload(ServerInfo *server, returnInfo *sync)
{
  NodeType *rcvNode = NULL;
  Socket *socket = server->socket;
  char *buffer;
  size_t size;
  const char *unl = NULL;
  const char *publisher = NULL;
  const char *type = NULL;
  const char *command = NULL;
  int result;
  unsigned subID = 0;
  OptionSet others = NULL;

  if (sync)
    {
      if (server->status.callType)
	{
	  *sync = server->status;
	  server->status.callType = NULL;
	  return 1;
	}
    }
  else
    {
      sync = &server->status;
    }
  do 
    {
      if (!SocketRecvRecord(socket, &buffer, &size))
	{
	  fprintf(stderr, "Error reading record from server.\n");
	  exit(1);
	}
#if SHOW_PROTOCOL
      printf("<- %s ", buffer);
#endif      
      if (size)
	{
	  if (!strcmp(buffer, "node"))
	    {
	      ReadNode(socket, &rcvNode, NULL, NULL);
#if SHOW_PROTOCOL
	      PrintNode(rcvNode, NULL);
#endif
	    }
	  else
	    {
	      char *value = NULL;
	      size_t size;
	      ReadString(socket, &value, &size);
#if SHOW_PROTOCOL
	      printf("%s\n", value);
#endif
	      if (!strcmp(buffer, "unl"))
		{
		  unl = InstanceString(value);
		  free(value);
		}
	      else if (!strcmp(buffer, "publisher"))
		{
		  publisher = InstanceString(value);
		  free(value);
		}
	      else if (!strcmp(buffer, "type"))
		{
		  type = InstanceString(value);
		  free(value);
		}
	      else if (type == s_update && !strcmp(buffer, "id"))
		{
		  subID = atoi(value);
		  free(value);
		}
	      else if (!strcmp(buffer, "command"))
		{
		  command = InstanceString(value);
		  free(value);
		}
	      else if (!strcmp(buffer, "result"))
		{
		  result = strtol(value, NULL, 16);
		  free(value);
		}
	      else
		{
		  if (!others)
		    others = NewOptionSet();
		  AddSetOption(others, InstanceString(buffer), value);
		}
	    }
	}
    } 
  while (size);
#if SHOW_PROTOCOL
  printf("\n");
#endif
  // Look at received node type and do appropriate action
  if (type == s_return)
    {
      sync->dataNode = rcvNode;
      sync->returnUNL = unl;
      sync->callType = command;
      sync->result = result;
      sync->others = others;
    }
  else if (type == s_update)
    {
      EnqueueUpdate(server->name, publisher, rcvNode, unl, subID);
    }
  else if (type == s_terminal || type == NULL)
    {
      return 0;
    }	      
  else
    {
      printf("doUnload::Unknown type of node received from server (%s).\n",
	     type);
      FreeNode(rcvNode);
    }
  return 1;
}

/* Called under requestLock */
static int UnloadResponse(ServerInfo *server, returnInfo *sync)
{
  sync->callType = NULL;
  while (1)
    {
      lock(server->ioLock);
      if (!doUnload(server, sync))
	{
	  unlock(server->ioLock);
	  return 0;
	}
      unlock(server->ioLock);
      if (sync->callType)
	return 1;
    }
}

/* If blocking is true, do not return until a new update is available. */
static void UnloadUpdates(int blocking, int updateSequenceSeen)
{
  int i;
  fd_set socketSet;

  // Only one thread may unload at a time
  lock(updateLock);

  // Another thread unloaded something.
  if (updateSequenceSeen != updateSequence)
    {
      unlock(updateLock);
      return;
    }

  // XXXX need to deal with connectionLock
  // XXXX need to account for the possibility of the list changing due to
  //  another thread.
  if (blocking)
    {
      SocketSelect(&socketSet, sockArray, NULL);
    }

  for (i = 0; i < ArrayLength(sockArray); i++)
    {
      ServerInfo *server = *((ServerInfo **) ArrayNth(servArray, i));
      
      // Only look at those fds that were set
      if (blocking && !SocketIncluded(&socketSet, server->socket))
	continue;

      lock(server->ioLock);
      // Loop until sock is empty
      while (SocketReadReady(server->socket, 0))
	{
	  doUnload(server, NULL);
	}
      unlock(server->ioLock);
    }
  unlock(updateLock);
}

// NULL on local failure or bad return value. Valid on other values.
static int HandleSyncReturn(ServerInfo *server, char *syncCallName, 
			    const char *UNL, NodeType **p_dataNode,
			    OptionSet *respOpts, const char **actual_UNL_p)
{
  returnInfo info;

  /*********** Get the Synchronous Return **************/
  
  info.returnUNL = NULL;
  if (!UnloadResponse(server, &info))
    return ENCHANT_ERROR;

  //printf("<- %s %s %x\n", info.callType, info.returnUNL, info.result);
  
  if (strcmp(info.callType, syncCallName))
    {
      if (SERVER_ERROR) 
	printf("Wrong type of call returned in result: %s should be %s.\n",
	       info.callType, syncCallName);
      
      return ENCHANT_ERROR;
    }
  if (info.result == ENCHANT_ERROR)
    {
      if (CLIENT_ERROR)
	printf("Bad request returned for %s on %s.\n", syncCallName, UNL);
      return ENCHANT_ERROR;
    }
  
  if (p_dataNode)
    *p_dataNode = info.dataNode;
  else
    FreeNode(info.dataNode);

  *respOpts = info.others;

  *actual_UNL_p = info.returnUNL;
  return info.result;
}

/* Called with requestLock */
static int SendRequest(ServerInfo *server, char *command, NodeType *request, 
		       const char *unl, WhiteboardInfo *info, 
		       const char **options)
{
  Socket *socket = server->socket;

  lock(server->ioLock);
  //printf("-> %s %s\n", command, unl);
  SocketSendRecord(socket, "command", 7);
  WriteString(socket, command, strlen(command));
#if SHOW_PROTOCOL
  printf("-> command %s\n", command);
#endif
  if (unl)
    {
      SocketSendRecord(socket, "unl", 3);
      WriteString(socket, unl, strlen(unl));
#if SHOW_PROTOCOL
      printf("-> unl %s\n", unl);
#endif
    }
  if (options)
    {
      int i;
      for (i = 0; options[i * 2]; i++)
	{
	  SocketSendRecord(socket, options[i * 2], strlen(options[i * 2]));
	  WriteString(socket, options[i * 2 + 1], strlen(options[i * 2 + 1]));
#if SHOW_PROTOCOL
	  printf("-> %s %s\n", options[i * 2], options[i * 2 + 1]);
#endif
	}
    }
  if (info->commonFlags & ENCHANT_FOLLOW_LINKS)
    {
      SocketSendRecord(socket, "links", 12);
#if SHOW_PROTOCOL
      printf("-> links ");
#endif
      if (info->commonFlags == ENCHANT_FOLLOW_LINKS)
	{
	  WriteString(socket, "f", 1);
#if SHOW_PROTOCOL
	  printf("f\n");
#endif
	}
      else
	{
	  WriteString(socket, "c", 1);
#if SHOW_PROTOCOL
	  printf("c\n");
#endif
	}
    }
  if (info->commonFlags & ENCHANT_DROP_AUTH)
    {
      SocketSendRecord(socket, K_DROP_AUTH, strlen(K_DROP_AUTH));
      WriteString(socket, "t", 1);
#if SHOW_PROTOCOL
      printf("-> " K_DROP_AUTH " t\n");
#endif
    }
  if (request)
    {
      SocketSendRecord(socket, "node", 4);
      WriteNode(socket, request, NULL, NULL);
#if SHOW_PROTOCOL
      printf("-> node ");
      PrintNode(request, NULL);
#endif
    }
  SocketSendRecord(socket, NULL, 0);
#if SHOW_PROTOCOL
  printf("->\n");
#endif
  SocketFlush(socket);
  unlock(server->ioLock);
  return 1;
}

/* Runs under connectionLock */
static void DisconnectFromServer(int i)
{
  ServerInfo *server = *((ServerInfo **) ArrayNth(servArray, i));
  SocketClose(server->socket);
  if (server->subs)
    ArrayFree(server->subs);
  free(server);
  ArrayDeleteAt(servArray, i);
  ArrayDeleteAt(sockArray, i);
}

void SetDefaultServer(const char *server)
{
  defaultServer = InstanceString(server);
}

void ShutDownClientIPC()
{
  // If we haven't initialized, we don't need to do anything.
  if (!ipcInit) 
    return;
  
  // Close down the socket connections and free memory
  // We leave this locked, since the lock will be reinitialized if needed.
  lock(connectionLock);
  if (servArray)
    {
      while (ArrayLength(servArray) > 0)
	{
	  DisconnectFromServer(0);
	}
      
      ArrayFree(servArray); // Deallocate the socket array
      ArrayFree(sockArray);
      servArray = NULL;
      sockArray = NULL;
    }
  
  // XXXX Free the update queue and its nodes

#if SOCK_STATS
  if (!quiet_client_ipc)
    PrintSockStats();
#endif 
  ipcInit = 0;
}

const char *NicknameUNL(const char *nickname)
{
  const char *temp = strchr(nickname, '@');
  const char *ret;
  const char *branchName = InstanceStringN(nickname, temp - nickname);
  ret = CreateUNL(temp + 1, "/servers");
  ret = AppendUNLPath(ret, branchName);
  return ret;
}

const char *GetServerName(const char *UNL)
{
  const char *ret = ExtractUNLServer(UNL);
  if (!ret)
    ret = defaultServer;
  return ret;
}

/**** Getting sockets ****/

static int ServerLocationByName(NetSocketLocation **locPtr, 
				const char *serverName)
{
  int port = DEFAULT_PORT;
  const char *hostname;
  const char *temp;
  NetSocketLocation *target = NULL;
  
  if ((temp = strchr(serverName, '@')))
    {
      port = atoi(serverName);
      if (port)
	hostname = temp + 1;
      else
	{
	  const char *serverUNL = NicknameUNL(serverName);
	  NodeType *serverNode = GetNode(serverUNL, NULL);
	  if (serverNode && (serverNode->type == T_SERVER))
	    {
	      target = serverNode->data.netsock_val;
	    }
	  else
	    {
	      printf("Didn't like server name '%s'\n", serverName);
	      return 0;
	    }
	}
    }
  else
    hostname = serverName;
  if (!target)
    {
      target = CreateNetSocketLocation(hostname, port);
      if (!target)
	return 0;
    }
  *locPtr = target;
  return 1;
}

static int GetSeparateSocket(Socket **sockPtr, const char *serverName)
{
  NetSocketLocation *target = NULL;
  int i;

  serverName = InstanceString(serverName);

  lock(connectionLock);
  for (i = 0; i < ArrayLength(servArray); i++) 
    {
      ServerInfo *server = *((ServerInfo **) ArrayNth(servArray, i));
      if (server->name == serverName)
	{
	  target = SocketRemote(server->socket);
	}
    }
  unlock(connectionLock);

  if (!target)
    {
      if (!ServerLocationByName(&target, serverName))
	return 0;
    }
  *sockPtr = SocketConnect(target);
  if (0) // XXXX
    free(target);

  return (!!*sockPtr);
}

int GetSignalProxySocket(Socket **sockPtr, const char *UNL)
{
  Socket *sock;
  const char *server;

  if (!ipcInit)
    initializeData();

  server = GetServerName(UNL);
  if (!GetSeparateSocket(&sock, server))
    return 0;
  *sockPtr = sock;
  return 1;
}

static int negotiateClient(Socket *socket)
{
  IntroMessage message;
  int ret;
  int did_type = 0;
  int done;
  int use_ssl = 0;
  int using_ssl = 0;
  int version = PROTOCOL_VERSION_NUMBER;
  int got_version = 0;
  do
    {
      done = 1;
      if (!did_type)
	{
	  message.format = &protocolInteraction;
	  message.value.string_val = PROTOCOL_TYPE_WHITEBOARD;
	  IntroSend(socket, &message);
	  message.format = &protocolVersion;
	  message.value.int_val = version;
	  IntroSend(socket, &message);
#ifdef WANT_SSL
	  message.format = &protocolEncrypt;
	  message.value.string_val = PROTOCOL_ENCRYPT_SSL;
	  IntroSend(socket, &message);
#endif
	  did_type = 1;
	  done = 0;
	}
      
      IntroDone(socket);

      if (!done)
	{
	  done = 1;

	  do
	    {
	      ret = IntroRecv(socket, serverCommonProperties, &message);
	      if (ret == -1)
		;
	      else if (message.format == &protocolEncrypt)
		{
#ifdef WANT_SSL
		  // In this case, we asked for it previously
		  use_ssl = 1;
#endif
		}
	      else if (message.format == &protocolVersion)
		{
		  if ((message.value.int_val & 0xFFFF00) !=
		      (version & 0xFFFF00))
		    {
		      fprintf(stderr, "protocol version mismatch: "
			      "%x remote %x local\n",
			      message.value.int_val, version);
		      IntroFree(&message);
		      return 0;
		    }
		  else
		    {
		      got_version = 1;
		      if (version > message.value.int_val)
			version = message.value.int_val;
		    }
		}
	      if (ret)
		{
		  done = 0;
		  IntroFree(&message);
		}
	    }
	  while (ret > 0);
	}
      if (use_ssl && !using_ssl)
	{
	  SocketSecure(socket, NULL, 0);
	  using_ssl = 1;
	}
    }
  while (!done);
  return got_version;
}

/* Must be called under the connection lock. */
/* Currently, we get new connections for different names for the same
 * server. */
static ServerInfo *LookupServer(const char *serverName,
				NetSocketLocation *target)
{
  int i;
  ServerInfo *server;
  for (i = ArrayLength(servArray) - 1; i >= 0; i--)
    {
      server = *((ServerInfo **)ArrayNth(servArray, i));
      if ((server)->name == serverName)
	return server;
    }
  return NULL;
}

/*
 * GetServerInfo makes the connection if necessary; FindServerInfo
 * only reports connects already made.
 */
static ServerInfo *GetServerInfo(const char *serverName)
{
  Socket *sock;
  ServerInfo *server;
  NetSocketLocation *target = NULL;
    
  if (!serverName) 
    return NULL;

  serverName = InstanceString(serverName);

  // Check if we already have a connection
  lock(connectionLock);
  server = LookupServer(serverName, NULL);
  unlock(connectionLock);

  if (server)
    return server;
  
  // Find the location; must not hold the connection lock, because it can
  // recurse.
  if (!ServerLocationByName(&target, serverName))
    return NULL;

  // Recheck if we already have a connection, because we may have been bypassed
  lock(connectionLock);
  server = LookupServer(serverName, NULL);
  if (server)
    {
      unlock(connectionLock);
      return server;
    }

  //printf("Must connect\n");
  sock = SocketConnect(target);
  if (!sock)
    {
      unlock(connectionLock);
      printf("Failed to connect to %s at ", serverName);
      PrintLocation(target);
      printf("\n");
      free(target);
      return NULL;
    }

  // Initialize the connection
  if (!negotiateClient(sock))
    {
      SocketClose(sock);
      unlock(connectionLock);
      printf("Failed to negotiate with %s at ", serverName);
      PrintLocation(target);
      printf("\n");
      free(target);
      return NULL;
    }

  server = malloc(sizeof(ServerInfo));
  server->name = serverName;
  server->socket = sock;
  server->status.callType = NULL;
#ifdef USE_THREADS
  pthread_mutex_init(&server->requestLock, NULL);
  pthread_mutex_init(&server->ioLock, NULL);
#endif
  server->subs = NULL;

  //printf("Connected\n");

  ArrayAppend(servArray, &server);
  ArrayAppend(sockArray, &sock);
  unlock(connectionLock);
  free(target);
  return server;
}

static ServerInfo *FindServerInfo(const char *serverName)
{
  ServerInfo *server;

  if (!serverName) 
    return NULL;

  serverName = InstanceString(serverName);

  // Check if we already have a connection
  lock(connectionLock);
  server = LookupServer(serverName, NULL);
  unlock(connectionLock);

  return server;
}

static int FindSubscriptionID(const char *UNL)
{
  ServerInfo *server = FindServerInfo(GetServerName(UNL));
  if (server && server->subs)
    {
      int i;
      const char *path = ExtractUNLPath(UNL);
      for (i = 0; i < ArrayLength(server->subs); i++)
	{
	  SubInfo *info = ArrayNth(server->subs, i);
	  if (info->UNL == path)
	    return info->subID;
	}
    }
  return 0;
}

int AttemptEncryption()
{
#ifdef USE_SSL
  return 1;
#else
  return 0;
#endif
}

static int GetLibraryInfo(const char *path, NodeType **p_retNode)
{
  int i;
  NodeType *ret = NULL;
  NodeType *sub = NULL;
  const char *headName;
  headName = GetUNLHead(path);

  //printf("Library asked for '%s'\n", headName);
  if (!strcmp(headName, ""))
    {
      ret = CreateBranchNode("local", T_BRANCH);
      if (GetLibraryInfo("/Info", &sub))
	InsertNode(ret, sub);
      if (GetLibraryInfo("/Servers", &sub))
	InsertNode(ret, sub);
    }
  else if (!strcmp(headName, "Servers"))
    {
      ret = CreateBranchNode("Servers", T_BRANCH);
      lock(connectionLock);
      for (i = 0; i < ArrayLength(servArray); i++)
	{
	  ServerInfo *info = *((ServerInfo **) ArrayNth(servArray, i));
	  NetSocketLocation *remote, *copy;
	  NodeType *server = 
	    CreateBranchNode(info->name, T_BRANCH);
	  remote = SocketRemote(info->socket);
	  copy = CopyNetSocketLocation(remote);
	  sub = CreateAddressNode("Address", T_NETSOCK, copy);
	  InsertNode(server, sub);
				  
	  //printf("Adding %d: %s\n", i, sub->name);
	  InsertNode(ret, server);
	}
      unlock(connectionLock);
    }
  *p_retNode = ret;
  return !!ret;
}

static DArray InitUNLArray(const char *UNL)
{
  DArray newArray = NULL;

  newArray = ArrayNew(sizeof(char *), kUNLArrayInitSize, NULL);
  if (newArray)
    {
      ArrayAppend(newArray, &UNL);
    }

  return newArray;
}

static void ApproveNewUNL(DArray visitedUNLs, const char *newUNL, int *CYCLIC)
{
  if (ArraySearch(visitedUNLs, &newUNL, ArrayEq, 0, 0) == NOT_FOUND)
    {
      ArrayAppend(visitedUNLs, &newUNL);
      *CYCLIC = 0;
    }
  else 
    {
      *CYCLIC = 1;
    }
}

static int FollowSymlinks(const char *UNL, char *command, NodeType *sendNode,
			  NodeType **p_retNode, WhiteboardInfo *info,
			  int checkOwned, const char **options, 
			  OptionSet *p_respOpts)
{
  int CYCLIC;
  const char *server;
  ServerInfo *serverInfo;
  DArray visitedUNLs = NULL;
  int result;
  NodeType *retNode;
  OptionSet localRespOpts = NULL;
  const char *retUNL;

  if (!p_respOpts)
    p_respOpts = &localRespOpts;
  else
    *p_respOpts = NULL;

  do
    {
      server = GetServerName(UNL);
      if (!*server)
	{
	  // Talking to the library...
	  //printf("Local\n");
	  info->actualUNL = UNL;
	  result = ENCHANT_LOCAL;
	  break;
	}
      serverInfo = GetServerInfo(server);
      if (!serverInfo)
	{
	  printf("Couldn't prepare\n");
	  result = ENCHANT_ERROR;
	  break;
	}
      lock(serverInfo->requestLock);
      if (!SendRequest(serverInfo, command, sendNode, ExtractUNLPath(UNL),
		       info, options))
	{
	  printf("Couldn't send\n");
	  result = ENCHANT_ERROR;
	  break;
	}
      result = HandleSyncReturn(serverInfo, command, UNL, 
				&retNode, p_respOpts, &retUNL);
      unlock(serverInfo->requestLock);
      if (!(result & ~0xff))
	{
	  if (info)
	    info->actualUNL = retUNL;
	  break;
	}
      else if (result == ENCHANT_SYMLINK)
	{
	  if (checkOwned)
	    {
	      if (!GetSetOption(*p_respOpts, InstanceString("controlled")))
		{
		  info->commonFlags |= ENCHANT_DROP_AUTH;
		}
	    }
	  UNL = InstanceString(retUNL);
	  
	  if (!visitedUNLs)
	    visitedUNLs = InitUNLArray(UNL);
	  else
	    {
	      ApproveNewUNL(visitedUNLs, UNL, &CYCLIC);
	      if (CYCLIC)
		{
		  printf("Cycle formed by symlinks\n");
		  result = ENCHANT_ERROR;
		  break;
		}
	    }
	}
      else
	{
	  break;
	}
    }
  while (1);
  if (visitedUNLs)
    ArrayFree(visitedUNLs);
  if (p_retNode)
    *p_retNode = retNode;
  if (info)
    {
      info->errorCode = result;
    }
  if (localRespOpts)
    {
      FreeOptionSet(localRespOpts);
    }
  return !(result & ~0xff);
}

/*************************** Public functions ************************/

int SecureUNL(const char *UNL)
{
  // NOTE: does not follow symlinks to determine whether the eventual
  // connection is secure.

  ServerInfo *serverInfo;
  const char *server;

  if (!ipcInit)
    initializeData();

  server = GetServerName(UNL);
  if (!*server)
    return 1; // unsuprisingly, we have a secure "connection" to the library
  serverInfo = GetServerInfo(server);
  if (!serverInfo)
    return 0; // No connection at all...
  
  return SocketIsSecure(serverInfo->socket);
}

NodeType *GetNode(const char *UNL, GetNodeInfo *info)
{
  int result;
  GetNodeInfo localInfo;
  NodeType *ret;

  if (!ipcInit)
    initializeData();
  if (!UNL || !info)
    {
      if (!info)
	info = &localInfo;
      info->common.commonFlags = ENCHANT_FOLLOW_LINKS;
    }
  if (!UNL)
    {
      return NULL;
    }
  
  result = FollowSymlinks(UNL, C_GET, NULL, &ret, &info->common, 
			  0, NULL, NULL);
  if (!result)
    {
      if (info->common.errorCode == ENCHANT_LOCAL)
	{
	  GetLibraryInfo(ExtractUNLPath(UNL), &ret);
	  return ret;
	}
      return NULL;
    }
  return ret;
}

int LockUNL(const char *UNL, LockUNLInfo *info)
{
  char *callType;
  LockUNLInfo lockInfo;
  int result;
  const char **options = NULL;

  if (!ipcInit)
    initializeData();
  if (!UNL || !info)
    {
      if (!info)
	info = &lockInfo;
      info->common.operationFlags = ENCHANT_LOCK_NORMAL;
      info->common.commonFlags = ENCHANT_CROSS_LINKS;
    }
  if (!UNL)
    return 0;
  
  callType = C_LOCK;

  // Package up the call in a node to send

  if (info->common.operationFlags & 3)
    {
      int i = 0;
      options = malloc(sizeof(char *) * 
		       (3 + (2 * !!(info->common.operationFlags & 4))));
      options[i] = K_LOCK_TYPE;
      switch (info->common.operationFlags & 3)
	{
	case 1:
	  options[i + 1] = "n";
	  break;
	case 2:
	  options[i + 1] = "r";
	  break;
	case 3:
	  options[i + 1] = "R";
	  break;
	}
      i += 2;
      if (info->common.operationFlags & 4)
	{
	  options[i] = K_LOCK_UPDATE;
	  options[i + 1] = "t";
	  i += 2;
	}
      options[i] = NULL;
    }

  result = FollowSymlinks(UNL, callType, NULL, NULL, &info->common, 
			  0, options, NULL);
  if (options)
    free(options);
  return result;
}

int GetUNLLock(const char *UNL, GetUNLLockInfo *info)
{
  int result;
  GetUNLLockInfo localInfo;
  NodeType *attrNode;
  OptionSet options;
  const char *option;

  if (!ipcInit)
    initializeData();
  if (!UNL || !info)
    {
      if (!info)
	info = &localInfo;
      info->common.commonFlags = ENCHANT_CROSS_LINKS;
    }
  if (!UNL)
    return 0;

  result = FollowSymlinks(UNL, C_GETLOCK, NULL, &attrNode, &info->common, 
			  0, NULL, &options);
  if (!result)
    return 0;
  info->lockOwner = 
    InstanceString(GetSetOption(options, InstanceString(K_GETLOCK_OWNER)));
  option = GetSetOption(options, InstanceString(K_LOCK_TYPE));
  switch (option ? option[0] : '\0')
    {
    case KV_LOCK_TYPE_REMOVE:
      info->lockType = ENCHANT_LOCK_REMOVE;
      break;
    case KV_LOCK_TYPE_RECURSIVE:
      info->lockType = ENCHANT_LOCK_RECURSIVE;
      break;
    case KV_LOCK_TYPE_NORMAL:
      info->lockType = ENCHANT_LOCK_NORMAL;
      break;
    default:
      info->lockType = ENCHANT_LOCK_NONE;
    }
  if (options)
    ArrayFree(options);
  return 1;
}

int PublishNode(const char *UNL, NodeType *node, PublishNodeInfo *info)
{
  int result;
  PublishNodeInfo localInfo;
  unsigned len;
  const char **options = NULL;
  
  if (!ipcInit)
    initializeData();
  if (!UNL || !info)
    {
      if (!info)
	info = &localInfo;
      info->common.commonFlags = ENCHANT_CROSS_LINKS;
      info->common.operationFlags = 
	ENCHANT_PUBLISH_REPLACE | ENCHANT_PUBLISH_SOFT;
    }
  if (!UNL)
    return 0;

  len = !!(info->common.operationFlags & ENCHANT_PUBLISH_HARD) +
    !!(info->common.operationFlags & 3);
  if (len)
    {
      options = malloc(sizeof(char *) * (len * 2 + 1));
      len = 0;
      if (info->common.operationFlags & ENCHANT_PUBLISH_HARD)
	{
	  options[len] = K_PUBLISH_HARD;
	  options[len + 1] = "t";
	  len += 2;
	}
      switch (info->common.operationFlags & 3)
	{
	case ENCHANT_PUBLISH_MERGE:
	  options[len] = K_PUBLISH_OPERATION;
	  options[len + 1] = KV_PUBLISH_OPERATION_MERGE;
	  len += 2;
	  break;
	case ENCHANT_PUBLISH_ADDONLY:
	  options[len] = K_PUBLISH_OPERATION;
	  options[len + 1] = KV_PUBLISH_OPERATION_ADDONLY;
	  len += 2;
	  break;
	}
      options[len] = NULL;
    }
  
  result = FollowSymlinks(UNL, C_PUBLISH, node, NULL, &info->common, 
			  1, options, NULL);
  if (options)
    free(options);
  return result;
}
 
int SubscribeUNL(const char *UNL, SubscribeUNLInfo *info)
{
  NodeType *dataNode = NULL;
  SubscribeUNLInfo localInfo;
  const char **others = NULL;
  OptionSet options;

  if (!ipcInit)
    initializeData();
  if (!UNL || !info)
    {
      if (!info)
	info = &localInfo;
      info->common.commonFlags = ENCHANT_FOLLOW_LINKS;
      info->common.operationFlags =
	ENCHANT_SUBSCRIBE_OTHERS | ENCHANT_SUBSCRIBE_NEW;
    }
  if (!UNL) return 0;

  if (info->common.operationFlags & ENCHANT_SUBSCRIBE_SELF)
    {
      others = malloc(3 * sizeof(char *));
      others[0] = K_SUBSCRIBE_SELF;
      others[1] = "t";
      others[2] = NULL;
    }

  if (FollowSymlinks(UNL, C_SUBSCRIBE, NULL, &dataNode, &info->common, 
		     0, others, &options))
    {
      const char *option = GetSetOption(options, InstanceString("id"));
      unsigned subID = option ? atoi(option) : 0;
      ServerInfo *server = FindServerInfo(GetServerName(UNL));
      SubInfo subInfo;
      if (dataNode)
	{
	  if (info->common.operationFlags & ENCHANT_SUBSCRIBE_INITIAL)
	    {
	      // Act like we got it in an update
	      EnqueueUpdate(GetServerName(UNL), NULL, dataNode, 
			    ExtractUNLPath(UNL), subID);
	    }
	  else
	    FreeNode(dataNode);
	}
      if (server->subs == NULL)
	{
	  server->subs = ArrayNew(sizeof(SubInfo), 1, NULL);
	}
      subInfo.subID = subID;
      subInfo.UNL = ExtractUNLPath(UNL);
      ArrayAppend(server->subs, &subInfo);
      // Stuff for putting the subscription in the subscription list.
      if (options)
	ArrayFree(options);
      if (others)
	free(others);
      return 1;
    }

  if (others)
    free(others);
  if (options)
    ArrayFree(options);
  return 0;
}

int UnsubscribeUNL(const char *UNL, UnsubscribeUNLInfo *info)
{
  UnsubscribeUNLInfo localInfo;
  int ret;
  OptionSet options;
  const char **others = NULL;
  ServerInfo *server;
  int i;
  unsigned subID;

  if (!ipcInit)
    initializeData();
  if (!UNL || !info)
    {
      if (!info)
	info = &localInfo;
      info->common.commonFlags = ENCHANT_FOLLOW_LINKS;
    }
  if (!UNL) return 0;

  server = FindServerInfo(GetServerName(UNL));
  subID = FindSubscriptionID(UNL);

  if (subID)
    {
      others = malloc(sizeof(char *) * 3);
      others[0] = "id";
      others[1] = InstanceNumber(subID);
      others[2] = NULL;
    }

  ret = FollowSymlinks(UNL, // We at least need to specify the server...
		       C_UNSUBSCRIBE, NULL, NULL, &info->common, 
		       0, others, &options);
  if (ret)
    {
      // Note that we remove the subscription the server tells us to,
      // regardless of what we determined above.
      const char *option = GetSetOption(options, InstanceString("id"));
      subID = option ? atoi(option) : 0;
      for (i = 0; i < ArrayLength(server->subs); i++)
	{
	  SubInfo *subInfo = ArrayNth(server->subs, i);
	  if (subInfo->subID == subID)
	    {
	      ArrayDeleteAt(server->subs, i);
	      break;
	    }
	}
    }
  if (options)
    ArrayFree(options);
  if (others)
    free(others);

  return ret;
}

NodeType *GetUNLMetadata(const char *UNL, char *metapath, 
			 GetUNLMetadataInfo *info)
{
  int result;
  GetUNLMetadataInfo localInfo;
  NodeType *ret;
  const char **headers;

  if (!ipcInit)
    initializeData();
  if (!UNL || !info)
    {
      if (!info)
	info = &localInfo;
      info->common.commonFlags = ENCHANT_IGNORE_LINKS;
    }
  if (!UNL && !metapath)
    {
      return NULL;
    }

  headers = malloc(sizeof(char *) * 3);
  headers[0] = "metapath";
  headers[1] = metapath;
  headers[2] = NULL;
  
  result = FollowSymlinks(UNL, C_GET_META, NULL, &ret, &info->common, 
			  0, headers, NULL);
  free(headers);
  return result ? ret : NULL;
}

int SetUNLMetadata(const char *UNL, const char *metapath, NodeType *meta,
		   SetUNLMetadataInfo *info)
{
  SetUNLMetadataInfo localInfo;
  const char **headers;
  
  if (!ipcInit)
    initializeData();
  if (!UNL || !info)
    {
      if (!info)
	info = &localInfo;
      info->common.commonFlags = ENCHANT_IGNORE_LINKS;
      info->common.operationFlags = 
	ENCHANT_PUBLISH_REPLACE | ENCHANT_PUBLISH_SOFT;
    }
  if (!UNL)
    return 0;
  
  headers = malloc(sizeof(char *) * 3);
  headers[0] = "metapath";
  headers[1] = metapath;
  headers[2] = NULL;
  
  return FollowSymlinks(UNL, C_SET_META, meta, NULL, &info->common, 
			1, headers, NULL);
}

/* The core of GetUpdate and WaitForUpdate. */
static void scanForUpdate(GetUpdateInfo *getInfo, const char *target, 
			  int blocking)
{
  UpdateInfo *info;
  UpdateInfo *result = NULL;
  int found = -1;
  int idx = 0;
  int updatesSentSeen;
  int updateSequenceSeen;
  int id;

  if (!ipcInit)
    initializeData();
  
  if (target)
    {
      id = FindSubscriptionID(target);
      if (!id)
	{
	  printf("No subscription found to %s\n", target);
	}
    }
  else
    {
      id = 0;
    }

  getInfo->UNL = NULL;
  getInfo->node = NULL;
  getInfo->publisher = NULL;

  if (!blocking)
    UnloadUpdates(0, updateSequence); // need a single check
  
  // Wait until there is something in the queue
  do
    {
      lock(updateQueueLock);
      updatesSentSeen = updatesSent;
      updateSequenceSeen = updateSequence;
      while (ArrayLength(updateQueue) > idx)
	{
	  info = *((UpdateInfo **)ArrayNth(updateQueue, idx));
	  if (id ? info->subID == id : !!info->path)
	    {
	      found = idx;
	      result = info;
	      break;
	    }
	  else
	    {
	      //printf("Skip %s %s\n", info->server, info->path);
	    }
	  idx++;
	}
      unlock(updateQueueLock);

      if (found >= 0 || !blocking)
	break;

      UnloadUpdates(1, updateSequenceSeen);

      // Move back to account for delivered updates Note that, if
      // something gets added to the end and delivered to another
      // thread, we'll back up inappropriately. We could fall off the
      // start, or we could look at a skipped update again; both are
      // fine.
      idx -= (updatesSent - updatesSentSeen);
      // Catch falling off the start
      if (idx < 0)
	idx = 0;
    }
  while (1);

  if (result)
    {      
      // Delete the pointer to this node and move indexes.
      ArrayDeleteAt(updateQueue, found);
      updatesSent++;
      getInfo->node = result->node;
      // XXXX use the same shortcuts that target uses
      getInfo->UNL = CreateUNL(result->server, result->path);
      getInfo->publisher = result->publisher;
      free(result);
    }
}

void GetUpdate(GetUpdateInfo *info, const char *UNL)
{
  scanForUpdate(info, UNL, 0);
}

NodeType *WaitForUpdate(GetUpdateInfo *info, const char *UNL)
{
  GetUpdateInfo localInfo;
  if (!info)
    info = &localInfo;
  scanForUpdate(info, UNL, 1);
  return info->node;
}

int SetUNLPermissions(const char *UNL, char operation, char **auth)
{
  NodeType *authNode = NULL;
  char *metapath = NULL;
  int i;
  if (auth)
    {
      NodeType *tokenNode;
      authNode = CreateBranchNode("root", T_BRANCH);
      for (i = 0; auth[i]; i++)
	{
	  tokenNode = CreateStringNode(InstanceNumber(i), T_STRING,
				       auth[i]);
	  InsertNode(authNode, tokenNode);
	}
    }
  switch (operation)
    {
    case 'r':
      metapath = "/Permissions/Read";
      break;
    case 'w':
      metapath = "/Permissions/Write";
      break;
    case 'l':
      metapath = "/Permissions/Lock";
      break;
    case 'a':
      metapath = "/Permissions/Admin";
      break;
    }
  if (metapath)
    return SetUNLMetadata(UNL, metapath, authNode, NULL);
  return 0;
}

int LinkUNL(const char *UNL, const char *target)
{
  NodeType *node =
    target ? CreateStringNode("", T_UNL, (char *) target) : NULL;
  int ret = SetUNLMetadata(UNL, "/symlink", node, NULL);
  FreeNode(node);
  return ret;
}

void WhiteboardVerbosity(int level)
{
  if (level == 0)
    quiet_client_ipc = 1;
  else
    quiet_client_ipc = 0;
}
