/*************************************************
 * File: server.c
 * Birth Date: 7/3/01
 *
 * Project: Enchantment IPC
 * 
 * Based on Enchantment IPC Whitepaper
 *   by Edward Keyes.
 * 
 * Copyright (C) 2002, 2003, 2004 MIT Media Lab
 *               2004 AWare Technologies, Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *************************************************/

/* This file contains the main server loop, identification of types
 * of clients, and server initialization. 
 */

#include <enchantment/types.h>
#include "whiteboard-protocol.h"
#include <enchantment/node.h>
#include <enchantment/unl.h>
#include <enchantment/darray.h>
#include <enchantment/dstring.h>
#include "socket.h"
#include <enchantment/signals.h>
#include <enchantment/whiteboard.h>

#include <enchantment/identity.h>

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <sys/time.h>
#include <sys/types.h>
#include <signal.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <dirent.h>
#include <stdarg.h>
#include <pwd.h>

#include "clients.h"
#include "signal-proxy.h"
#include "tree-manip.h"
#include "symbols.h"
#include "logging.h"
#include "journal.h"
#include "request.h"

#include "protocol.h"
#include "introduction.h"
#include "ipc.h"

#ifdef USE_THREADS

#include <pthread.h>
#include <semaphore.h>

#endif

/**** Globals ****/
ClientIDType *serverID;

static SocketFactory *g_servSock = NULL;
static DArray g_sockets = NULL;

DArray g_clients = NULL;

static int quiet;

/*** Structures ***/

typedef struct
{
  Signal *signalCon;
  Signal *signalGen;
} SignalProxyInfo;

/**** Function Prototypes ****/

int       InitGlobalData(char *filename);
void      SetupSignals();
int       ReadIdentity(char *filename);
void      PublishNicknames(NetSocketLocation *location, DArray nicknames);

void      StartAcceptingClients();
void      AcceptNewClients();
int       HandleInitialContact(Socket *socket);

void      ShutDownServer(int param);

void      ServiceInput();

typedef struct
{
  NodeType *clientNode;
  NodeType *rcvNode;
} RequestType;

#ifdef USE_THREADS
pthread_mutex_t clientLock;
#endif

/*** Filenames ***/

#define DEFAULT_LOG_FILE "enchant-server-log"
#define DEFAULT_CERT_FILE "server-cert.pem"
#define DEFAULT_KEY_FILE "server-key.pem"

/* In state directory... */
#define WHITEBOARD_FILE "whiteboard"
#define DOCUMENT_DIR "documents"
#define CERT_FILE "cert.pem"
#define KEY_FILE "key.pem"

int main(int argc, char* argv[])
{
  int daemon_flag = 0;
  int opt;
  int help = 0;
  int serv_port = DEFAULT_PORT;

  char *filename = NULL;
  DArray nicknames = NULL;
  NetSocketLocation *servLocation;

  assert(sizeof(time_t) == sizeof(long));

  do
    {
      opt = getopt(argc, argv, "df:p:hn:");
      if (opt == 'd')
	{
	  daemon_flag = 1;
	  quiet = 1;
	}
      if (opt == 'f')
	filename = optarg;
      if (opt == 'p')
	serv_port = atoi(optarg);
      if (opt == 'h')
	help = 1;
      if (opt == 'n')
	{
	  if (!nicknames)
	    nicknames = ArrayNew(sizeof(char *), 0, NULL);
	  ArrayAppend(nicknames, &optarg);
	}
    }
  while (opt > 0);

  if (help)
    {
      printf("Usage: %s\n", argv[0]);
      printf(" -f <state>: enable state, using <state> for storage\n");
      printf(" -p <port>: listen on <port> instead of %d\n", DEFAULT_PORT);
      printf(" -n <nickname>: publish address to another server\n");
      printf(" -d: become daemon\n");
      exit(0);
    }

  g_servSock = SocketListen(serv_port, 0);
  if (!g_servSock)
    {
      if (SERVER_ERROR)
	printf("Main::Listening socket creation failed. Exiting.\n");
      exit(-1);
    }

  if (!InitGlobalData(filename))
    {
      if (SERVER_ERROR)
	printf("Main::Failed to initialze global data. Exiting.\n");
      exit(-1);
    }

  if (daemon_flag)
    {
      NodePathReturnType npReturn;
      int pid;
      int fd;
      pid = fork();
      switch (pid) 
	{
	case -1:
	  printf("daemon() failed\n");
	  exit(-1);
	case 0:
	  break;
	default:
	  printf("%d\n", pid);
	  exit(0);
	}
      if (setsid() == -1)
	{
	  printf("daemon() failed\n");
	  exit(-1);
	}
      chdir("/");
      fd = open("/dev/null", O_RDWR, 0);
      if (fd == -1)
	{
	  printf("daemon() failed\n");
	  exit(-1);
	}
      dup2(fd, 0);
      dup2(fd, 1);
      dup2(fd, 2);
      if (fd > 2)
	close(fd);
      // Set the pid in the server's info to the new pid
      npReturn.UNL = "/clients/Server/pid";
      npReturn.getSubscriptions = 0;
      GetNodePathInfo(serverID, &npReturn, 0);
      npReturn.existNode->data.int_val = getpid();
    }
  else
    {
      printf("ready\n");
    }

  SetupSignals();  

  if (nicknames)
    {
      servLocation = CreateNetSocketLocation(NULL, serv_port);
      PublishNicknames(servLocation, nicknames);
    }

#ifdef USE_THREADS
  StartRequestThread();
#endif
  while (1)
    {
      ServiceInput();
#ifndef USE_THREADS
      ProcessRequests();
#endif
    }
}

void ServiceInput()
{
  static fd_set sockSet;
           
  // Check for new clients and add to the clients group.  This
  // could go after the new line, but by putting it here, the new
  // client will be serviced more quickly.  The downside is that
  // we are probing blindly for a new client which is a little
  // more expensive.
  AcceptNewClients();

  // Block on sockets ready to be read; this includes the server listening
  // socket so that the function will return when a new client is ready. 
  //printf("About to select\n");
  SocketSelect(&sockSet, g_sockets, g_servSock);

  //printf("selected\n");
  // Now, service each of the clients that has data, as indicated by being 
  // set in the sockSet.
#ifndef USE_THREADS
  ServiceClients(&sockSet);
  ServiceSignals(&sockSet);
#endif
}

int TestStateDirectory(char *filename)
{
  struct stat info;
  struct dirent *entry;
  DIR *dir;
  int okay;
  if (stat(filename, &info))
    {
      if (errno == ENOENT) // new directory entirely
	{
	  mkdir(filename, 0777);
	  return 1;
	}
    }
  if (!S_ISDIR(info.st_mode))
    {
      return 0;
    }
  dir = opendir(filename);
  do
    {
      entry = readdir(dir);
      if (entry)
	{
	  okay = 0;
	  if (!strcmp(entry->d_name, ".") ||
	      !strcmp(entry->d_name, ".."))
	    okay = 1;
	  if (!strcmp(entry->d_name, WHITEBOARD_FILE))
	    okay = 1;
	  if (!strcmp(entry->d_name, DOCUMENT_DIR))
	    okay = 1;
	  if (!strcmp(entry->d_name, LOG_FILE))
	    okay = 1;
	  if (!strcmp(entry->d_name, CERT_FILE))
	    okay = 1;
	  if (!strcmp(entry->d_name, KEY_FILE))
	    okay = 1;
	  if (!okay)
	    {
	      printf("Didn't expect %s in state directory\n", entry->d_name);
	      closedir(dir);
	      return 0;
	    }
	}
    }
  while (entry);
  closedir(dir);
  return 1;
}

int ReadIdentity(char *filename)
{
  char *certname, *keyname;
  if (filename)
    {
      certname = malloc(strlen(filename) + strlen(CERT_FILE) + 2);
      strcpy(certname, filename);
      strcpy(certname + strlen(filename), "/" CERT_FILE);
      keyname = malloc(strlen(filename) + strlen(KEY_FILE) + 2);
      strcpy(keyname, filename);
      strcpy(keyname + strlen(filename), "/" KEY_FILE);
    }
  else
    {
      certname = DEFAULT_CERT_FILE;
      keyname = DEFAULT_KEY_FILE;
    }
  return UseIdentity(certname, keyname);
}

int ConstructGlobalTree(void)
{
  ClientSelector selector;
  NodeType *node, *timeNode, *serverNode;
  NodeType *clientGroup;
  struct passwd *info;
  char *auth;

  serverID = (ClientIDType *)malloc(sizeof(ClientIDType));
  serverID->subscribedNodes = NULL;
  serverID->socket = NULL;
  serverID->name = "Server";
  serverID->auth = ArrayNew(sizeof(char *), 0, NULL);
  serverID->locks = NULL;
  auth = "server";
  ArrayAppend(serverID->auth, &auth);

  serverNode = CreateBranchNode("Server", (T_BRANCH));

  PublishLocal(NULL, CreateBranchNode("global", T_BRANCH), 0);

  node = CreateIntegerNode("pid", T_INT32, getpid());
  InsertNode(serverNode, node);

  node = CreateIntegerNode("uid", T_INT32, getuid());
  InsertNode(serverNode, node);

  info = getpwuid(getuid());
  if (info)
    {
      node = CreateStringNode("username", T_STRING, info->pw_name);
      InsertNode(serverNode, node);
    }

  clientGroup = CreateBranchNode("clients", T_BRANCH);

  PublishLocal("/clients", clientGroup, 0);

  timeNode = CreateTimeNode("time", T_TIMEVAL, NULL);

  SelectorSome(&selector, "server", NULL);
  LocalSetPathPermissions(clientGroup->name, 'w', &selector);
  LocalSetPathPermissions(clientGroup->name, 'l', &selector);
  LocalSetPathPermissions(clientGroup->name, 'a', &selector);

  PublishLocal("/clients/Server", serverNode, 0);

  PublishLocal("/time", timeNode, 0);

  IsSpecial(timeNode) = 1;

  LocalSetPathPermissions("/time", 'w', &selector);
  LocalSetPathPermissions("/time", 'l', &selector);
  LocalSetPathPermissions("/time", 'a', &selector);

  return 1;
}

int InitGlobalData(char *filename)
{
  InitSymbols();
  InitRequestQueue();
  if (1)
    {
      g_sockets = ArrayNew(sizeof(Socket *), 0, NULL);

      g_clients = ArrayNew(sizeof(ClientIDType *), 0, NULL);

#ifdef USE_THREADS
      pthread_mutex_init(&clientLock, NULL);
#endif
    }

  if (!ConstructGlobalTree())
    {
      if(SERVER_ERROR) printf("Failed to construct global tree.\n");
      exit(-1);
    }

  if (filename)
    {
      if (!TestStateDirectory(filename))
	{
	  printf("%s seems to be something else.\n", filename);
	  exit(-1);
	}
      if (!LoadStateFile(quiet, filename))
	{
	  if (SERVER_ERROR) printf("Failed to load persistant state.\n");
	  exit(-1);
	}
    }
  if (!ReadIdentity(filename))
    {
      fprintf(stderr, 
	      "Could not read identity files; disabling SSL support.\n");
    }
  if (KEEP_LOG)
    OpenLogFile(filename);

  // signal stuff here
  return 1;
}

void PublishNicknames(NetSocketLocation *location, DArray nicknames)
{
  int i;
  NodeType *node = CreateAddressNode("", T_SERVER, location);
  for (i = 0; i < ArrayLength(nicknames); i++)
    {
      const char *UNL = NicknameUNL(*((char **)ArrayNth(nicknames, i)));
      printf("Publish to %s\n", UNL);
      PublishNode(UNL, node, NULL);
    }
  FreeNode(node);
  ShutDownClientIPC();
}

void AddSocket(Socket *socket)
{
  ArrayAppend(g_sockets, &socket);
}

void RemoveSocket(Socket *socket)
{
  int i = ArraySearch(g_sockets, &socket, ArrayEq, 0, 0);
  if (i != NOT_FOUND)
    ArrayDeleteAt(g_sockets, i);
}

void AcceptNewClients()
{
  Socket *incoming;
#ifdef USE_THREADS
  pthread_t thread;
#endif

  // Loop handling one client connection request at a time
  while ((incoming = SocketAccept(g_servSock)))
    {
#ifdef USE_THREADS
      pthread_create(&thread, NULL, 
		     (void *(*)(void *)) HandleInitialContact, incoming);
#else
      HandleInitialContact(incoming);
#endif
      plogf("Initial contact handled\n");
    }
}

int HandleInitialContact(Socket *socket)
{
  // In the non-threaded case, this returns after the connection is in
  // the list. In the threaded case, it continues to serve the client,
  // waiting for the condition, and only returns after the client has
  // closed.
 
  IntroMessage message;
  int done;
  int type = 0;
  int ret;
  int use_ssl = 0;
  int using_ssl = 0;
  int version = PROTOCOL_VERSION_NUMBER;

  plogf("Handle new connection... ");

  do
    {
      done = 1;
      do
	{
	  ret = IntroRecv(socket, clientCommonProperties, &message);
	  if (ret < -1)
	    ;
	  else if (message.format == &protocolInteraction)
	    {
	      if (!strcmp(message.value.string_val, PROTOCOL_TYPE_WHITEBOARD))
		type = 1;
	      else if (!strcmp(message.value.string_val, PROTOCOL_TYPE_SIGNAL))
		type = 2;
	    }
	  else if (message.format == &protocolVersion)
	    {
	      if ((message.value.int_val & 0xFFFF00) !=
		  (version & 0xFFFF00))
		{
		  fprintf(stderr, 
			  "protocol version mismatch: %x remote %x local\n",
			  message.value.int_val,
			  version);
		  IntroFree(&message);
		  SocketClose(socket);
		  return 0;
		}
	      else
		{
		  if (version > message.value.int_val)
		    version = message.value.int_val;
		}
	    }
	  else if (message.format == &protocolEncrypt)
	    {
	      if (!strcmp(message.value.string_val, PROTOCOL_ENCRYPT_SSL))
		use_ssl = 1;
	    }
	  if (ret)
	    {
	      done = 0;
	      IntroFree(&message);
	    }
	}
      while (ret > 0);
      if (!done)
	{
	  done = 1;
	  if (type)
	    {
	      message.format = &protocolInteraction;
	      switch (type)
		{
		case 1:
		  message.value.string_val = PROTOCOL_TYPE_WHITEBOARD;
		  break;
		case 2:
		  message.value.string_val = PROTOCOL_TYPE_SIGNAL;
		  break;
		}
	      IntroSend(socket, &message);
	      message.format = &protocolVersion;
	      message.value.int_val = version;
	      IntroSend(socket, &message);
	      done = 0;
	    }
	  if (use_ssl && !using_ssl)
	    {
#ifdef WANT_SSL
	      if (DefaultIdentity())
		{
		  message.format = &protocolEncrypt;
		  message.value.string_val = PROTOCOL_ENCRYPT_SSL;
		  IntroSend(socket, &message);
		  done = 0;
		}
	      else
		use_ssl = 0;
#else
	      use_ssl = 0;
#endif
	    }
	  IntroDone(socket);
	}
      if (use_ssl && !using_ssl)
	{
	  SocketSecure(socket, NULL, 1);
	  using_ssl = 1;
	}
    }
  while (!done && !type);

  if (type == 1)
    {
      return HandleNewClient(socket);
    }
  else if (type == 2)
    {
      return HandleNewSignalProxy(socket);
    }

  SocketClose(socket);
  return 0;
}

void SetupSignals()
{
  if (1)
    {
      signal(SIGQUIT, ShutDownServer);
      signal(SIGSEGV, ShutDownServer);
      signal(SIGHUP, ShutDownServer);
      signal(SIGABRT, ShutDownServer);
      signal(SIGTERM, ShutDownServer);
    }
}

void ShutDownServer(int param)
{
  // Waits a second for each client to disconnect
  ClientIDType *clientID = NULL;
  int i;

  // First of all, don't catch fatal signals any more
  signal(SIGSEGV, SIG_DFL);
  signal(SIGTERM, SIG_DFL);

  SocketFactoryClose(g_servSock);

  PrintSockStats();

  for (i = 0; i < ArrayLength(g_clients); i++)
    {
      clientID = *((ClientIDType **)ArrayNth(g_clients, i));
      if (clientID == serverID) continue;
      SocketSendRecord(clientID->socket, "type", 4);
      WriteString(clientID->socket, S_TERMINAL, strlen(S_TERMINAL));
      SocketSendRecord(clientID->socket, NULL, 0);
      
      if (SocketReadReady(clientID->socket, 1000000))
	printf("Removed %s\n", clientID->name);
      else
	printf("Closing connection on %s\n", clientID->name);
      SocketClose(clientID->socket);
    }

  closeLog();
  exit(0);
}
