/*************************************************
 * File: tree-manip.c
 * Birth Date: 12/3/2003
 *
 * Project: Enchantment IPC
 * 
 * Based on Enchantment IPC Whitepaper
 *   by Edward Keyes.
 * 
 * Copyright (C) 2002 MIT Media Lab,
 *               2004 AWare Technologies, Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *************************************************/

/* This file implements manipulations of the server's tree state. */

#include "tree-manip.h"
#include "logging.h"

#include "whiteboard-protocol.h"
#include <enchantment/whiteboard.h>

#include <enchantment/unl.h>
#include "symbols.h"

#include "clients.h"
#include "journal.h"

#include <stdlib.h>
#include <string.h>

#include <stdarg.h>

#include <assert.h>

NodeType *g_globalNode = NULL;

extern ClientIDType *serverID;

static int SendUpdates(DArray subscribers, NodeType *node, const char *UNL,
		       ClientIDType *publisher, ClientIDType *only);

int FitsSelector(ClientSelector *selector, ClientIDType *client)
{
  int i, j;
  if (!selector->tokens)
    {
      if (SHOW_AUTH && client != serverID)
	printf("Any\n");
      return 1;
    }
  for (i = 0; selector->tokens[i]; i++)
    {
      for (j = 0; j < ArrayLength(client->auth); j++)
	{
	  if (!strcmp(selector->tokens[i],
		      *((char **)ArrayNth(client->auth, j))))
	    {
	      if (SHOW_AUTH && client != serverID)
		printf("Matched %s\n", selector->tokens[i]);
	      return 1;
	    }
	  if (SHOW_AUTH && client != serverID)
	    printf("%s != %s\n", selector->tokens[i],
		   *((char **)ArrayNth(client->auth, j)));
	}
    }
  if (SHOW_AUTH && client != serverID)
    {
      if (selector->tokens)
	{
	  int i;
	  printf("[");
	  for (i = 0; selector->tokens[i]; i++)
	    {
	      if (i)
		printf(", ");
	      printf("%s", selector->tokens[i]);
	    }
	  printf("]\n");
	}
      else
	printf("(any)\n");
    }
  return 0;
}

int CheckPermission(SchemaNode *schema, ClientIDType *client, char type)
{
  int ret;
  ClientSelector *selector;
  if (SHOW_AUTH && client != serverID)
    printf("Checking permission %c on %p for %s\n", type, schema, 
	   client->name);
  switch (type)
    {
    case 'w':
      // Locks prevent writes to a node
      if (schema->lock.owner && schema->lock.owner != client)
	return 0;
      selector = &schema->canWrite;
      break;
    case 'r':
      selector = &schema->canRead;
      break;
    case 'l':
      selector = &schema->canLock;
      break;
    case 'a':
      selector = &schema->canAdmin;
      break;
    default:
      return 0;
    }
  ret = FitsSelector(selector, client);
  if (SHOW_AUTH && client != serverID)
    printf("%s\n", ret ? "Allowed" : "Denied");
  return ret;
}

void SelectorNone(ClientSelector *selector)
{
  static const char **none = NULL;
  if (!none)
    {
      none = malloc(sizeof(char *));
      none[0] = NULL;
    }
  selector->tokens = none;
}

void SelectorAny(ClientSelector *selector)
{
  selector->tokens = NULL;
}

void SelectorSome(ClientSelector *selector, ...)
{
  va_list args;
  char *auth;
  int i;
  DArray auths = ArrayNew(sizeof(char *), 0, NULL);
  va_start(args, selector);
  do
    {
      auth = va_arg(args, char *);
      ArrayAppend(auths, &auth);
    }
  while (auth);
  va_end(args);
  selector->tokens = malloc(sizeof(char *) * ArrayLength(auths));
  for (i = 0; i < ArrayLength(auths); i++)
    selector->tokens[i] = *((const char **)ArrayNth(auths, i));
  ArrayFree(auths);
}

static SchemaNode *SchemaDefault(SchemaNode *parent, const char *name)
{
  SchemaNode *schema = (SchemaNode *)malloc(sizeof(SchemaNode));
  schema->lock.owner = NULL;
  schema->lockTrails = NULL;
  schema->subscribers = NULL;
  schema->hardNode = 0;
  schema->type = T_NONE;
  schema->hardSchema = 0;
  schema->parent = parent;
  schema->children = ArrayNew(sizeof(SchemaChild), 0, NULL);
  schema->special = 0;
  schema->implicitNode = 0;
  schema->distinct = 0;
  schema->symlink = NULL;
  schema->currentNode = NULL;
  if (SHOW_AUTH)
    printf("Created schema %p for %s\n", schema, name);
  if (parent)
    {
      SchemaChild child;
      child.name = name;
      child.node = schema;
      ArrayAppend(parent->children, &child);
      schema->canRead = parent->canRead;
      schema->canWrite = parent->canWrite;
      schema->canLock = parent->canLock;
      schema->canAdmin = parent->canAdmin;
    }
  else
    {
      SelectorAny(&schema->canRead);
      SelectorAny(&schema->canWrite);
      SelectorAny(&schema->canLock);
      SelectorAny(&schema->canAdmin);
    }
  return schema;
}

SchemaNode *GetSchemaChild(SchemaNode *parent, const char *name)
{
  int i;
  for (i = 0; i < ArrayLength(parent->children); i++)
    {
      SchemaChild *child =
	(SchemaChild *)ArrayNth(parent->children, i);
      if (!strcmp(child->name, name))
	return child->node;
    }
  return NULL;
}

typedef int (*TravSchemaFN)(void *data, SchemaNode *node, const char *UNL);

int TraverseSchema(SchemaNode *node, void *data, 
		   TravSchemaFN pre, TravSchemaFN post, const char *UNL)
{
  int i;
  int ret = 1;
  if (pre)
    {
      ret = (*pre)(data, node, UNL);
    }
  if (!ret)
    return 0;
  if (ret != 2)
    {
      for (i = 0; i < ArrayLength(node->children); i++)
	{
	  SchemaChild *child =
	    (SchemaChild *)ArrayNth(node->children, i);
	  const char *childUNL = UNL ? AppendUNLPath(UNL, child->name) : NULL;
	  ret = TraverseSchema(child->node, data, pre, post, childUNL);
	  if (!ret)
	    return ret;
	}
    }
  if (post)
    ret = (*post)(data, node, UNL);
  return ret;
}

int CheckNodePermission(NodeType *node, ClientIDType *client, char type)
{
  if (SHOW_AUTH && client != serverID)
    printf("For %s, ", node->name);
  return CheckPermission(SchemaOf(node), client, type);
}

void SetPermissions(SchemaNode *schema, char type, ClientSelector *selector)
{
  ClientSelector *target;
  switch (type)
    {
    case 'w':
      target = &schema->canWrite;
      break;
    case 'r':
      target = &schema->canRead;
      break;
    case 'l':
      target = &schema->canLock;
      break;
    case 'a':
      target = &schema->canAdmin;
      break;
    default:
      return;
    }
  *target = *selector;
  if (SHOW_AUTH)
    {
      printf("Set %c permission on %p to ", type, schema);
      if (selector->tokens)
	{
	  int i;
	  printf("[");
	  for (i = 0; selector->tokens[i]; i++)
	    {
	      if (i)
		printf(", ");
	      printf("%s", selector->tokens[i]);
	    }
	  printf("]\n");
	}
      else
	printf("(any)\n");
    }
}

void SetFixedType(SchemaNode *schema, TypeID type)
{
  schema->type = type;
}

void SetNodePermissions(NodeType *node, char type, ClientSelector *selector)
{
  SetPermissions(SchemaOf(node), type, selector);
}

struct checkPermData
{
  ClientIDType *clientID;
  char perm;
};

static int checkPermPre(struct checkPermData *data, SchemaNode *node,
			char *path)
{
  if (!CheckPermission(node, data->clientID, data->perm))
    return 0;
  return 1;
}

int CheckDescendantPermission(NodeType *node, ClientIDType *clientID,
			      char type)
{
  struct checkPermData data;
  data.perm = type;
  data.clientID = clientID;
  return TraverseSchema(SchemaOf(node), &data,
			(TravSchemaFN) checkPermPre, NULL, NULL);
}

/* Checks write permission for the nodes which would be modified. */
static int mayAddRec(ClientIDType *clientID, NodeType *node, 
		     SchemaNode *schema)
{
  //printf("Checking on %s\n", node->name);
  if (!CheckPermission(schema, clientID, 'w'))
    return 0;
  //printf("Checking on %s\n", node->name);
  if (GetFormat(node->type) == T_BRANCH)
    {
      DArray nodes = node->data.branch_val;
      int index = 0;
      NodeType *child;
      SchemaNode *schemaChild;
      while (index < ArrayLength(nodes))
	{
	  child = *((NodeType **)ArrayNth(node->data.branch_val, index));
	  schemaChild = GetSchemaChild(schema, child->name);
	  if (schemaChild)
	    {
	      if (!mayAddRec(clientID, child, schemaChild))
		return 0;
	    }
	  index++;
	}
    }
  return 1;
}


/* This tests in advance if we are permitted to do the
 * operation. Note: only needed for replace/delete, since the others
 * can simply do nothing if not permitted. Operation is therefore
 * determined by whether dataNode is NULL. (Does not check locks)
 */
static int mayPublish(ClientIDType *clientID, NodeType *dataNode,
		      NodePathReturnType *npReturn, int hard)
{
  if (dataNode)
    {
      // We need to traverse the schema and dataNode and determine:

      if (!npReturn->targetSchema)
	return ENCHANT_SUCCESS;
      if (!mayAddRec(clientID, dataNode, npReturn->targetSchema))
	return ENCHANT_LOCKED;
      // XXXX whether we will be allowed to publish everything in
      // dataNode to the UNLs they will get.

      // XXXX whether we are trying to publish the parent of a
      // required child without publishing the child.

      // We need to make sure we can write to every existing node in
      // the current value, if it exists
      if (npReturn->existNode &&
	  !CheckDescendantPermission(npReturn->existNode, clientID, 'w'))
	return ENCHANT_LOCKED;
    }
  else
    {
      // If there's already nothing there, fine.
      if (!npReturn->existNode)
	return ENCHANT_SUCCESS;

      // We need to make sure we can write to every existing node in
      // the current value.
      if (!CheckDescendantPermission(npReturn->existNode, clientID, 'w'))
	return ENCHANT_LOCKED;

      // We need to make sure we aren't removing a required child at
      // the root.
    }
  return ENCHANT_SUCCESS;
}

/**** Operations on tree structure ****/

#define LINK_HARD (1 << 0)
#define LINK_IMPLICIT (1 << 1)

struct unlinkInfo
{
  ClientIDType *clientID;
  // The deepest ancestor which will remain
  NodeType *lastLeft;
};

static int UnlinkNodePost(struct unlinkInfo *info, NodeType *node, 
			  NodeType *parent, const char *UNL)
{
  if (!parent)
    return 1; // We're replacing this node; don't unlink

  if (!InTree(node))
    {
      printf("Node already unlinked!\n");
      return 1;
    }

  SchemaOf(node)->currentNode = NULL;

  // We need to send NULL updates to subscribers, and update nodes in
  // subscriptions.
  SendUpdates(SchemaOf(node)->subscribers, NULL, UNL, info->clientID, NULL);

  free(node->local);
  node->local = NULL;

  RemoveNode(parent, node);
  FreeNode(node);
  return 1;
}

static void UnlinkNode(NodeType *parent, NodeType *node, const char *UNL, 
		       ClientIDType *clientID)
{
  // Anyone subscribed to this node or a descendant will be notified in
  // transit, when the subscribed node is unlinked.
  struct unlinkInfo info = { clientID, parent };
  TraverseTree(node, parent, &info, NULL, (TravTreeFN) UnlinkNodePost, UNL);
}

struct linkInfo
{
  ClientIDType *clientID;
  int flags;
};

static int LinkNodePre(struct linkInfo *info, NodeType *node, 
		       NodeType *parent, const char *UNL)
{
  SchemaNode *schema;
  SchemaNode *schemaParent;
  assert(node->name);

  if (InTree(node)) // just need the children here.
    return 1;

  schema = NULL;
  if (parent)
    {
      schemaParent = SchemaOf(parent);
      schema = GetSchemaChild(schemaParent, node->name);
    }
  else
    schemaParent = NULL;

  if (!schema)
    {
      schema = SchemaDefault(schemaParent, node->name);
      schema->hardNode = info->flags & LINK_HARD;
      schema->hardSchema = schema->hardNode;
    }

  node->local = malloc(sizeof(PathTreeDataType));
  memset(node->local, 0, sizeof(PathTreeDataType));
  SchemaOf(node) = schema;
  ParentOf(node) = parent;
  schema->currentNode = node;
  if (!(info->flags & LINK_IMPLICIT))
    SendUpdates(schema->subscribers, node, UNL, info->clientID, NULL);
  return 1;
}

/* Can fail if parent isn't a branch */
static int LinkNode(NodeType *parent, NodeType *node, const char *UNL, 
		    ClientIDType *clientID, int flags)
{
  struct linkInfo info;
  info.flags = flags;
  info.clientID = clientID;
  if (parent && !InsertNode(parent, node))
    return 0;
  TraverseTree(node, parent, (void *)&info, (TravTreeFN) LinkNodePre, 
	       NULL, UNL);
  ParentOf(node) = parent;
  return 1;
}

static void SwapNodes(NodeType *node1, NodeType *node2)
{
  EnchantData data = node1->data;
  const char *name = node1->name;
  long type = node1->type;

  node1->data = node2->data;
  node1->name = node2->name;
  node1->type = node2->type;

  node2->data = data;
  node2->name = name;
  node2->type = type;
}

static void ReplaceNode(NodeType *old, NodeType *new, const char *UNL, 
			ClientIDType *clientID, int hard)
{
  struct unlinkInfo info = { clientID, old };
  NodeType *parent = ParentOf(old);

  struct linkInfo link;
  link.clientID = clientID;
  link.flags = hard ? LINK_HARD : 0;
  // Unlink all of the descendants of old, but not old itself
  // NULL parent tells UnlinkNodePost not to do it.
  TraverseTree(old, NULL, &info, NULL, (TravTreeFN) UnlinkNodePost, UNL);
  SwapNodes(old, new);
  FreeNode(new);
  TraverseTree(old, parent, &link, (TravTreeFN) LinkNodePre, NULL, UNL);

  SendUpdates(SchemaOf(old)->subscribers, old, UNL, clientID, NULL);

  SchemaOf(old)->hardNode = hard;
  if (hard)
    SchemaOf(old)->hardSchema = 1;
}

static int MergeNode(const char *path, NodeType *intoNode, NodeType *fromNode,
		     ClientIDType *permittedOwner, int DESTRUCTIVE, int hard)
{
  int mayWrite;
  /*
  printf("merge\n");
  PrintNode(fromNode, NULL);
  printf("into\n");
  PrintNode(intoNode, NULL);
  */

  // A recursive lock we don't own stops things right away.
  if (SchemaOf(intoNode)->lock.owner &&
      SchemaOf(intoNode)->lock.recurse &&
      SchemaOf(intoNode)->lock.owner != permittedOwner)
    return 1;
  mayWrite = CheckNodePermission(intoNode, permittedOwner, 'w');
  if (GetFormat(intoNode->type) == T_BRANCH && 
      GetFormat(fromNode->type) == T_BRANCH)
    {
      // They're both branches; we're not just replacing
      int index;
      NodeType *mergeChild = NULL;
      if ((DESTRUCTIVE ||
	   (hard && !SchemaOf(intoNode)->hardNode)) && 
	  mayWrite)
	{
	  // We want to change things and we're allowed to. So we
	  // replace information local to the branch node.
	  intoNode->type = fromNode->type;
	  // It can't be implicit after this
	  IsImplicit(intoNode) = 0;
	}
      // We can't do anything with the children of a node we can't read
      if (!CheckNodePermission(intoNode, permittedOwner, 'r'))
	return 1;
      for (index = 0; (mergeChild = BranchIndex(fromNode, index)); index++)
	{
	  NodeType *existChild = GetBranchChild(intoNode, mergeChild->name);
	  if (existChild)
	    {
	      const char *newPath = AppendUNLPath(path, mergeChild->name);
	      MergeNode(newPath, existChild, mergeChild, permittedOwner, 
			DESTRUCTIVE, hard);
	    }
	  else
	    {
	      if (mayWrite)
		{
		  NodeType *copyNode = NULL;
		  SchemaNode *childSchema =
		    GetSchemaChild(SchemaOf(intoNode), mergeChild->name);
		  if (childSchema && 
		      !CheckPermission(childSchema, permittedOwner, 'w'))
		    return 1;

		  copyNode = CopyNode(mergeChild);
		  //printf("With UNL %s:\n", newPath);
		  //PrintNode(copyNode, NULL);
		  LinkNode(intoNode, copyNode, 
			   AppendUNLPath(path, copyNode->name),
			   permittedOwner, hard ? LINK_HARD : 0);
		}
	    }
	}
      //Send updates to subscribed nodes
      if (InTree(intoNode))
	SendUpdates(SchemaOf(intoNode)->subscribers, intoNode, path,
		    permittedOwner, NULL);
      return 1;
    }
  else
    {
      NodeType *copyNode;
      // We are going to try to replace the node; stop if the
      // operation is add-only or if we can't write here.
      if (!DESTRUCTIVE)
	return 1;
      if (!mayWrite)
	return 1;
      // Branches with locked children are protected, even against the
      // owner
      if (GetFormat(intoNode->type) == T_BRANCH && 
	  SchemaOf(intoNode)->lockTrails && 
	  ArrayLength(SchemaOf(intoNode)->lockTrails) > 0)
	return 1;
      // Can't replace a persistant node with a non-persistant one
      if (!hard && SchemaOf(intoNode)->hardNode)
	return 1;
      copyNode = CopyNode(fromNode);
      ReplaceNode(intoNode, copyNode, path, permittedOwner, hard);
      return 1;
    }
}

/**** Returning information about paths ****/

// Concatenate two subscription lists for npReturn
static void ConcatSubscriptions(DArray *p_subscriptArray, NodeType *node,
				const char *UNL)
{
  int i, len;
  SubscriptionType *p_subscriber = NULL;

  DArray subscriberArray = SchemaOf(node)->subscribers;

  if (!subscriberArray || !(len = ArrayLength(subscriberArray)))
    return;
  
  if (!*p_subscriptArray) 
    *p_subscriptArray = ArrayNew(sizeof(SubscriptionType *), len, NULL);
      
  for (i = 0; i < len; i++)
    {
      p_subscriber = *((SubscriptionType **) ArrayNth(subscriberArray, i));
      ArrayAppend(*p_subscriptArray, &p_subscriber);
    } 
}

static void applySchema(SchemaNode *current, NodePathReturnType *npReturn)
{
  ServerLockType *currLock = &current->lock;
  // A recursive lock has different effects from lacking write permission:
  // on any ancestor it prevents changes, not just on a few relevant nodes.
  if (currLock->owner && currLock->recurse)
    {
      if (currLock->owner != npReturn->clientID)
	{
	  npReturn->mayCreateRemove = 0;
	  npReturn->mayModify = 0;
	}
      npReturn->reportLock = currLock;
    }

  if (currLock->owner == npReturn->clientID)
    {
      // We own a lock on this node, either a recursive lock on an
      // ancestor, or a lock on the node itself.
      if (currLock->recurse || current == npReturn->targetSchema)
	{
	  npReturn->controlled = 1;
	}
    }

  if (!CheckPermission(current, npReturn->clientID, 'r'))
    {
      npReturn->mayRead = 0;
      if (current != npReturn->targetSchema)
	{
	  npReturn->mayCreateRemove = 0;
	  npReturn->mayModify = 0;
	}
    }

  if (!CheckPermission(current, npReturn->clientID, 'w'))
    {
      if (current == npReturn->targetSchema)
	{
	  // Can't write to node
	  npReturn->mayCreateRemove = 0;
	  npReturn->mayModify = 0;
	}
      if (current == npReturn->closestSchema)
	{
	  // Can't write to the ancestor which has the list of children which
	  // would change to include this UNL.
	  npReturn->mayCreateRemove = 0;
	}
    }
}

static void GetSchemaRec(SchemaNode *current, const char *path,
			 NodePathReturnType *npReturn, int createSchema)
{
  if (current->symlink &&
      (path ? npReturn->cross_symlinks : 
       npReturn->follow_symlinks))
    {
      // The UNL is at or descended from a symlink
      npReturn->targetSchema = current;
      
      npReturn->symlinkCont = path;
      npReturn->symlinkTarget = current->symlink;
      
      npReturn->getSubscriptions = 0;
    }
  if (!path)
    {
      npReturn->targetSchema = current;
      if (current->lock.owner)
	npReturn->reportLock = &current->lock;
    }
  else
    {
      const char *head = GetUNLHead(path);
      const char *tail = GetUNLTail(path);
      SchemaNode *child = GetSchemaChild(current, head);

      if (!child && createSchema)
	child = SchemaDefault(current, head);
      if (child)
	{
	  GetSchemaRec(child, tail, npReturn, createSchema);
	}
      if (!npReturn->closestSchema)
	{
	  npReturn->closestSchema = current;
	}
    }
  applySchema(current, npReturn);
}

static void GetNodePathInfoRec(NodeType *currNode, const char *path,
			       NodePathReturnType *npReturn,
			       int createSchema)
{
  const char *currUNL;
  const char *headName, *tailName;
  NodeType *pathNode;
   
  if (!path || !path[0])
    {
      // We're at the end of the path.
      npReturn->existNode = currNode;
      npReturn->targetSchema = SchemaOf(currNode);
    }
  else
    {
      headName = GetUNLHead(path);
      tailName = GetUNLTail(path);
  
      // UNL of the current node
      currUNL = InstanceStringN(npReturn->UNL, 
				strlen(npReturn->UNL) - 
				strlen(path));
  
      // On the way down....
      pathNode = GetBranchChild(currNode, headName);
      
      if (!pathNode)
	{
	  // We've come to the end of the tree without coming to the
	  // end of the UNL. This is the closest ancestor, and the
	  // node doesn't exist.
	  
	  npReturn->existNode = NULL;
	  npReturn->closestUNL = currUNL;
	  npReturn->closestAncestorNode = currNode;
	  GetSchemaRec(SchemaOf(currNode), path, npReturn, createSchema);
	}
      else
	{
	  // Got a valid path to follow, so recurse downwards
	  GetNodePathInfoRec(pathNode, tailName, npReturn, createSchema);

	  if (SchemaOf(pathNode)->symlink &&
	      (tailName ? npReturn->cross_symlinks : 
	       npReturn->follow_symlinks))
	    {
	      // The UNL is at or descended from a symlink
	      npReturn->existNode = pathNode;
	      npReturn->targetSchema = SchemaOf(pathNode);
	      
	      npReturn->closestUNL = currUNL;
	      npReturn->closestAncestorNode = currNode;
	      npReturn->symlinkCont = tailName;
	      npReturn->symlinkTarget = SchemaOf(pathNode)->symlink;
	      
	      npReturn->getSubscriptions = 0;
	    }

	  // If the existNode was set and the existParent is still NULL, then
	  // the current current node must be the existParent.
	  if (npReturn->existNode && !npReturn->existParent)
	    {
	      // The node exists, and its parent hasn't been found yet;
	      // therefore, this node is that parent.
	      npReturn->existParent = currNode;
	      
	      // Also, it must be the closest existing ancestor, since it
	      // exists and is the closest possible ancestor.
	      npReturn->closestAncestorNode = currNode;
	      npReturn->closestUNL = currUNL;

	      npReturn->closestSchema = SchemaOf(currNode);
	    }
	}
    }
  applySchema(SchemaOf(currNode), npReturn);
  // We don't list subscribers to the target node, because updates for
  // the target node, missing ancestors of the target node, and
  // descendants of the target node are handled separately.
  if (npReturn->existNode != currNode)
    {
      if (npReturn->getSubscriptions &&
	  SchemaOf(currNode)->subscribers)
	{
	  ConcatSubscriptions(&npReturn->subscriptions, currNode, currUNL);
	}
    }
}

void GetNodePathInfo(ClientIDType *clientID,
		     NodePathReturnType *npReturn,
		     int createSchema)
{
  npReturn->existNode = NULL;
  npReturn->existParent = NULL;
  npReturn->closestUNL = NULL;
  npReturn->closestAncestorNode = NULL;
  npReturn->clientID = clientID;
  npReturn->targetSchema = NULL;
  npReturn->closestSchema = NULL;
  npReturn->closestSchemaUNL = NULL;
  npReturn->controlled = 0;

  npReturn->reportLock = NULL;
  npReturn->mayCreateRemove = 1;
  npReturn->mayModify = 1;
  npReturn->mayRead = 1;
  npReturn->subscriptions = NULL;
  npReturn->symlinkTarget = NULL;
  npReturn->symlinkCont = NULL;

  GetNodePathInfoRec(g_globalNode, npReturn->UNL, npReturn, createSchema);

  if (npReturn->existNode && 
      SchemaOf(npReturn->existNode)->lock.owner)
    {
      npReturn->reportLock = &SchemaOf(npReturn->existNode)->lock;
    }
}

/**** Subscriptions ****/

static int SendUpdates(DArray subscribers, NodeType *node, const char *UNL,
		       ClientIDType *publisher, ClientIDType *only)
{
  int i, iLen;
  if (subscribers)
    {
      iLen = ArrayLength(subscribers);
      for (i = 0; i < iLen; i++)
	{
	  SubscriptionType *subscription =
	    *((SubscriptionType **) ArrayNth(subscribers, i));
	  // If we only send updates to a single client, skip others
	  if (only && subscription->subscriber != only)
	    continue;
	  // If we don't report our own publications, skip ourself.
	  if (publisher == subscription->subscriber &&
	      !subscription->self)
	    continue;
	  SendUpdate(subscription, node, UNL, 
		     publisher->name, NULL, NULL);
	}
      return iLen;
    }
  return 0;
}

// Send messages to subscribers of existing nodes above the target
static int InformSubscribers(ClientIDType *clientID, 
			     NodePathReturnType *npReturn,
			     NodeType *replace, NodeType *with)
{
  // Note that this sends all the updates where more of the tree than
  // just the changed node is sent, because these are the
  // subscriptions to an extant ancestor.
  if (npReturn->subscriptions)
    {
      int i, len = ArrayLength(npReturn->subscriptions);
      SubscriptionType *p_subscription = NULL;
      for (i = 0; i < len; i++)
	{
	  p_subscription = *((SubscriptionType **) 
			     ArrayNth(npReturn->subscriptions, i));
	  plogf("Inform %s\n", p_subscription->subscriber);
	  // If this is a publication to an update-only lock, and the
	  // publisher doesn't hold it, skip anyone other than the
	  // publisher.
	  if (npReturn->reportLock && npReturn->reportLock->update &&
	      npReturn->reportLock->owner != clientID &&
	      npReturn->reportLock->owner != p_subscription->subscriber)
	    continue;
	  // If we don't see our own publications, skip ourself.
	  if (!p_subscription->self &&
	      p_subscription->subscriber == clientID)
	    continue;
	  SendUpdate(p_subscription, 
		     p_subscription->schema->currentNode,
		     npReturn->UNL, clientID->name,
		     replace, with);
	}
      ArrayFree(npReturn->subscriptions);
      npReturn->subscriptions = NULL;
      return len;
    }
  return 0;
}

/* Note that removing subscriptions is split, because subscriptions
 * are often removed when a node or a client is also removed, in which
 * case only one function needs to be called, since the whole other
 * array will be deleted together afterwards. 
 */

/* Remove subscription from subscriber's list */
void ForgetSubscription(ClientIDType *subscriber, SubscriptionType *sub)
{
  DArray subscribedNodes = subscriber->subscribedNodes;
  int j, jLen = ArrayLength(subscribedNodes);
  for(j = 0; j < jLen; j++)
    {
      SubscriptionType *subscription = 
	*((SubscriptionType **) ArrayNth(subscribedNodes, j));
      if(subscription == sub)
	{
	  if (jLen == 1)
	    {
	      ArrayFree(subscribedNodes);
	      subscriber->subscribedNodes = NULL;
	    }
	  else
	    ArrayDeleteAt(subscribedNodes, j);
	  return;
	}
    }
  printf("Didn't find subscription!\n");
}

/* Remove subscription from a subscribed node's list. */
void ForgetSubscriber(SubscriptionType *subscription)
{
  DArray subscribers = subscription->schema->subscribers;
  SchemaNode *schema = subscription->schema;
  int j, jLen;
  if (subscribers)
    {
      jLen = ArrayLength(subscribers);
      for (j = 0; j < jLen; j++)
	{
	  SubscriptionType *sub = 
	    *((SubscriptionType **) ArrayNth(subscribers, j));
	  if (sub == subscription)
	    {
	      free(sub);
	      if (jLen == 1)
		{
		  ArrayFree(subscribers);
		  schema->subscribers = NULL;
		}
	      else 
		ArrayDeleteAt(subscribers, j);
	      return;
	    }
	}
    }
  printf("Didn't find subscriber!\n");
}

/* Returns 0 to indicate that it didn't do anything, in which case
 * the operation was not needed. */
int Subscribe(ClientIDType *clientID, NodePathReturnType *npReturn, 
	      int SELF_SUBSCRIBE)
{
  int i, iLen;
  DArray subscribers, subscribedNodes;
  SubscriptionType *newSubscriber;

  subscribers = npReturn->targetSchema->subscribers;
  if (!subscribers)
    {
      npReturn->targetSchema->subscribers = 
	ArrayNew(sizeof(SubscriptionType *), 1, NULL);
      subscribers = npReturn->targetSchema->subscribers;
    }

  subscribedNodes = clientID->subscribedNodes;
  if (!subscribedNodes)
    {
      clientID->subscribedNodes = 
	ArrayNew(sizeof(SubscriptionType *), 1, NULL);
      subscribedNodes = clientID->subscribedNodes;
    }
  else
    {
      // Only need to check for duplicates if there were subscriptions before
      iLen = ArrayLength(subscribedNodes);
      for (i = 0; i < iLen; i++)
	{
	  SubscriptionType *sNode =
	    *(SubscriptionType **) ArrayNth(subscribedNodes, i);
	  if (sNode->UNL == npReturn->UNL)
	    {
	      sNode->self = SELF_SUBSCRIBE;
	      return 0;
	    }
	}
    }
  newSubscriber = malloc(sizeof(SubscriptionType));
  newSubscriber->subscriber = clientID;
  newSubscriber->self = SELF_SUBSCRIBE;
  newSubscriber->UNL = npReturn->UNL;
  newSubscriber->schema = npReturn->targetSchema;
  newSubscriber->id = clientID->nextID;
  clientID->nextID++;
  ArrayAppend(subscribers, &newSubscriber);
  ArrayAppend(subscribedNodes, &newSubscriber);
  return newSubscriber->id;
}

SubscriptionType *IsSubscribed(ClientIDType *clientID, const char *UNL)
{
  DArray subscribedNodes = clientID->subscribedNodes;
  if (subscribedNodes)
    {
      int i, iLen = ArrayLength(subscribedNodes);
      for (i = 0; i < iLen; i++)
	{
	  SubscriptionType *sNode = 
	    *(SubscriptionType **) ArrayNth(subscribedNodes, i);
	  //printf("Looking for sub: %s ?= %s\n", sNode->UNL, UNL);
	  if (sNode->UNL == UNL)
	    return sNode;
	}
    }
  return NULL;
}

/**** Locks ****/

static int ContainsUnownedLockTrails(NodeType *node, ClientIDType *owner)
{
  int len;
  DArray lockTrails;

  lockTrails = SchemaOf(node)->lockTrails;

  if (!lockTrails || (len = ArrayLength(lockTrails)) == 0)
    {
      return 0;
    }
  else if (len > 1)
    {
      LockTrailType *p_lockTrail = (LockTrailType *) ArrayNth(lockTrails, 0);

      // Can't both be the owner's
      printf("multiple lock trails; first by %s\n", p_lockTrail->owner->name);
      return 1;
    }
  else
    {
      LockTrailType *p_lockTrail = (LockTrailType *) ArrayNth(lockTrails, 0);
      
      return (p_lockTrail->owner != owner);
    }
}

static void RecurseRemoveLockTrail(NodeType *node, const char *path, 
				   LockTrailType *p_endLockTrail)
{
  DArray lockTrails = NULL;
  int i, len;
  const char *head, *tail;
  NodeType *recurseNode = NULL;

  printf("Clearing locks for %s %s\n", path, p_endLockTrail->owner->name);
  
  lockTrails = SchemaOf(node)->lockTrails;

  len = ArrayLength(lockTrails);

  for (i = 0; i < len; i++)
    {
      LockTrailType *p_trail = (LockTrailType *) ArrayNth(lockTrails, i);
      
      if (p_trail->owner == p_endLockTrail->owner)
	{
	  assert(p_trail->count >= p_endLockTrail->count);
	  p_trail->count -= p_endLockTrail->count;
	  if (p_trail->count == 0)
	    ArrayDeleteAt(lockTrails, i);
	  
	  if (p_trail == p_endLockTrail) 
	    return;
	  break;
	}
    }
     
  head = GetUNLHead(path);
  tail = GetUNLTail(path);
      
  // Get the next node on the path
  recurseNode = GetBranchChild(node, head);
      
  RecurseRemoveLockTrail(recurseNode, tail, p_endLockTrail);
}

static int RemoveLocksPre(ClientIDType *clientID, SchemaNode *node, 
			  char *UNL)
{
  int i, len, ret;
  DArray lockTrails;

  if (node->lock.owner == clientID)
    {
      if (node->lock.remove)
	{
	  NodePathReturnType npReturn;
	  npReturn.UNL = UNL;
	  npReturn.follow_symlinks = 0;
	  npReturn.cross_symlinks = 0;
	  npReturn.getSubscriptions = 1;
	  GetNodePathInfo(clientID, &npReturn, 0);
	  // We need to remove the node, not just the lock
	  if (mayPublish(clientID, NULL, &npReturn, 0) == ENCHANT_SUCCESS)
	    {
	      UnlinkNode(node->parent->currentNode, node->currentNode, 
			 UNL, clientID);
	      InformSubscribers(clientID, &npReturn, NULL, NULL);
	    }
	}
      node->lock.owner = NULL;
    }

  lockTrails = node->lockTrails;
      
  if (!lockTrails)
    return 2;

  len = ArrayLength(lockTrails);

  ret = 2;
  for (i = len - 1; i >= 0; i--)
    {
      LockTrailType *p_lockTrail = (LockTrailType *) ArrayNth(lockTrails, i);
	  
      if (p_lockTrail->owner == clientID)
	{
	  ArrayDeleteAt(lockTrails, i);
	  ret = 1;
	  break;
	}
    }

  return ret;
}

static int RemoveLocksPost(ClientIDType *clientID, SchemaNode *node, 
			   char *UNL)
{
  NodeType *current = node->currentNode;
  if (!current)
    return 1;
  if (!IsImplicit(current))
    return 1;
  // If this leaves an empty implicit branch, remove it.
  if (current && IsImplicit(current) && !ArrayLength(current->data.branch_val))
    UnlinkNode(node->parent->currentNode, current, UNL, clientID);
  return 1;
}

static int RemoveLocksByClient(ClientIDType *clientID)
{
  TraverseSchema(SchemaOf(g_globalNode), clientID, 
		 (TravSchemaFN) RemoveLocksPre, 
		 (TravSchemaFN) RemoveLocksPost, "");
  if (clientID->locks)
    {
      ArrayFree(clientID->locks);
    }
  return 1;
}

static void DecrementLockTrail(SchemaNode *node, ClientIDType *owner)
{
  int i, len;
  DArray lockTrails = node->lockTrails;
  len = ArrayLength(lockTrails);

  for (i = 0; i < len; i++)
    {
      LockTrailType *p_lockTrail = (LockTrailType *) ArrayNth(lockTrails, i);
      if (p_lockTrail->owner == owner) 
	{
	  if (p_lockTrail->count > 1) 
	    p_lockTrail->count--;
	  else
	    ArrayDeleteAt(lockTrails, i);
	  return;
	}
    }
  printf("Couldn't find the lock trail for the removed lock!\n");
}

static void IncrementLockTrail(SchemaNode *node, ClientIDType *owner)
{
  int i, len;
  LockTrailType newLockTrail;
  DArray *p_lockTrails = &node->lockTrails;

  // Lazy allocate child Lock Owners array for the parent node
  if (!(*p_lockTrails))
    {
      *p_lockTrails = ArrayNew(sizeof(LockTrailType), 1, NULL);
      len = 0;
    }
  else
    len = ArrayLength(*p_lockTrails);

  for (i = 0; i < len; i++)
    {
      LockTrailType *p_lockTrail = 
	(LockTrailType *) ArrayNth(*p_lockTrails, i);
      if (p_lockTrail->owner == owner) 
	{
	  p_lockTrail->count++;
	  return;
	}
    }

  newLockTrail.owner = owner;
  newLockTrail.count = 1;
  
    // Put the owner into this path node
  ArrayAppend(*p_lockTrails, &newLockTrail); 
}

static int RecurseLockNodePath(SchemaNode *parent, const char *path, 
			       ServerLockType *p_newLock)
{
  ServerLockType *p_existLock = &parent->lock;
  const char *head, *tail;
  SchemaNode *child;
  
  if (!path)
    {
      // We've gotten to the node we want to lock

      // Already locked (by us or someone else)
      if (p_existLock->owner)
	return 0;

      // We're not allowed to lock this node
      if (!CheckPermission(parent, p_newLock->owner, 'l'))
	return 0;

      *p_existLock = *p_newLock;
      ArrayAppend(p_existLock->owner->locks, &p_existLock);
      return 1;
    }

  // Recursive check: The path is not null, so the node we want to
  // lock is farther down the path.  However, before we can
  // descend, we need to make sure that the current parent node
  // does not have any recursive locks on it.
  
  if (p_existLock->owner && p_existLock->recurse)
    {
      // There's a recursive lock here
      return 0;
    }

  // Not allowed to read the children
  if (!CheckPermission(parent, p_newLock->owner, 'r'))
    return 0;
      
  head = GetUNLHead(path);
  tail = GetUNLTail(path);
      
  // Get the next node on the path
  child = GetSchemaChild(parent, head);
      
  if (!child)
    {
      // There's no such node...
      return 0;
    }
      
  // Recursive case: We need to recurse down the path and get the
  // result so that the proper changes can be made to this node
  // upon success.
  if (!RecurseLockNodePath(child, tail, p_newLock))
    {
      // Problems further down
      return 0;
    }
      
  // SUCCESSFUL lock: we need to leave the lock trail in this parent.
  IncrementLockTrail(parent, p_newLock->owner);
  return 1;
}

static int RecurseUnlockNodePath(SchemaNode *schema, const char *path,
				 ClientIDType *owner)
{
  const char *head, *tail;
  int i;
  
  if (path) // This indicates that the desired node is below this one
    { 
      SchemaNode *child;
      
      head = GetUNLHead(path);
      tail = GetUNLTail(path);
      
      // Get the next node on path
      child = GetSchemaChild(schema, head);

      if (!child)
	{
	  // No such node
	  printf("No such node %s\n", head);
	  return 0;
	}

      // Path is good so far, so keep on descending
      if (!RecurseUnlockNodePath(child, tail, owner))
	{
	  // Problems below
	  return 0;
	}

      // Since the node was successfully unlocked below, we
      // should be able to find and decrement a LockTrailType
      // entry from this node above.
      
      DecrementLockTrail(schema, owner);
      return 1;
    }

  // Test this end node to see if passed in owner owns it
  if (schema->lock.owner != owner)
    {
      printf("Didn't own it\n");
      return 0;
    }

  // The owner does indeed own this node, so unlock it and indicate
  // this in the result. 
  for (i = 0; i < ArrayLength(schema->lock.owner->locks); i++)
    {
      if (*((ServerLockType **)ArrayNth(schema->lock.owner->locks, i)) ==
	  &schema->lock)
	{
	  ArrayDeleteAt(schema->lock.owner->locks, i);
	  break;
	}
    }
  schema->lock.owner = NULL;
  return 1;
}

int LockNodePath(const char *path, ServerLockType *lock)
{
  if (!lock->owner->locks)
    lock->owner->locks = ArrayNew(sizeof(ServerLockType *), 0, NULL);
  return RecurseLockNodePath(SchemaOf(g_globalNode), path, lock);
}

int UnlockNodePath(const char *path, ClientIDType *owner)
{
  return RecurseUnlockNodePath(SchemaOf(g_globalNode), path, owner);
}

/**** Utilities for generating automatically generated nodes ****/

// Returns a name not used in parent
static void DistinctName(const char **name_p, NodeType *parent)
{
  static int id = 0; 
  char buffer[20];
  // XXXX this needs to take into account hard published children on a restart
  sprintf(buffer, "%d", id);
  id++;
  *name_p = InstanceString(buffer);
}

static void CreateAncestors(NodePathReturnType *npReturn, 
			    ClientIDType *clientID)
{
  const char *path = DescendsFrom(npReturn->closestUNL, npReturn->UNL);
  const char *headName, *tailName, *UNL;
  NodeType *branch, *parent;

  parent = npReturn->closestAncestorNode;

  headName = GetUNLHead(path);
  tailName = GetUNLTail(path);

  while (tailName)
    {
      branch = CreateBranchNode(headName, T_BRANCH);
      // Never hard
      UNL = AppendUNLPath(npReturn->closestUNL, headName);
      LinkNode(npReturn->closestAncestorNode, branch, UNL, clientID, 
	       LINK_IMPLICIT);
      ConcatSubscriptions(&npReturn->subscriptions, branch, UNL);
      IsImplicit(branch) = 1;
      npReturn->closestAncestorNode = branch;
      npReturn->closestUNL = UNL;
      path = tailName;
      headName = GetUNLHead(path);
      tailName = GetUNLTail(path);
    }
}

// Given a path and a node, returns a created branch which puts the
// node at the path.
static NodeType *EmbedNode(const char *path, NodeType *node)
{
  const char *headName = NULL, *tailName = NULL;
  NodeType *branch;

  //printf("Working on %s\n", path);
  headName = GetUNLHead(path);
  tailName = GetUNLTail(path);
  //printf("Head %s tail %s\n", headName, tailName);

  if (!tailName)
    {
      if (headName)
	node->name = headName;
      return node;
    }
  branch = CreateBranchNode(headName, T_BRANCH);
  InsertNode(branch, EmbedNode(tailName, node));
  //PrintNode(branch, NULL);
  return branch;
}

static int doPublish(ClientIDType *clientID, NodeType *dataNode,
		     NodePathReturnType *npReturn, const char *op,
		     int hard, int publish_if_locked)
{
  int ret;
  const char *target = npReturn->UNL;

  if (!publish_if_locked && npReturn->controlled)
    {
      printf("Publish from untrusted symlink failed on locked node.\n");
      return ENCHANT_LOCKED;
    }

  if (!npReturn->mayModify)
    {
      if (npReturn->reportLock && npReturn->reportLock->update)
	return ENCHANT_SENT;
      return ENCHANT_LOCKED;
    }
  
  if (npReturn->targetSchema && npReturn->targetSchema->type && dataNode &&
      dataNode->type != npReturn->targetSchema->type)
    {
      // Tried to change type of node with fixed type
      printf("Fixed type\n");
      return ENCHANT_TYPE;
    }

  // Trying to create a client-chosen child of a distinct UNL
  if (dataNode && !npReturn->existNode && npReturn->closestAncestorNode &&
      SchemaOf(npReturn->closestAncestorNode)->distinct)
    {
      return ENCHANT_TYPE;
    }
  
  if (npReturn->targetSchema && npReturn->targetSchema->distinct && dataNode)
    {
      DistinctName(&dataNode->name, npReturn->existNode);
      // Move the reported UNL
      target = AppendUNLPath(target, dataNode->name);
      if (npReturn->existNode)
	{
	  npReturn->closestUNL = npReturn->UNL;
	  npReturn->closestAncestorNode = npReturn->existNode;
	}
      npReturn->UNL = target;
      npReturn->existParent = NULL;
      npReturn->existNode = NULL;
    }

  if (npReturn->existNode)
    {
      if (!hard && SchemaOf(npReturn->existNode)->hardNode)
	{
	  printf("Hard published\n");
	  return ENCHANT_LOCKED;
	}

      if (op == kv_merge || op == kv_addonly)
	{
	  // Merge/Add
	  if (MergeNode(target, npReturn->existNode, dataNode, 
			clientID, op == kv_merge, hard))
	    {
	      if (hard)
		{
		  HardPublish(dataNode, target);
		}
	      FreeNode(dataNode);
	      return ENCHANT_SUCCESS;
	    }
	  return ENCHANT_ERROR;
	}

      // Now we're down to replacement, which needs to check that
      // children are modifiable.
      ret = mayPublish(clientID, dataNode, npReturn, hard);
      if (ret != ENCHANT_SUCCESS)
	return ret;

      if (ContainsUnownedLockTrails(npReturn->existNode, clientID))
	{
	  printf("Contains locked children\n");
	  // There are locked children, no replace or delete
	  return ENCHANT_LOCKED;
	}
      if (dataNode)
	{
	  if (hard)
	    {
	      HardPublish(dataNode, target);
	    }
	  ReplaceNode(npReturn->existNode, dataNode, target, 
		      clientID, hard);

	  return ENCHANT_SUCCESS;
	}

      // Removing extant node.
      if (!npReturn->mayCreateRemove)
	{
	  // Parent is locked, couldn't change child list
	  printf("parent locked\n");
	  return ENCHANT_LOCKED;
	}
      // if the node isn't hard published, don't hard unpublish it
      // ???? hard unpublication of soft node with hard descendants?
      if (!SchemaOf(npReturn->existNode)->hardNode)
	hard = 0;
      // Go up through implicit parents, but don't change report.
      while (IsImplicit(npReturn->existParent) &&
	     ArrayLength(npReturn->existParent->data.branch_val) == 1)
	{
	  target = ExtractParentUNL(target);
	  npReturn->existNode = npReturn->existParent;
	  npReturn->existParent = ParentOf(npReturn->existParent);
	}
      UnlinkNode(npReturn->existParent, npReturn->existNode, 
		 target, clientID);

      if (hard)
	HardPublish(NULL, target);
      return ENCHANT_SUCCESS;
    }
  else
    {
      // Node doesn't exist.

      ret = mayPublish(clientID, dataNode, npReturn, hard);
      if (ret != ENCHANT_SUCCESS)
	return ret;

      if (!dataNode)
	{
	  // It's not there, as you requested...
	  return ENCHANT_NO_EFFECT;
	}

      // Otherwise, operation doesn't matter, since it's new

      // The closest existing ancestor must be modified; it must not
      // be locked, and its ancestors must not be locked recursively.
      if (!npReturn->mayCreateRemove)
	{
	  // Path was locked
	  return ENCHANT_LOCKED;
	}

      CreateAncestors(npReturn, clientID);
      dataNode->name = 
	GetUNLHead(DescendsFrom(npReturn->closestUNL, target));
      if (!LinkNode(npReturn->closestAncestorNode, dataNode, target, 
		    clientID, hard ? LINK_HARD : 0))
	{
	  // npReturn->closestAncestorNode isn't a branch.
	  printf("Error\n");
	  return ENCHANT_ERROR;
	}

      // XXXX mark created nodes as implicit

      if (hard)
	{
	  HardPublish(dataNode, target);
	}
      return ENCHANT_SUCCESS;
    }
}

int AddNodeToTree(ClientIDType *clientID, NodeType *dataNode,
		  NodePathReturnType *npReturn, const char *op,
		  int hard, int publish_if_locked)
{
  int ret;
  if (dataNode)
    dataNode->name = ExtractUNLEndNode(npReturn->UNL);
  ret = doPublish(clientID, dataNode, npReturn, op, hard, publish_if_locked);
  if (ret == ENCHANT_SUCCESS)
    {
      InformSubscribers(clientID, npReturn, NULL, NULL);
    }
  if (ret == ENCHANT_SENT)
    {
      int sent = 0;
      sent += SendUpdates(npReturn->targetSchema->subscribers,
			  dataNode, npReturn->UNL, clientID, 
			  npReturn->reportLock->owner);
      sent += InformSubscribers(clientID, npReturn, npReturn->existNode,
				dataNode);
      if (!sent)
	ret = ENCHANT_LOCKED;
    }
  return ret;
}

void PublishLocal(const char *path, NodeType *insertNode, int hard)
{
  /* This uses MERGE mode */
  /* Does not dereference symlinks */
  NodePathReturnType npReturn;
  if (!path && !g_globalNode)
    {
      g_globalNode = insertNode;
      LinkNode(NULL, g_globalNode, NULL, serverID, 0);
      return;
    }
  npReturn.UNL = path;
  npReturn.follow_symlinks = 0;
  npReturn.cross_symlinks = 0;
  npReturn.getSubscriptions = 1;
  GetNodePathInfo(serverID, &npReturn, 0);
  if (AddNodeToTree(serverID, insertNode, &npReturn, kv_replace, hard, 1) !=
      ENCHANT_SUCCESS)
    {
      printf("Local %s failed!\n", insertNode ? "Publish" : "Unpublish");
      FreeNode(insertNode);
      return;
    }
}

int LocalSetPathPermissions(const char *path, char type, 
			    ClientSelector *selector)
{
  NodePathReturnType npReturn;
  npReturn.UNL = path;
  npReturn.follow_symlinks = 0;
  npReturn.cross_symlinks = 0;
  npReturn.getSubscriptions = 0;
  GetNodePathInfo(serverID, &npReturn, 1);
  if (!npReturn.targetSchema)
    return 0;
  SetPermissions(npReturn.targetSchema, type, selector);
  return 1;
}

void RemoveClientTreeEffects(ClientIDType *clientID)
{
  SubscriptionType *subscription;
  int i, iLen;
  if (clientID->subscribedNodes)
    {
      iLen = ArrayLength(clientID->subscribedNodes);
      for(i = 0; i < iLen; i++)
	{
	  subscription = 
	    *((SubscriptionType **) ArrayNth(clientID->subscribedNodes, i));
	  ForgetSubscriber(subscription);
	}
      ArrayFree(clientID->subscribedNodes);
    }
  RemoveLocksByClient(clientID);
}

static NodeType *CreateClientSelectorNode(const char *name,
					  ClientSelector *selector)
{
  NodeType *ret;
  int i;
  if (!selector->tokens)
    return NULL;
  
  ret = CreateBranchNode(name, T_BRANCH);
  for (i = 0; selector->tokens[i]; i++)
    {
      InsertNode(ret, CreateStringNode(InstanceNumber(i),
				       T_AUTH,
				       (char *) selector->tokens[i]));
    }
  return ret;
}

NodeType *GetMetadataTree(SchemaNode *node, const char *metapath)
{
  const char *head = GetUNLHead(metapath);
  if (!*head)
    {
      NodeType *ret = CreateBranchNode("", T_BRANCH);
      InsertNode(ret, GetMetadataTree(node, "/fixed"));
      InsertNode(ret, GetMetadataTree(node, "/symlink"));
      InsertNode(ret, GetMetadataTree(node, "/distinct"));
      InsertNode(ret, GetMetadataTree(node, "/Permissions"));
      InsertNode(ret, GetMetadataTree(node, "/persistant"));
      return ret;
    }
  else if (!strcmp(head, "Permissions"))
    {
      const char *next = GetUNLTail(metapath);
      if (!next)
	{
	  NodeType *ret = CreateBranchNode(head, T_BRANCH);
	  InsertNode(ret, GetMetadataTree(node, "/Permissions/Read"));
	  InsertNode(ret, GetMetadataTree(node, "/Permissions/Write"));
	  InsertNode(ret, GetMetadataTree(node, "/Permissions/Lock"));
	  InsertNode(ret, GetMetadataTree(node, "/Permissions/Admin"));
	  return ret;
	}
      head = GetUNLHead(next);
      if (!strcmp(head, "Read"))
	{
	  return CreateClientSelectorNode(head, &node->canRead);
	}
      else if (!strcmp(head, "Write"))
	{
	  return CreateClientSelectorNode(head, &node->canWrite);
	}
      else if (!strcmp(head, "Lock"))
	{
	  return CreateClientSelectorNode(head, &node->canLock);
	}
      else if (!strcmp(head, "Admin"))
	{
	  return CreateClientSelectorNode(head, &node->canAdmin);
	}
    }
  else if (!strcmp(head, "fixed"))
    {
      if (!node->type)
	return NULL;
      return CreateTypeNode(head, T_TYPE, node->type);
    }
  else if (!strcmp(head, "symlink"))
    {
      if (!node->symlink)
	return NULL;
      return CreateStringNode(head, T_UNL, (char *) node->symlink);
    }
  else if (!strcmp(head, "distinct"))
    {
      if (!node->distinct)
	return NULL;
      return CreateBooleanNode(head, T_BOOL, 1);
    }
  else if (!strcmp(head, "persistant"))
    {
      return NULL;
    }
  return NULL;
}

int SetMetadataTree(SchemaNode *node, const char *metapath, char *op,
		    NodeType *value)
{
  const char *head = GetUNLHead(metapath);
  if (!strcmp(head, "fixed"))
    {
      if (value->type != T_TYPE)
	return ENCHANT_TYPE;
      node->type = value->data.type_val;
      return ENCHANT_SUCCESS;
    }
  if (!strcmp(head, "distinct"))
    {
      if (value && value->type != T_BOOL)
	return ENCHANT_TYPE;
      node->distinct = value && value->data.int_val;
      return ENCHANT_SUCCESS;
    }
  if (!strcmp(head, "symlink"))
    {
      if (value)
	{
	  if (value->type != T_UNL)
	    return ENCHANT_TYPE;
	  node->symlink = InstanceString(value->data.string_val);
	}
      else
	{
	  node->symlink = NULL;
	}
      return ENCHANT_SUCCESS;
    }
  if (!strcmp(head, "Permissions"))
    {
      const char *next = GetUNLTail(metapath);
      ClientSelector *target;
      if (!next)
	{
	  return ENCHANT_ERROR;
	}
      head = GetUNLHead(next);
      if (!strcmp(head, "Read"))
	{
	  target = &node->canRead;
	}
      else if (!strcmp(head, "Write"))
	{
	  target = &node->canWrite;
	}
      else if (!strcmp(head, "Lock"))
	{
	  target = &node->canLock;
	}
      else if (!strcmp(head, "Admin"))
	{
	  target = &node->canAdmin;
	}
      else
	{
	  return ENCHANT_NOT_EXIST;
	}
      if (!value)
	SelectorAny(target);
      else
	{
	  int i = 0;
	  NodeType *tokenNode;
	  target->tokens = malloc(sizeof(char *) *
				  (ArrayLength(value->data.branch_val) + 1));
	  for (i = 0; (tokenNode = BranchIndex(value, i)); i++)
	    {
	      target->tokens[i] = InstanceString(tokenNode->data.string_val);
	    }
	  target->tokens[i] = NULL;
	}
      return ENCHANT_SUCCESS;
    }
  return ENCHANT_ERROR;
}
