/*************************************************
 * File: test.c
 * Author: Daniel Barkalow
 * Birth Date: 12/30/02
 *
 * Project: Enchantment IPC
 * 
 * Based on Enchantment IPC Whitepaper
 *   by Edward Keyes.
 * 
 * Copyright (C) 2002 MIT Media Lab,
 *               2004 AWare Technologies, Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *************************************************/

#include <enchantment/types.h>
#include <enchantment/node.h>
#include <enchantment/whiteboard.h>
#include <enchantment/unl.h>
#include <enchantment/dstring.h>

#include <getopt.h>
#include <errno.h>
#include <unistd.h>
#include <sys/socket.h>

#include <stdlib.h>
#include <string.h>

#include "test-utils.h"

int cleanup = 0;
int noclean = 0;
int failures = 0;
char *server;
char *head = "/Testing";

int subclient; // socket to other client
int issubclient = 0;

#define HEAD "/Testing"
#define BR1 "/Foo"
#define BR2 "/Baz"
#define LF1 "/bar"
#define LF2 "/quux"
#define STR1 "test1"
#define STR2 "test2"

void clean()
{
  const char *unl;
  if (cleanup && !noclean && !issubclient)
    {
      unl = CreateUNL(server, HEAD);
      if (!PublishNode(unl, NULL, NULL))
	{
	  fprintf(stderr, "Error cleaning up %s\n", HEAD);
	}
    }
}

void done()
{
  if (!issubclient)
    {
      clean();
      write(subclient, "turn", 5);
    }
  // Note: this is for testing, not because it is actually necessary.
  ShutDownClientIPC();
}

void checkNode(int cond, NodeType *node, char *format, ...)
{
  va_list args;
  va_start(args, format);
  vcheck(cond, format, args);
  if (!cond)
    {
      // XXXX wrong place
      printf("Instead, node is:\n");
      PrintNode(node, NULL);
    }
  va_end(args);
}

void requireUNL(const char *got, const char *right, char *format, ...)
{
  va_list args;
  va_start(args, format);
  vrequire((!got == !right) && !strcmp(got, right),
	   format, args);
  va_end(args);
}

void checkUNL(const char *got, const char *right, char *format, ...)
{
  va_list args;
  va_start(args, format);
  vcheck((!got == !right) && !strcmp(got, right),
	 format, args);
}

void requireNode(int cond, NodeType *node, char *format, ...)
{
  va_list args;
  va_start(args, format);
  if (cond)
    {
      if (format[0] != '?')
	{
	  fprintf(stderr, "[pass] ");
	  vfprintf(stderr, format, args);
	}
      va_end(args);
    }
  else
    {
      fprintf(stderr, "[FAIL] ");
      if (format[0] == '?')
	format++;
      vfprintf(stderr, format, args);
      va_end(args);
      // XXXX wrong place
      printf("Instead, node is:");
      PrintNode(node, NULL);
      done();
      exit(1);
    }
}

void reqGetNode(const char *unl, NodeType **nodep)
{
  *nodep = GetNode(unl, NULL);
  if (!nodep)
    {
      fprintf(stderr, "[FAIL] get %s\n", unl);
      done();
      exit(1);
    }
}

/********************** Tests follow **************************/

void testUNL()
{
  char *complete = "server:/path";
  char *no_slash = "server:path";
  char *no_server = "/path";
  char *word = "path";
  char *mid_slash = "path/path";
  char *local = ":/local";

  info("Testing UNL\n");

  require(!strcmp(ExtractUNLServer(complete), "server"),
	  "correct server for %s\n", complete);
  require(!strcmp(ExtractUNLServer(no_slash), "server"),
	  "correct server for %s\n", no_slash);
  require(!ExtractUNLServer(no_server),
	  "correct server for %s\n", no_server);
  require(!ExtractUNLServer(word),
	  "correct server for %s\n", word);
  require(!ExtractUNLServer(mid_slash),
	  "correct server for %s\n", mid_slash);
  require(!strcmp(ExtractUNLServer(local), ""),
	  "correct server for %s\n", local);

  require(!strcmp(ExtractUNLPath(complete), "/path"),
	  "correct path for %s\n", complete);
  require(!strcmp(ExtractUNLPath(no_slash), "/path"),
	  "correct path for %s\n", no_slash);
  require(!strcmp(ExtractUNLPath(no_server), "/path"),
	  "correct path for %s\n", no_server);
  require(!strcmp(ExtractUNLPath(word), "/path"),
	  "correct path for %s\n", word);
  require(!strcmp(ExtractUNLPath(mid_slash), "/path/path"),
	  "correct path for %s\n", mid_slash);
  require(!strcmp(ExtractUNLPath(local), "/local"),
	  "correct path for %s\n", local);
}

void testPublish()
{
  const char *unl;
  NodeType *node, *subnode;
  PublishNodeInfo pubInfo;
  info("Testing publication\n");

  node = CreateStringNode("bar", T_STRING, STR1);
  unl = CreateUNL(server, HEAD BR1 LF1);
  PublishNode(NULL, NULL, &pubInfo);
  pubInfo.common.operationFlags |= ENCHANT_PUBLISH_MERGE;
  require(PublishNode(unl, node, &pubInfo), 
	  "Publish a string to %s\n", unl);
  reqGetNode(unl, &node);
  requireNode(node->type == T_STRING && 
	      !strcmp(node->data.string_val, STR1),
	      node,
	      "%s is a string with value %s\n", 
	      unl, STR1);
  unl = CreateUNL(server, HEAD BR1);
  reqGetNode(unl, &node);
  requireNode(node->type == T_BRANCH &&
	      BranchSize(node) == 1,
	      node,
	      "%s is a branch with one child\n", unl);
  node = CreateBranchNode("Foo", T_BRANCH);
  subnode = CreateStringNode("quux", T_STRING, STR2);
  InsertNode(node, subnode);
  require(PublishNode(unl, node, &pubInfo),
	  "Publish a branch to %s in merge mode\n", unl);
  reqGetNode(unl, &node);
  requireNode(node->type == T_BRANCH &&
	      BranchSize(node) == 2,
	      node,
	      "%s is now a branch with two children\n", unl);
  unl = CreateUNL(server, HEAD BR1 LF1);
  reqGetNode(unl, &node);
  requireNode((node->type == T_STRING) && 
	      !strcmp(node->data.string_val, STR1),
	      node,
	      "%s is still a string with value %s\n", 
	      unl, STR1);
  unl = CreateUNL(server, HEAD BR1 LF2);
  reqGetNode(unl, &node);
  requireNode((node->type == T_STRING) && 
	      !strcmp(node->data.string_val, STR2),
	      node,
	      "%s is a string with value %s\n", 
	      unl, STR2);
  node = CreateBranchNode("Foo", T_BRANCH);
  subnode = CreateStringNode("quux", T_STRING, STR1);
  InsertNode(node, subnode);
  unl = CreateUNL(server, HEAD BR1);
  pubInfo.common.operationFlags = ENCHANT_PUBLISH_REPLACE;
  require(PublishNode(unl, node, &pubInfo),
	  "Publish a branch to %s in replace mode\n", unl);
  reqGetNode(unl, &node);
  requireNode((node->type == T_BRANCH) &&
	      BranchSize(node) == 1,
	      node,
	      "%s is now a branch with one child\n", unl);
  unl = CreateUNL(server, HEAD BR1 LF1);
  require(!GetNode(unl, NULL), "%s does not exist\n", unl);
  unl = CreateUNL(server, HEAD BR1 LF2);
  reqGetNode(unl, &node);
  requireNode((node->type == T_STRING) && 
	      !strcmp(node->data.string_val, STR1),
	      node,
	      "%s is a string with value %s\n", 
	      unl, STR1);
  node = CreateBranchNode("Foo", T_BRANCH);
  subnode = CreateStringNode("quux", T_STRING, STR2);
  InsertNode(node, subnode);
  subnode = CreateStringNode("bar", T_STRING, STR1);
  InsertNode(node, subnode);
  unl = CreateUNL(server, HEAD BR1);
  pubInfo.common.operationFlags = ENCHANT_PUBLISH_ADDONLY;
  require(PublishNode(unl, node, &pubInfo),
	  "Publish a branch to %s in addonly mode\n", unl);
  reqGetNode(unl, &node);
  requireNode((node->type == T_BRANCH) &&
	      BranchSize(node) == 2,
	      node,
	      "%s is now a branch with two children\n", unl);
  unl = CreateUNL(server, HEAD BR1 LF1);
  reqGetNode(unl, &node);
  requireNode((node->type == T_STRING) && 
	      !strcmp(node->data.string_val, STR1),
	      node,
	      "%s is a string with value %s\n", 
	      unl, STR1);
  unl = CreateUNL(server, HEAD BR1 LF2);
  reqGetNode(unl, &node);
  requireNode((node->type == T_STRING) && 
	      !strcmp(node->data.string_val, STR1),
	      node,
	      "%s is still a string with value %s\n", 
	      unl, STR1);
  unl = CreateUNL(server, HEAD BR1);
  require(PublishNode(unl, NULL, NULL), "Unpublish %s\n", unl);
  require(!GetNode(unl, NULL), "%s does not exist\n", unl);
  unl = CreateUNL(server, HEAD);
  reqGetNode(unl, &node);
  requireNode((node->type == T_BRANCH) && 
	      BranchSize(node) == 0, node,
	      "%s is an empty branch\n", unl);
  unl = CreateUNL(server, HEAD BR1);
  node = CreateEmptyNode("Notes", T_EMPTY);
  pubInfo.common.operationFlags = ENCHANT_PUBLISH_MERGE;
  require(PublishNode(unl, node, &pubInfo),
	  "Publish an empty node to %s in merge mode\n", unl);
  node = CreateBranchNode("Foo", T_BRANCH);
  subnode = CreateStringNode("quux", T_STRING, STR2);
  InsertNode(node, subnode);
  subnode = CreateStringNode("bar", T_STRING, STR1);
  InsertNode(node, subnode);
  require(PublishNode(unl, node, &pubInfo),
	  "Publish a branch to %s in merge mode\n", unl);
  require(PublishNode(unl, node, &pubInfo),
	  "Publish a branch to %s in merge mode again\n", unl);
  require(PublishNode(unl, NULL, NULL), "Unpublish %s\n", unl);
}

void testSubscribe()
{
  const char *unl;
  const char *rightUNL;
  NodeType *node;
  NodeType *branch;

  PublishNodeInfo pubInfo;
  SubscribeUNLInfo subInfo;
  GetUpdateInfo upInfo;

  info("Testing subscription\n");

  SubscribeUNL(NULL, &subInfo);
  subInfo.common.operationFlags = ENCHANT_SUBSCRIBE_SELF;

  unl = CreateUNL(server, HEAD BR1);
  node = CreateBranchNode("Foo", T_BRANCH);
  require(PublishNode(unl, node, NULL),
	  "publish %s\n", unl);
  require(SubscribeUNL(unl, NULL), 
	  "Subscribe to extant %s without self\n", unl);
  reqGetNode(unl, &node);
  require((node->type == T_BRANCH) &&
	  BranchSize(node) == 0, "Existing node is right\n");
  require(PublishNode(unl, node, NULL), 
	  "Publish to subscribed node\n");
  GetUpdate(&upInfo, NULL);
  requireNode(upInfo.publisher == NULL && upInfo.node == NULL,
	      upInfo.node,
	      "no update received without self-subscribe\n");

  unl = CreateUNL(server, HEAD BR1);
  require(PublishNode(unl, NULL, NULL),
	  "Unpublish %s\n", unl);
  GetUpdate(&upInfo, NULL);
  requireNode(upInfo.publisher == NULL && upInfo.node == NULL,
	      upInfo.node,
	      "no update received without self-subscribe\n");

  unl = CreateUNL(server, HEAD BR1);
  
  require(UnsubscribeUNL(unl, NULL), "Unsub from %s\n", unl);

  require(SubscribeUNL(unl, &subInfo), "Subscribe to missing %s\n", unl);
  GetUpdate(&upInfo, NULL);
  requireNode(upInfo.publisher == NULL && upInfo.node == NULL,
	      upInfo.node,
	      "no update received from %s to %s\n", upInfo.publisher,
	      upInfo.UNL);
  unl = CreateUNL(server, HEAD BR1);
  node = CreateBranchNode("Foo", T_BRANCH);
  require(PublishNode(unl, node, NULL),
	  "publish %s with mode replace\n", unl);
  rightUNL = unl;
  GetUpdate(&upInfo, NULL);
  requireUNL(upInfo.UNL, rightUNL, "?got correct update\n");
  checkNode(upInfo.node && (upInfo.node->type == T_BRANCH) &&
	    BranchSize(upInfo.node) == 0, node, 
	    "got correct update\n");
  GetUpdate(&upInfo, NULL);
  requireNode(upInfo.publisher == NULL && upInfo.node == NULL,
	      upInfo.node,
	      "no duplicate update received\n");

  unl = CreateUNL(server, HEAD BR1);
  node = CreateBranchNode("Foo", T_BRANCH);
  require(PublishNode(unl, node, NULL),
	  "publish %s with mode replace again\n", unl);
  node = CreateStringNode("bar", T_STRING, STR2);
  unl = CreateUNL(server, HEAD BR1 LF1);

  require(SubscribeUNL(unl, &subInfo), "Subscribe to child %s\n", unl);

  PublishNode(NULL, NULL, &pubInfo);
  pubInfo.common.operationFlags = ENCHANT_PUBLISH_MERGE;
  require(PublishNode(unl, node, &pubInfo), 
	  "Publish a string to %s in merge mode\n", unl);
  
  rightUNL = unl;
  GetUpdate(&upInfo, unl);
  checkNode(upInfo.UNL && !strcmp(upInfo.UNL, rightUNL) && 
	    (upInfo.node->type == T_STRING), upInfo.node, 
	    "scan got matching update past non-matching update\n");
  GetUpdate(&upInfo, unl);
  requireNode(upInfo.publisher == NULL,
	      upInfo.node,
	      "non-matching update not returned\n");
  unl = CreateUNL(server, HEAD BR1);
  rightUNL = unl;
  GetUpdate(&upInfo, NULL);
  checkUNL(upInfo.UNL, rightUNL, "?got passed update 1\n");
  checkNode(upInfo.UNL && !strcmp(upInfo.UNL, rightUNL) && 
	    (upInfo.node->type == T_BRANCH) &&
	    BranchSize(upInfo.node) == 0, upInfo.node, 
	    "got passed update 1\n");
  GetUpdate(&upInfo, NULL);
  rightUNL = CreateUNL(server, HEAD BR1 LF1);
  checkUNL(upInfo.UNL, rightUNL, "?got passed update 2\n");
  checkNode(upInfo.UNL && !strcmp(upInfo.UNL, rightUNL) && 
	    (upInfo.node->type == T_BRANCH) &&
	    BranchSize(upInfo.node) == 1, upInfo.node, 
	    "got passed update 2\n");
  GetUpdate(&upInfo, NULL);
  requireNode(upInfo.publisher == NULL && upInfo.node == NULL,
	      upInfo.node, "no duplicate update received\n");

  unl = CreateUNL(server, HEAD BR1 LF1);
  require(UnsubscribeUNL(unl, NULL), "Unsubscribe from child %s\n", unl);

  unl = CreateUNL(server, HEAD BR1);
  node = CreateBranchNode("Foo", T_BRANCH);
  require(PublishNode(unl, node, &pubInfo),
	  "republish %s with mode merge\n", unl);
  rightUNL = unl;
  GetUpdate(&upInfo, NULL);
  checkUNL(upInfo.UNL, rightUNL, "?got correct update again\n");
  checkNode(upInfo.UNL && !strcmp(upInfo.UNL, rightUNL) &&
	    (upInfo.node->type == T_BRANCH) &&
	    BranchSize(upInfo.node) == 1, upInfo.node, 
	    "got correct update again\n");

  unl = CreateUNL(server, HEAD BR2);
  node = CreateBranchNode("Foo", T_BRANCH);
  require(PublishNode(unl, node, NULL),
	  "publish %s with mode replace\n", unl);
  GetUpdate(&upInfo, NULL);
  requireNode(upInfo.publisher == NULL && upInfo.node == NULL,
	      upInfo.node,
	      "no misguided update received\n");
  node = CreateStringNode("bar", T_STRING, STR2);
  unl = CreateUNL(server, HEAD BR1 LF1);
  require(PublishNode(unl, node, &pubInfo), 
	  "Publish a string to %s\n", unl);
  rightUNL = CreateUNL(server, HEAD BR1 LF1);
  unl = NULL;
  GetUpdate(&upInfo, NULL);
  checkUNL(upInfo.UNL, rightUNL, "?got correct update");
  checkNode(upInfo.node && (upInfo.node->type == T_BRANCH) &&
	    BranchSize(upInfo.node) == 1, 
	    upInfo.node,
	    "got correct update\n");

  node = CreateStringNode("bar", T_STRING, STR2);
  unl = CreateUNL(server, HEAD BR1 LF1);
  require(PublishNode(unl, node, &pubInfo), 
	  "Publish a string to %s in merge mode\n", unl);
  rightUNL = CreateUNL(server, HEAD BR1 LF1);
  unl = NULL;
  GetUpdate(&upInfo, NULL);
  checkUNL(upInfo.UNL, rightUNL,
	   "?got correct update when only a child changed\n");
  checkNode(upInfo.node && (upInfo.node->type == T_BRANCH) &&
	    BranchSize(upInfo.node) == 1, 
	    upInfo.node,
	    "got correct update when only a child changed\n");

  unl = CreateUNL(server, HEAD BR2);
  require(PublishNode(unl, NULL, NULL),
	  "Unpublish %s\n", unl);
  GetUpdate(&upInfo, NULL);
  requireNode(upInfo.publisher == NULL && upInfo.node == NULL,
	      upInfo.node,
	      "no misguided update received\n");

  unl = CreateUNL(server, HEAD BR1);
  require(PublishNode(unl, NULL, NULL),
	  "Unpublish %s\n", unl);
  rightUNL = CreateUNL(server, HEAD BR1);
  unl = NULL;
  GetUpdate(&upInfo, NULL);
  checkUNL(upInfo.UNL, rightUNL, "got correct update\n");
  check(upInfo.node == NULL,
	"got correct update\n");
  unl = rightUNL;
  require(UnsubscribeUNL(unl, NULL), "Unsub from %s\n", rightUNL);

  unl = CreateUNL(server, HEAD BR1 LF1);
  node = CreateStringNode("bar", T_STRING, STR2);
  require(PublishNode(unl, node, &pubInfo), 
	  "Publish a string to %s\n", unl);
  require(SubscribeUNL(unl, &subInfo), "Subscribe to %s\n", unl);

  branch = CreateBranchNode("Foo", T_BRANCH);
  node = CreateStringNode("bar", T_STRING, STR1);
  InsertNode(branch, node);
  unl = CreateUNL(server, HEAD BR1);

  require(PublishNode(unl, branch, &pubInfo),
	  "Merge %s\n", unl);
  rightUNL = CreateUNL(server, HEAD BR1 LF1);
  unl = NULL;
  GetUpdate(&upInfo, NULL);

  checkNode(upInfo.UNL && !strcmp(upInfo.UNL, rightUNL) &&
	    (upInfo.node->type == T_STRING) &&
	    !strcmp(upInfo.node->data.string_val, STR1),
	    upInfo.node,
	    "got correct update; %s should be %s\n",
	    upInfo.UNL, rightUNL);

  unl = CreateUNL(server, HEAD BR1);
  node = CreateStringNode("bar", T_STRING, STR2);
  require(PublishNode(unl, node, NULL),
	  "Replace %s\n", unl);
  rightUNL = CreateUNL(server, HEAD BR1);
  unl = NULL;
  GetUpdate(&upInfo, NULL);

  rightUNL = CreateUNL(server, HEAD BR1 LF1);
  checkUNL(upInfo.UNL, rightUNL, "got correct update\n");
  check(upInfo.node == NULL,
	"got correct update\n");
  unl = CreateUNL(server, HEAD BR1 LF1);
  require(UnsubscribeUNL(unl, NULL), "Unsubscribe from %s\n", unl);
}

void testLock()
{
  const char *unl;
  NodeType *node;
  
  LockUNLInfo lockInfo;
  
  info("Testing locking\n");
  node = CreateBranchNode("Foo", T_BRANCH);
  unl = CreateUNL(server, HEAD BR1);
  require(PublishNode(unl, NULL, NULL), "Unpublish %s\n", unl);
  LockUNL(NULL, &lockInfo);
  lockInfo.common.operationFlags = ENCHANT_LOCK_RECURSIVE;
  require(LockUNL(unl, &lockInfo),
	  "Can lock non-existant %s recursive\n", unl);
  lockInfo.common.operationFlags = ENCHANT_LOCK_NONE;
  require(LockUNL(unl, &lockInfo),
	  "Can unlock non-existant %s\n", unl);
  lockInfo.common.operationFlags = ENCHANT_LOCK_NORMAL;
  require(LockUNL(unl, &lockInfo),
	  "Can lock non-existant %s non-recursive\n", unl);
  lockInfo.common.operationFlags = ENCHANT_LOCK_NONE;
  require(LockUNL(unl, &lockInfo),
	  "Can unlock non-existant %s\n", unl);
  require(PublishNode(unl, node, NULL),
	  "Publish branch to %s\n", unl);
  lockInfo.common.operationFlags = ENCHANT_LOCK_RECURSIVE;
  require(LockUNL(unl, &lockInfo),
	  "Lock %s recursive\n", unl);
  require(!!GetNode(unl, NULL),
	  "Self can read %s\n", unl);
  turn("Other can read %s\n", unl);
  turn("Other sees %s is locked\n", unl);
  turn("Other cannot publish %s\n", unl);
  unl = CreateUNL(server, HEAD BR1 LF1);
  require(GetUNLLock(unl, NULL), "Owner sees %s locked\n", unl);
  turn("Other cannot publish %s\n", unl);
  node = CreateStringNode("bar", T_STRING, STR1);
  require(PublishNode(unl, node, NULL),
	  "Publish string to %s\n", unl);
  turn("Other still cannot publish %s\n", unl);
  reqGetNode(unl, &node);
  requireNode((node->type == T_STRING) && 
	      !strcmp(node->data.string_val, STR1),
	      node,
	      "%s is still a string with value %s\n", 
	      unl, STR1);
  unl = CreateUNL(server, HEAD BR1);
  lockInfo.common.operationFlags = ENCHANT_LOCK_NONE;
  require(LockUNL(unl, &lockInfo), "Unlock %s\n", unl);
  unl = CreateUNL(server, HEAD BR1 LF1);
  turn("Other can now publish %s\n", unl);
  reqGetNode(unl, &node);
  requireNode((node->type == T_STRING) && 
	      !strcmp(node->data.string_val, STR2),
	      node,
	      "%s is now a string with value %s\n", 
	      unl, STR2);
  unl = CreateUNL(server, HEAD BR1);
  require(LockUNL(unl, NULL), "Lock %s non-recursive\n", unl);
  unl = CreateUNL(server, HEAD BR1 LF1);
  turn("Other can still publish %s\n", unl);
  unl = CreateUNL(server, HEAD BR1);
  require(PublishNode(unl, NULL, NULL),
	  "Owner can remove locked %s\n", unl);
}

void subtestLock()
{
  const char *unl;
  NodeType *node;
  PublishNodeInfo pubInfo;
  GetUNLLockInfo getLockInfo;

  info("Testing locking\n");
  unl = CreateUNL(server, HEAD BR1);
  require(!!GetNode(unl, NULL), "Other can read %s\n");
  GetUNLLock(NULL, &getLockInfo);
  check(GetUNLLock(unl, &getLockInfo) && getLockInfo.lockType,
	"Other sees %s is locked\n");

  node = CreateBranchNode("Foo", T_BRANCH);
  require(!PublishNode(unl, node, NULL),
	  "Other cannot publish %s\n");
  unl = CreateUNL(server, HEAD BR1 LF1);
  node = CreateStringNode("bar", T_STRING, STR2);
  require(!PublishNode(unl, node, NULL),
	  "Other cannot publish %s\n");
  require(!PublishNode(unl, node, NULL),
	  "Other still cannot publish %s\n");
  require(PublishNode(unl, node, NULL),
	  "Other can now publish %s\n");
  node = CreateStringNode("bar", T_STRING, STR1);
  PublishNode(NULL, NULL, &pubInfo);
  pubInfo.common.operationFlags = ENCHANT_PUBLISH_MERGE;
  require(PublishNode(unl, node, &pubInfo), "Other can still publish %s\n");
}

void subtestUNL()
{
  info("Testing UNL\n");
}

void subtestPublish()
{
  info("Testing publication\n");
}

void subtestSubscribe()
{
  info("Testing subscription\n");
}

test tests[] =
{
  {"unl", testUNL, subtestUNL},
  {"publish", testPublish, subtestPublish},
  {"subscribe", testSubscribe, subtestSubscribe},
  {"lock", testLock, subtestLock},
  {NULL, NULL, NULL}
};

int main(int argc, char **argv)
{
  const char *unl;
  NodeType *node;
  int opt;
  int anyway = 0;
  int i;
  char *cases = NULL;

  do
    {
      opt = getopt(argc, argv, "+uac:");
      if (opt == 'a')
	anyway = 1;
      if (opt == 'u')
	noclean = 1;
      if (opt == 'c')
	cases = optarg;
    }
  while (opt > 0);

  if (argc > optind)
    server = argv[optind];
  else
    server = "localhost";

  splitSubclient(cases);

  // Our test suite is as follows:
  // check for /Testing; if it is there, exit with no result (stuff in the way)
  // publish /Testing as a branch
  // check for /Testing, must be an empty branch
  // publish /Testing/Foo/bar as a string
  // check for /Testing/Foo/bar, must be a string, must be the right string
  // check for /Testing/Foo, must be a branch with one child
  // publish /Testing/Foo as a branch[quux:string], merge
  // check for /Testing/Foo/bar, must be the same
  // check for /Testing/Foo/quux, must be a string, must be the right string
  // publish /Testing/Foo as a branch[quux:string2], replace
  // check for /Testing/Foo: 1 child
  // check for /Testing/Foo/bar: doesn't exist
  // check for /Testing/Foo/quux, must be the new string
  // publish /Testing/Foo as a branch[bar:string, quux:string], addonly
  // check for /Testing/Foo: 2 children
  // check for /Testing/Foo/bar, must be the right string
  // check for /Testing/Foo/quux, must be the old string

  unl = CreateUNL(server, "/");
  require(!!GetNode(unl, NULL), "Root is available\n");
  unl = CreateUNL(server, HEAD);
  if (!anyway)
    {
      require(!GetNode(unl, NULL), 
	      "%s doesn't exist beforehand\n", HEAD);
    }
  cleanup = 1;
  node = CreateBranchNode("Testing", T_BRANCH);
  require(PublishNode(unl, node, NULL),
	  "Initial publication of %s\n", unl);
  reqGetNode(unl, &node);
  requireNode((node->type == T_BRANCH) && BranchSize(node) == 0,
	      node,
	      "%s is an empty branch\n", unl);

  i = 0;
  while (tests[i].name)
    {
      if (!cases || strstr(cases, tests[i].name))
	(*tests[i].main)();
      i++;
    }

  done();
  if (!failures)
    printf("Everything passes.\n");
  else
    printf("Had %d non-terminal failures\n", failures);
  return 0;
}
