/*****************************************************************************
 *
 * Author: Daniel Barkalow, MIT Wearable Computing Group
 * http://www.media.mit.edu/wearables/
 * Copyright (C) 2003 MIT Media Lab, 
 *               2004 AWare Technologies, Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *****************************************************************************/

#include <stdlib.h>
#include <enchantment/whiteboard.h>
#include <enchantment/signals.h>
#include <enchantment/time-utils.h>
#include <getopt.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <string.h>

#define USE_THREADS

#ifdef USE_THREADS
#include <pthread.h>
#include <semaphore.h>
#endif

#define CERT_FILE "cert.pem"
#define KEY_FILE "key.pem"

#ifdef USE_THREADS
pthread_mutex_t fileLock;
pthread_mutex_t ipcLock;
#define lock(name) pthread_mutex_lock(&name)
#define unlock(name) pthread_mutex_unlock(&name)
#else
#define lock(name)
#define unlock(name)
#endif

/*
inline long time_diff(struct timeval *t1, struct timeval *t2)
{
  return (t1->tv_sec - t2->tv_sec) * 1000000 + (t1->tv_usec - t2->tv_usec);
}

inline void time_add(struct timeval *time, signed long usecs)
{
  time->tv_usec += usecs;
  while (time->tv_usec >= 1000000)
    {
      time->tv_sec++;
      time->tv_usec -= 1000000;
    }
}
*/

void usage(char *arg0)
{
  fprintf(stderr, "Usage: %s [OPTION]...\n", arg0);
  fprintf(stderr, 
	  "Perform various buffering and replay operations on signals.\n\n");
  fprintf(stderr, "  -i [UNL]	read from [UNL]\n");
  fprintf(stderr, "  -o [UNL]	write to [UNL]\n");
  fprintf(stderr, "  -r [FILE]	read from [FILE]\n");
  fprintf(stderr, "  -w [FILE]	write to [FILE]\n");
  fprintf(stderr, "  -a [FILE]	append to [FILE]\n");
  fprintf(stderr, 
	  "  -d		when writing to a file, put all the info with each\n");
  fprintf(stderr, 
	  "		datagram, so the file can be split easily\n");
  fprintf(stderr, "  -c [AUTH]	hold data if there is no connected reader;\n");
  fprintf(stderr, "           	when reading, send receipts after the data\n");
  fprintf(stderr, "           	is committed.\n");
  fprintf(stderr, "  -t [FILE]	use [FILE] for temporary storage if needed\n");
  fprintf(stderr, "  -s [SVR]	make [SVR] the default enchantment server\n");
  fprintf(stderr, "  -f       	fast; ignore original speed\n");
  fprintf(stderr, "  -p         proxy through the server\n");
  //fprintf(stderr, "  -n       	shift the timestamps to start now.\n");
  fprintf(stderr, "\n");
  fprintf(stderr, 
	  "Exactly one source (-i, -r) must be given. If no sink (-o, -w)\n"
	  "is specified, standard out is used. In caching mode, enables\n"
	  "the reliable delivery of a signal when the receiver might get\n"
	  "disconnected temporarily, provided this client does not get\n"
	  "disconnected from the signal.\n");
  fprintf(stderr, "\n");
}

Signal *outSig = NULL;
FILE *outfile = NULL;
char *auth = NULL;

FILE *infile = NULL;

int fast = 0;

void got_signal_data(char *name, struct timeval timestamp, void *buffer,
		     Signal *inSig)
{
  if (outSig)
    {
      SignalGenAccept(outSig);
      SignalSend(outSig, timestamp, buffer);
    }
  if (outfile)
    {
      lock(fileLock);
      fprintf(outfile, "vector ");
      if (name)
	fprintf(outfile, "%s ", name);
      PrintSignalElements(outfile, inSig, timestamp, buffer);
      if (auth)
	SignalRecvd(inSig, timestamp);
      fflush(outfile);
      unlock(fileLock);
    }
}

void loop_from_file()
{
  int timeset = 0;
  long offset;
  struct timeval timestamp;
  void *buffer = malloc(SignalBufferSize(outSig));
  do
    {
      if (fscanf(infile, " vector ") < 0)
	{
	  exit(0);
	}
      while (fgetc(infile) != ' ')
	;
      if (!ScanSignalElements(infile, outSig, &timestamp, buffer))
	{
	  printf("Error reading file\n");
	  exit(1);
	}
      if (!fast)
	{
	  struct timeval now;
	  if (!timeset)
	    {
	      gettimeofday(&now, NULL);
	      offset = timeval_sub(&now, &timestamp);
	      timeset = 1;
	    }
	  else
	    {
	      struct timeval until = timestamp;
	      timeval_add(&until, offset);
	      wait_until(&until, NULL);
	    }
	}
      got_signal_data(NULL, timestamp, buffer, NULL);
    }
  while (1);
}

int proxy = 0;

void loop_from_signal(char *name)
{
  struct timeval timestamp;
  void *buffer = NULL;
  Signal *inSig = NewSignalUNL(name);
  do
    {
      // Loop on getting data
      do
	{
	  // Loop on connecting
	  while (!SignalConnected(inSig))
	    {
	      //lock(ipcLock);
	      printf("Waiting for %s\n", name);
	      if (!(proxy ? SignalConnectByProxy(inSig) :
		    SignalConnectWait(inSig)))
		{
		  return;
		}
	      printf("Done waiting for %s\n", name);
	      //unlock(ipcLock);
	      if (!buffer)
		{
		  printf("Properties set\n");
		  buffer = malloc(SignalBufferSize(inSig));
		  if (outSig)
		    SignalCopyProperties(inSig, outSig);
		}
	    }
	}
      while (!SignalRecv(inSig, &timestamp, buffer));
      got_signal_data(name, timestamp, buffer, inSig);
    }
  while (1);
}

int main(int argc, char **argv)
{
  int opt;
  char *inUNL = NULL;
  DArray inUNLs = NULL;
  char *outUNL = NULL;
  char *inFile = NULL;
  char *outFile = NULL;
  int append = 0;
  int help = 0;
  void *buffer = NULL;
  struct timeval timestamp;
  struct timeval base;

  do
    {
      opt = getopt(argc, argv, "I:i:o:r:w:a:t:c:s:fp");
      if (opt == 's')
	{
	  SetDefaultServer(optarg);
	}
      else if (opt == 'i')
	{
	  if (inFile)
	    {
	      fprintf(stderr, "Only one input source allowed\n\n");
	      help = 1;
	    }
	  if (inUNLs || inUNL)
	    {
	      if (!inUNLs)
		{
		  inUNLs = ArrayNew(sizeof(char *), 2, free);
		  ArrayAppend(inUNLs, &inUNL);
		  inUNL = NULL;
		}
	      ArrayAppend(inUNLs, &optarg);
	    }
	  else
	    {
	      inUNL = optarg;
	    }
	}
      else if (opt == 'p')
	{
	  proxy = 1;
	}
      else if (opt == 'o')
	{
	  if (outUNL)
	    fprintf(stderr, "Only one output UNL currently supported\n");
	  outUNL = optarg;
	}
      else if (opt == 'r')
	{
	  if (inFile || inUNL)
	    {
	      fprintf(stderr, "Only one input source allowed\n\n");
	      help = 1;
	    }
	  inFile = optarg;
	}
      else if (opt == 'w')
	{
	  if (outFile)
	    fprintf(stderr, "Only one output file currently supported\n");
	  outFile = optarg;
	}
      else if (opt == 'a')
	{
	  if (outFile)
	    fprintf(stderr, "Only one output file currently supported\n");
	  outFile = optarg;
	  append = 1;
	}
      else if (opt == 'f')
	{
	  fast = 1;
	}
      else if (opt == 'c')
	{
	  auth = optarg;
	}
      else if (opt == 'I')
	{
          char *certname = malloc(strlen(optarg) + strlen(CERT_FILE) + 1);
          char *keyname = malloc(strlen(optarg) + strlen(KEY_FILE) + 1);
          strcpy(certname, optarg);
          strcpy(certname + strlen(optarg), CERT_FILE);
          strcpy(keyname, optarg);
          strcpy(keyname + strlen(optarg), KEY_FILE);
          if (!UseIdentity(certname, keyname))
            printf("Identity in %s and %s could not be used.\n",
                   certname, keyname);
	}
      else if (opt > 0)
	{
	  printf("Unknown option\n");
	  help = 1;
	}
    }
  while (opt > 0);
  if (!inUNL && !inFile && !inUNLs)
    {
      fprintf(stderr, "Input source required\n");
      help = 1;
    }
  /*
  if (!outUNL && !outFile)
    {
      fprintf(stderr, "At least one output sink required\n");
      help = 1;
    }
  */
  if (help)
    {
      usage(argv[0]);
      exit(0);
    }
  if (outUNL)
    {
      outSig = NewSignalGenUNL(outUNL);
      if (auth)
	SignalGenGetReceipts(outSig, auth);
    }
  if (outFile)
    {
      outfile = fopen(outFile, append ? "a" : "w");
    }
  if (!outFile && !outUNL)
    {
      outfile = stdout;
    }

  if (inUNL)
    {
      loop_from_signal(inUNL);
    }
  else if (inUNLs)
    {
      int i;
      pthread_mutex_init(&ipcLock, NULL);
      for (i = 0; i < ArrayLength(inUNLs) - 1; i++)
	{
	  pthread_t thread;
	  pthread_create(&thread, NULL,
			 (void *(*)(void *)) loop_from_signal, 
			 *(char **)ArrayNth(inUNLs, i));
	}
      loop_from_signal(*(char **)ArrayNth(inUNLs, i));
    }
  else if (inFile)
    {
      infile = fopen(inFile, "r");
      buffer = malloc(7);
      fread(buffer, 1, 7, infile);
      if (!strncmp(buffer, "vector ", 7))
	{
	  while (fgetc(infile) != ' ')
	    ;
	  ScanSignalElements(infile, outSig, &base, NULL);
	  if (SignalGetUnitSamples(outSig))
	    {
	      long diff;
	      if (fscanf(infile, " vector ") < 0)
		{
		  exit(0);
		}
	      while (fgetc(infile) != ' ')
		;
	      ScanSignalElements(infile, outSig, &timestamp, NULL);

	      diff = timeval_sub(&timestamp, &base);

	      if (diff)
		SignalSetSampleRate(outSig, 
				    1000000 * SignalGetUnitSamples(outSig) / 
				    timeval_sub(&timestamp, &base),
				    SignalGetUnitSamples(outSig));
	      printf("Sample rate: %d\n", SignalGetSampleRate(outSig));
	    }
	}
      free(buffer);
      buffer = NULL;
      fseek(infile, 0, SEEK_SET);
      loop_from_file();
    }
  return 0;
}
