/*************************************************
 * File: ench.c
 * Author: Daniel Barkalow
 * Birth Date: 12/30/02
 *
 * Project: Enchantment IPC
 * 
 * Based on Enchantment IPC Whitepaper
 *   by Edward Keyes.
 * 
 * Copyright (C) 2002 MIT Media Lab,
 *               2004 AWare Technologies, Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *************************************************/

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <enchantment/node.h>
#include <enchantment/whiteboard.h>
#include <enchantment/unl.h>
#include <ctype.h>
#include <enchantment/dstring.h>

#ifdef WANT_SSL
#include <enchantment/identity.h>
#endif

#define CERT_FILE "cert.pem"
#define KEY_FILE "key.pem"

void usageCommon(char *arg0)
{
  fprintf(stderr, "Usage: %s [options]", arg0);
}

void usage(char *arg0)
{
  usageCommon(arg0);
  fprintf(stderr, " command [command-args]\n");
  fprintf(stderr, " Commands:\n");
  fprintf(stderr, "  shell\n");
  fprintf(stderr, "  display\n");
  fprintf(stderr, "  secure\n");
  fprintf(stderr, "  publish\n");
  fprintf(stderr, "  unpublish\n");
  fprintf(stderr, "  subscribe\n");
  fprintf(stderr, "  lock\n");
  fprintf(stderr, "  getattr\n");
  fprintf(stderr, " Options:\n");
  fprintf(stderr, "  -s server-name: use the named context server; default\n"
	  "   is $ENCHANTHOST or \"localhost\"\n");
#ifdef WANT_SSL
  fprintf(stderr, "  -i identity-file-base: use {identity-file-base}" 
	  CERT_FILE "\n"
	  "   and {identity-file-base}" KEY_FILE 
	  " for the client identity info\n");
#endif
}

void usageSecure(char *arg0)
{
  usageCommon(arg0);
  fprintf(stderr, " secure UNL ...\n");
}

void usageChmod(char *arg0)
{
  usageCommon(arg0);
  fprintf(stderr, " chmod [cmdopts] UNL operation authentication...\n");
  fprintf(stderr, " Options:\n");
  fprintf(stderr, "  -a: allow all clients\n");
  fprintf(stderr, " Operations:\n");
  fprintf(stderr, "  r: read\n");
  fprintf(stderr, "  w: write\n");
  fprintf(stderr, "  l: lock\n");
  fprintf(stderr, "  a: administer\n");
}

void usageGetattr(char *arg0)
{
  usageCommon(arg0);
  fprintf(stderr, " getattr UNL ...\n");
}

void usageDisplay(char *arg0)
{
  usageCommon(arg0);
  fprintf(stderr, " display [cmdopts] UNL ...\n");
  fprintf(stderr, " Options:\n");
  fprintf(stderr, "  -q: display only the direct contents of the node \n"
	  "   without complicated formatting.\n");
  fprintf(stderr, "  -s: only use secure connections.\n");
}

void usagePublish(char *arg0)
{
  usageCommon(arg0);
  fprintf(stderr, " publish UNL type value\n");
  fprintf(stderr, " Types:\n");
  fprintf(stderr, "  string\n");
  fprintf(stderr, "  int\n");
  fprintf(stderr, "  float\n");
  fprintf(stderr, " Options:\n");
  fprintf(stderr, "  -q: produce no output\n");
  fprintf(stderr, "  -h: publish persistently\n");
  fprintf(stderr, "  -s: only use secure connections.\n");
}

void usageUpload(char *arg0)
{
  usageCommon(arg0);
  fprintf(stderr, " upload UNL <filename>\n");
  fprintf(stderr, "  If no filename is givenm, reads from standard in\n");
  fprintf(stderr, " Options:\n");
  fprintf(stderr, "  -q: produce no output\n");
  fprintf(stderr, "  -h: publish persistently\n");
}

void usageUnpublish(char *arg0)
{
  usageCommon(arg0);
  fprintf(stderr, " unpublish UNL\n");
  fprintf(stderr, "  -s: only use secure connections.\n");
}

void usageSubscribe(char *arg0)
{
  usageCommon(arg0);
  fprintf(stderr, " subscribe UNL\n");
  fprintf(stderr, " Options:\n");
  fprintf(stderr, "  -q: display only the direct contents of the node \n"
	  "   without complicated formatting.\n");
  fprintf(stderr, "  -m: stay connected after getting updates.\n");
  fprintf(stderr, "  -s: only use secure connections.\n");
}

void usageLock(char *arg0)
{
  usageCommon(arg0);
  fprintf(stderr, " lock [cmdopts] UNL\n");
  fprintf(stderr, " Options:\n");
  fprintf(stderr, "  -r: lock recursively\n");
  fprintf(stderr, "  -R: remove node on client exit\n");
  fprintf(stderr, "  -u: unlock node\n");
}

void usageType(char *arg0)
{
  usageCommon(arg0);
  fprintf(stderr, " type [cmdopts] UNL [type]\n");
}

void setup()
{
  AttemptEncryption();
}

void done()
{
  // Not actually necessary these days, but could be useful if you wanted to
  // disconnect but keep running
  ShutDownClientIPC();
}

void waitForMessage(int quiet, int multi)
{
  GetUpdateInfo upInfo;
  do
    {
      // Get an update
      WaitForUpdate(&upInfo, NULL);

      // Print it
      if (quiet)
	PrintValue(upInfo.node, NULL);
      else
	{
	  printf("Found out about a publication to %s by %s\n",
		 upInfo.UNL, upInfo.publisher);
	  PrintNode(upInfo.node, NULL);
	  fflush(stdout);
	}
      FreeNode(upInfo.node);
    }
  while (multi);// && node);
}

int doCommand(int argc, char **argv)
{
  int opt;
  char *command = argv[0];
  char **cmdargv = argv;
  int cmdargc = argc;
  char *unl;
  char *name;
  char *typename;
  int quiet = 0;
  int multi = 0;
  int wait = 0;
  int secure = 0;
  NodeType *node;
  DataType data;
  int type;
  if (!strcmp(command, "display"))
    {
      optind = 0;
      do
 	{
	  opt = getopt(cmdargc, cmdargv, "+qws");
	  if (opt == 'q')
	    {
	      quiet = 1;
	      // This makes the client library not print stuff
	      WhiteboardVerbosity(0);
	    }
	  if (opt == 'w')
	    {
	      wait = 1;
	    }
	  if (opt == 's')
	    {
	      secure = 1;
	    }
	} 
      while (opt > 0);
      if (cmdargc < optind + 1)
	{
	  usageDisplay(argv[0]);
	  return -1;
	}
      setup();
      do
	{
	  unl = cmdargv[optind];
	  if (!secure || SecureUNL(unl))
	    {
	      // Actual operation to fetch information from the server
	      // XXXX use 'wait'
	      node = GetNode(unl, NULL);
	      if (!node)
		{
		  fprintf(stderr, "Error getting %s\n", unl);
		  return 2;
		}
	      if (!quiet)
		printf("value of %s\n", unl);
	      if (quiet)
		PrintValue(node, NULL);
	      else
		PrintNode(node, NULL);
	      FreeNode(node);
	    }
	  else
	    {
	      fprintf(stderr, "Secure connection not available to %s.\n",
		      unl);
	    }
	  optind++;
	}
      while (cmdargv[optind]);
      return 0;
    }
  if (!strcmp(command, "getattr"))
    {
      /*
      int flags = 0;
      GetNodeAttrReturnType ret;
      optind = 0;
      do
	{
	  opt = getopt(cmdargc, cmdargv, "+");
	}
      while (opt > 0);
      if (cmdargc < optind + 1)
	{
	  usageGetattr(argv[0]);
	}
      if (!GetNodeAttrF(cmdargv[optind], flags, &ret))
	{
	  printf("Could not get attributes for %s.\n", cmdargv[optind]);
	}
      else
	{
	  if (ret.lock.owner)
	    {
	      printf("Locked by %s%s.\n",
		     ret.lock.owner,
		     ret.lock.RECURSE ?
		     (ret.lock.REMOVE ? " remove" : " recursive") : "");
	    }
	  if (ret.permissions)
	    {
	      printf("Permissions: %s\n", ret.permissions);
	    }
	}
      return 0;
      */
    }
  if (!strcmp(command, "lock"))
    {
      LockUNLInfo info;
      optind = 0;
      LockUNL(NULL, &info);
      do
	{
	  opt = getopt(cmdargc, cmdargv, "+rRu");
	  if (opt == 'r')
	    {
	      info.common.operationFlags = ENCHANT_LOCK_RECURSIVE;
	    }
	  else if (opt == 'R')
	    {
	      info.common.operationFlags = ENCHANT_LOCK_REMOVE;
	    }
	  else if (opt == 'u')
	    {
	      info.common.operationFlags = ENCHANT_LOCK_NONE;
	    }
	}
      while (opt > 0);
      if (cmdargc < optind + 1)
	{
	  usageLock(argv[0]);
	  return -1;
	}
      if (!LockUNL(cmdargv[optind], &info))
	{
	  printf("Error locking %s\n", cmdargv[optind]);
	}
      return 0;
    }
  if (!strcmp(command, "chmod"))
    {
      int all = 0;
      optind = 0;
      do
	{
	  opt = getopt(cmdargc, cmdargv, "+a");
	  if (opt == 'a')
	    all = 1;
	}
      while (opt > 0);
      if (cmdargc < optind + 2)
	{
	  usageChmod(argv[0]);
	}
      setup();
      //printf("Set %c for %s\n", *(cmdargv[optind + 1]), cmdargv[optind]);
      if (!SetUNLPermissions(cmdargv[optind], *(cmdargv[optind + 1]), 
			     all ? NULL : (cmdargv + optind + 2)))
	{
	  printf("Chmod failed\n");
	}
      return 0;
    }
  if (!strcmp(command, "type"))
    {
      optind = 0;
      do
	{
	  opt = getopt(cmdargc, cmdargv, "+");
	}
      while (opt > 0);
      if (cmdargc < optind)
	{
	  usageType(argv[0]);
	}
      setup();
      if (cmdargc < optind + 2)
	{
	  if (!SetUNLFixedType(cmdargv[optind], T_NONE))
	    printf("Clearing type failed.\n");
	}
      else
	{
	  printf("Setting type.\n");
	  if (!SetUNLFixedType(cmdargv[optind],
			       TypeByName(cmdargv[optind + 1])))
	    printf("Setting type failed.\n");
	}
      return 0;
    }
  if (!strcmp(command, "secure"))
    {
      optind = 0;
      do
 	{
	  opt = getopt(cmdargc, cmdargv, "+");
	} 
      while (opt > 0);
      if (cmdargc < optind + 1)
	{
	  usageSecure(argv[0]);
	  return -1;
	}
      setup();
      do
	{
	  unl = cmdargv[optind];
	  printf("%s %s secure.\n", unl, SecureUNL(unl) ? "is" : "is not");
	  optind++;
	}
      while (cmdargv[optind]);
      return 0;
    }
  if (!strcmp(command, "unpublish"))
    {
      PublishNodeInfo info;
      optind = 0;
      PublishNode(NULL, NULL, &info);
      do
	{
	  opt = getopt(cmdargc, cmdargv, "+hqs");
	  if (opt == 'q')
	    {
	      quiet = 1;
	      WhiteboardVerbosity(0);
	    }
	  if (opt == 'h')
	    {
	      info.common.operationFlags |= ENCHANT_PUBLISH_HARD;
	    }
	  if (opt == 's')
	    {
	      secure = 1;
	    }
	}
      while (opt > 0);
      if (cmdargc < optind + 1)
	{
	  usageUnpublish(argv[0]);
	  return -1;
	}
      unl = cmdargv[optind];
      setup();
      if (secure && !SecureUNL(unl))
	{
	  fprintf(stderr, "Secure connection not available to %s\n", unl);
	  return -1;
	}
      if (PublishNode(unl, NULL, &info))
	{
	  fprintf(stderr, "Error unpublishing %s\n", unl);
	  return 2;
	}
      if (!quiet)
	printf("Unpublished %s\n", unl);
      return 0;
    }
  if (!strcmp(command, "subscribe"))
    {
      SubscribeUNLInfo info;
      SubscribeUNL(NULL, &info);
      optind = 0;
      do
 	{
	  opt = getopt(cmdargc, cmdargv, "+qms");
	  if (opt == 'q')
	    {
	      quiet = 1;
	      WhiteboardVerbosity(0);
	    }
	  if (opt == 'm')
	    {
	      multi = 1;
	    }
	  if (opt == 's')
	    {
	      secure = 1;
	    }
	}
      while (opt > 0);
      if (cmdargc < optind + 1)
	{
	  usageSubscribe(argv[0]);
	  return -1;
	}
      unl = cmdargv[optind];
      setup();
      // Subscribe to UNL
      if (secure && !SecureUNL(unl))
	{
	  fprintf(stderr, "Secure connection not available to %s\n", unl);
	  return -1;
	}
      if (!SubscribeUNL(unl, &info))
	{
	  fprintf(stderr, "Error subscribing to %s\n", unl);
	  return 2;
	}
      waitForMessage(quiet, multi);
      return 0;
    }
  if (!strcmp(command, "publish"))
    {
      PublishNodeInfo info;
      PublishNode(NULL, NULL, &info);
      optind = 0;
      do
 	{
	  opt = getopt(cmdargc, cmdargv, "+qhrams");
	  if (opt == 'q')
	    {
	      quiet = 1;
	      WhiteboardVerbosity(0);
	    }
	  if (opt == 'h')
	    {
	      info.common.operationFlags |= ENCHANT_PUBLISH_HARD;
	    }
	  if (opt == 'r')
	    {
	    }
	  if (opt == 'a')
	    {
	      info.common.operationFlags |= ENCHANT_PUBLISH_ADDONLY;
	    }
	  if (opt == 'm')
	    {
	      info.common.operationFlags |= ENCHANT_PUBLISH_MERGE;
	    }
	  if (opt == 's')
	    secure = 1;
	}
      while (opt > 0);
      if (cmdargc < optind + 2)
	{
	  usagePublish(argv[0]);
	  return -1;
	}
      unl = cmdargv[optind];
      name = "";
      type = -1;
      node = NULL;
      typename = cmdargv[optind + 1];
      type = TypeByName(typename);
      if (((type != T_DISTINCT && type != -1 &&
	    type != T_BRANCH) &&
	   (cmdargc < optind + 3)) ||
	  ((type == T_MATRIX) && (cmdargc < optind + 4)))
	{
	  // we need the third argument
	  usagePublish(argv[0]);
	  return -1;
	}
      if (type == T_SERVER)
	{
	  unsigned port = 11111;
	  char *svr;
	  char *delim = strchr(cmdargv[optind + 2], ':');
	  if (!delim)
	    delim = strchr(cmdargv[optind + 2], '@');
	  if (delim)
	    {
	      svr = delim + 1;
	      port = atoi(cmdargv[optind + 2]);
	    }
	  else
	    svr = cmdargv[optind + 2];
	  data.netsock_val = CreateNetSocketLocation(svr, port);
	  if (!data.netsock_val)
	    return -1;
	}
      else if (type == T_STRING)
	{
	  data.string_val = strdup(cmdargv[optind + 2]);
	}
      else if (type == T_SYMLINK)
	{
	  data.string_val = strdup(cmdargv[optind + 2]);
	}
      else if (type == T_INT32)
	{
	  data.int_val = atoi(cmdargv[optind + 2]);
	}
      else if (type == T_CHARACTER)
	{
	  data.int_val = cmdargv[optind + 2][0];
	}
      else if (type == T_FLOAT)
	{
	  data.float_val = (float)atof(cmdargv[optind + 2]);
	}
      else if (type == T_DOUBLE)
	{
	  data.doubleptr_val = malloc(sizeof(double));
	  *data.doubleptr_val = atof(cmdargv[optind + 2]);
	}
      else if (type == T_DISTINCT)
	{
	  node = CreateBranchNode(name, type);
	}
      else if (type == T_BRANCH)
	{
	  node = CreateBranchNode(name, type);
	}
      else if (type == T_VECTOR)
	{
	  data.vector_val = ReadVector(stdin, atoi(cmdargv[optind + 2]));
	}
      else if (type == T_MATRIX)
	{
	  data.matrix_val = ReadMatrix(stdin, atoi(cmdargv[optind + 2]),
				       atoi(cmdargv[optind + 3]));
	}
      else
	{
	  if (type)
	    printf("Unhandled type: %x\n", type);
	  else
	    printf("Unknown type: %s\n", typename);
	  return -1;
	}
      if (!node)
	node = CreateTypedNode(name, type, &data);
      setup();
      if (secure && !SecureUNL(unl))
	{
	  fprintf(stderr, "Secure connection not available to %s\n", unl);
	  return -1;
	}
      if (!PublishNode(unl, node, &info))
	{
	  fprintf(stderr, "Error publishing to %s\n", unl);
	  PrintNode(node, NULL);
	  FreeNode(node);
	  return 2;
	}
      if (!quiet)
	{
	  printf("Published to %s\n", unl);
	  PrintNode(node, NULL);
	}
      FreeNode(node);
      return 0;
    }
  if (!strcmp(command, "upload"))
    {
      /*
      Socket *sendSock;
      char *buffer = malloc(1024);
      int infd;
      int cnt;

      optind = 0;
      do
 	{
	  opt = getopt(cmdargc, cmdargv, "+qh");
	  if (opt == 'q')
	    {
	      quiet = 1;
	      WhiteboardVerbosity(0);
	    }
	  if (opt == 'h')
	    {
	      hard = 1;
	    }
	}
      while (opt > 0);
      if (cmdargc < optind + 2)
	{
	  usageUpload(argv[0]);
	  return -1;
	}
      unl = cmdargv[optind];
      CreateDocument(unl, &node, &sendSock);
      infd = open(cmdargv[optind + 1], O_RDONLY);
      while ((cnt = read(infd, buffer, 1024)) > 0)
	{
	  SocketSend(sendSock, buffer, cnt);
	}
      SocketClose(sendSock);
      PublishNode(unl, node, PUB_REPLACE, TRUE);
      if (!quiet)
	{
	  printf("Created handle at %s:\n", unl);
	  PrintNode(node, NULL);
	}
      free(buffer);
      */
      return 0;
    }
  fprintf(stderr, "Unknown command: %s\n", command);
  return -1;
}

int commandLoop()
{
  char cmd[256];
  char *argv[256];
  int argc;
  int i;
  int initial = 1;
  do
    {
      printf("> ");
      *cmd = '\0';
      fgets(cmd, 255, stdin);
      argc = 0;
      if (!*cmd)
	printf("\n");
      for (i = 0; cmd[i]; i++)
	{
	  if (isspace(cmd[i]))
	    {
	      if (!initial)
		{
		  argc++;
		  initial = 1;
		}
	      cmd[i] = '\0';
	    }
	  else if (initial)
	    {
	      initial = 0;
	      argv[argc] = cmd + i;
	    }
	}
      if (*cmd)
	{
	  argv[argc] = NULL;
	  /*
	  for (i = 0; i < argc; i++)
	    printf("%s ", argv[i]);
	  printf("\n");
	  */
	  doCommand(argc, argv);
	}
    }
  while (*cmd);
  return 0;
}

int main(int argc, char **argv)
{
  int opt;
  int ret;
  do
    {
      opt = getopt(argc, argv, "+s:i:");
      // If the command line has a server on it, 
      // set the default server to match
      if (opt == 's')
	SetDefaultServer(optarg);
      if (opt == 'i')
	{
#ifdef WANT_SSL
	  char *certname = malloc(strlen(optarg) + strlen(CERT_FILE) + 1);
	  char *keyname = malloc(strlen(optarg) + strlen(KEY_FILE) + 1);
	  strcpy(certname, optarg);
	  strcpy(certname + strlen(optarg), CERT_FILE);
	  strcpy(keyname, optarg);
	  strcpy(keyname + strlen(optarg), KEY_FILE);
	  if (!UseIdentity(certname, keyname))
	    printf("Identity in %s and %s could not be used.\n",
		   certname, keyname);
#else
	  printf("Encryption not supported.\n");
#endif
	}
    }
  while (opt > 0);
  if (argc <= optind)
    {
      usage(argv[0]);
      return -1;
    }
  if (!strcmp(argv[optind], "shell"))
    ret = commandLoop();
  else
    ret = doCommand(argc - optind, argv + optind);
  done();
  return ret;
}
